unit krpkr;

interface

uses Sysutils, Windows, ZLib, uMisc;

function EvaluateKRPKR16(pboard16:pbytearray; eval:boolean): integer;


implementation


const
 tbsize = 8*48*64*64*16 div 8;   // 3145728 bytes

var
 ktdata: array[0..tbsize-1] of byte;

const
 C_A = 0;
 C_B = 1;
 C_C = 2;
 C_D = 3;
 C_E = 4;
 C_F = 5;
 C_G = 6;
 C_H = 7;

VROOKPOS: array[0..63] of byte = (
 3,3,3,3,3,3,3,3,
 3,2,2,2,2,2,2,3,
 3,2,1,1,1,1,2,3,
 3,2,1,0,0,1,2,3,
 3,2,1,0,0,1,2,3,
 3,2,1,1,1,1,2,3,
 3,2,2,2,2,2,2,3,
 3,3,3,3,3,3,3,3
);

VYPAWN: array[0..7] of byte = (
0,75,85,100,120,145,180,0);

apawncode: array[0..63] of byte = (
  99,99,99,99,99,99,99,99,
  99,99,99,99,99,99,99,99,
  99,99,99,99,99,99,99,99,
  15,14,13,12,99,99,99,99,
  11,10,09,08,99,99,99,99,
  07,06,05,04,99,99,99,99,
  03,02,01,00,99,99,99,99,
  99,99,99,99,99,99,99,99
 );

relativepos: array[0..23] of byte = (
 0,0,0,1,2,3,0,0,
 0,0,0,4,0,5,0,0,
 0,0,0,6,7,8,0,0
);

TO16: array[0..63] of byte = (
 00, 01, 02, 03, 04, 05, 06, 07,
 16, 17, 18, 19, 20, 21, 22, 23,
 32, 33, 34, 35, 36, 37, 38, 39,
 48, 49, 50, 51, 52, 53, 54, 55,
 64, 65, 66, 67, 68, 69, 70, 71,
 80, 81, 82, 83, 84, 85, 86, 87,
 96, 97, 98, 99, 100,101,102,103,
 112,113,114,115,116,117,118,119
);

function getindex(ak,ar,ap,bk,br:integer): integer;
var apx,ac,i: integer;
begin
   result := -1;
   apx := ap and 7;
   if apx >= C_E then begin
      ap := (ap and $f8) + (7 - (ap and 7));
      ak := (ak and $f8) + (7 - (ak and 7));
      bk := (bk and $f8) + (7 - (bk and 7));
      ar := (ar and $f8) + (7 - (ar and 7));
      br := (br and $f8) + (7 - (br and 7));
      //messagebeep(0);
   end;
   ac := apawncode[ap];
   if (ac = 99) or (distboard(ap,ak) <> 1) then exit;
   dec(bk,16);
   if (bk < 0) then begin
      result := -2;
      exit;
   end;
   i := ak-ap+12;
   if (i<0) or (i>23) then raise exception.create('index');
   ak := relativepos[i];
   result := ak + bk*8 + ar*(8*48) + (br*8*48*64) + (ac*8*48*64*64);
end;

function krpkr_ending(ak,ar,ap,bk,br:integer): integer;
var index,bit: integer;
begin
   index := getindex(ak,ar,ap,bk,br);
   if index < 0 then begin
      result := index;
      exit;
   end;
   bit := 1 shl (index and 7);
   index := index shr 3;
   if (ktdata[index] and bit) <> 0 then result := 1 else result := 0;
end;

procedure kt_opendata;
var
 f: file;
 buftemp: pointer;
begin
   try
    filemode := 0;
    getmem(buftemp,240000);
    assignfile(f,ExtractFilePath(paramstr(0)) + 'Delfi.dat');
    reset(f,1);
    blockread(f,buftemp^,237987);
    closefile(f);
    DecompressToUserBuf(buftemp,237987,@ktdata,3145728);
    freemem(buftemp);
   except
    messagebox(0,'Cannot find the file "Delfi.dat"','ERROR',MB_ICONERROR);
    halt;
   end;
end;

// ++++++++++++++++++++++++++++++++++++++++++++++++++


function EvaluateKRPKR16(pboard16:pbytearray; eval:boolean): integer;
var
 ak,bk,ar,br,ap,i,d: integer;
 apx,apy,bry,bkx,bky,aky: integer;
 saferank,isblack: boolean;
 t: byte;
begin
   ap := 0; ar := 0; br := 0;
   ak := 0; bk := 0;
   isblack := false;
   for i := 0 to 63 do begin
       t := pboard16[TO16[i]];
       if t = 0 then continue;
       case t of
        AKING: ak := i;
        BKING: bk := i;
        AROOK: ar := i;
        BROOK: br := i;
        APAWN: ap := i;
        BPAWN: begin ap := i; isblack := true; end;
       end;
   end;
   if isblack then begin
      ap := VFlipSquare8(ap);
      i := ak;
      ak := VFlipSquare8(bk);
      bk := VFlipSquare8(i);
      i := ar;
      ar := VFlipSquare8(br);
      br := VFlipSquare8(i);
   end;
   apx := ap and 7;  apy := ap shr 3;
   bry := br shr 3;
   aky := ak shr 3;
   bkx := bk and 7;  bky := bk shr 3;

   result := VROOKPOS[ar] - VROOKPOS[br] + VYPAWN[apy];
   if (apx = 0) or (apx = 7) then dec(result,10);
   result := result - distboard(ak,ap)*4 + distboard(bk,ap+8)*4;

   if not eval then begin
      i := krpkr_ending(ak,ar,ap,bk,br);
      if (i = 1) or (i = -2) then begin
         result := result*2;
         exit;
      end;
      if (i = -1) then begin
         result := 0;
         exit;
      end;
      if (i = 0) then dec(result,8);
   end;

   if (apy >= _R5) then begin
      if (bk = ap+8) or (bk = ap+16) then dec(result,8);
      exit;
   end;
   saferank := (bky = _R5) or (bky = _R6);
   if saferank then dec(result,4);
   d := abs(bkx-apx);
   if (apx <= C_B) or (apx >= C_G) then dec(d,2);
   if apy = _R4 then begin
      if (d > 2) and (distboard(ak,ap) <= 2) then inc(result,8);
      if (d <= 2) and (bry = _R8) and (aky < _R6) and saferank
         then dec(result,8);
   end;
   if apy = _R3 then begin
      if (d > 3) and (distboard(ak,ap) <= 2) then inc(result,8);
      if (d <= 3) and (bry = _R8) and (aky < _R6)
         and saferank then dec(result,8);
   end;
end;


initialization

 kt_opendata;

end.
