unit uBook;

interface

uses Sysutils, Windows, uMisc, uBoard, uGenMove;

type

TChessBook = class
private
 data: array [0..499000] of integer;
 nentry: integer;
public
 function GetBookMove(pcb:PChessbrd; idx,elo:integer; btm:boolean): integer;
 procedure LoadBook;
end;


implementation


function MoveTO16(m:integer): integer;
begin
  result := TO16X[m and 63] + (TO16X[(m shr 6) and 63] shl 7);
end;

function TChessBook.GetBookMove(pcb:PChessbrd; idx,elo:integer; btm:boolean): integer;
var
 i,index,hb,hp,moves,m: integer;
 cboard: TChessBrd;
 mfound: array[0..63] of integer;
 weakmode: boolean;
const
 MAXIDXBOOK = 25*2;
 BOOKBAD = $40000000;
 BOOKWEAK = $20000000;
begin
  result := 0;
  index := MAXIDXBOOK;
  weakmode := (elo >= 1000) and (elo <= 2300);
  if weakmode then index := Elo div 100 + random(10);
  if (nentry < 1) or (idx >= index) then exit;
  cboard := pcb^;
  if btm then BoardVerticalFlipSide(@cboard);
  BoardUpdateData(@cboard);
  hb := cboard.HashB1;
  hp := cboard.HashP1;
  moves := 0;
  for i := 0 to nentry-1 do begin
      index := i*3;
      if (hb <> data[index+2]) or (hp <> data[index+1]) or (moves > 50)
         then continue;
      m := data[index+3];
      if (m >= BOOKBAD) then begin
         if not weakmode then continue;
         mfound[moves] := m;
         inc(moves);
         continue;
      end;
      mfound[moves] := m;
      inc(moves);
      if (m >= BOOKWEAK) and not weakmode then continue;
      mfound[moves] := m;
      inc(moves);
      if (m >= BOOKWEAK) then continue;
      mfound[moves] := m;
      inc(moves);
  end;
  m := 0;
  if moves > 0 then m := mfound[random(moves)] and $fff;
  if btm and (m <> 0) then m := InvertMove8(m);
  m := MoveTO16(m and $1fff);
  if not CheckMoveX(pcb,m,btm) then m := 0;
  result := m;
end;

procedure TChessBook.LoadBook;
var
 sname: string;
 f: file;
 nbytes: integer;
begin
  nentry := 0;
  sname := ExtractFilePath(ParamStr(0)) + 'Delfi.bok';
  if not FileExists(sname) then begin
     nentry := 0;
     exit;
  end;
  FileMode := 0;
  AssignFile(f,sname);
  Reset(f,1);
  nbytes := filesize(f);
  BlockRead(f,data,nbytes);
  CloseFile(f);
  nentry := data[0];
end;


end.
