unit uBrdList;

interface

uses SysUtils, Windows, uBoard, uMisc, uEval, uPriV;

const
 MAXBRDLIST = 500;
 MAXPV = 64;

type

 TPonderStatus = (psNone,psWait,psFail,psHit);

 TPollFunction = procedure(bl:pointer; dostats:boolean);

 TChessBoardList = class
 public
  PCurBrd: PChessBrd;
  PV: TPrincipalVar;
  PollFunction: TPollFunction;
  PollParam: pointer;
  PonderStatus: TPonderStatus;
  Idx,SearchNodes,D1MoveCount,D1Current,D1Depth,D1Score,Cut1,CutX,PollCnt: integer;
  TickStart,TickAbort: cardinal;
  AbortSearch,Thinking,WBAbsoluteScore,D1BlackToMove: boolean;
  PBrdList: array[0..MAXBRDLIST] of PChessBrd;
  BrdList: array[0..MAXBRDLIST] of TChessBrd;
  PVMoves: array[0..MAXPV] of integer;
  D1Moves,D1Scores: array[0..99] of integer;
  WBStatPonder: string;
  constructor Create;
  destructor Destroy; override;
  function SetPosition(sfen:string; var wmove:boolean): boolean;
  procedure PlayMove(m:integer);
  procedure TakeBack;
  procedure StorePV(offs:integer);
  function IsRepetition(triple: boolean): boolean;
  function Search(timethink,timegame,maxdepth,elo:integer; btm,ponder:boolean): integer;
  function SearchEasyMove(btm: boolean): integer;
  function SearchRoot(depth:integer; btm:boolean): integer;
  function SearchTree(depth,aa,bb:integer; btm,isnull:boolean; incheck:integer): integer;
  function SearchQuiesce(aa,bb:integer; btm:boolean): integer;
  function CountNodes(depth,incheck:integer; btm:boolean): int64;
 private
  SearchIdx,SearchMaxDepth,StatCnt,EloSearch: integer;
  MateKillers,PosKillers: array[0..63] of integer;
  procedure ClearKillers;
  procedure EvaluateMoves(var movelist:array of integer; n:integer; btm:boolean);
 end;

 function WBInfoString2(bl: TChessBoardList): string;
 function UCIInfoString2(bl: TChessBoardList): string;
 function WBInfoStringX(bl: TChessBoardList): string;
 function UCIInfoStringX(bl: TChessBoardList): string;

 
implementation


uses uGenMove, uGenCaps, uSee, uTable, uEndgame;

{ TBoardList }


procedure QSort(var arr:array of integer; n:integer); forward;

constructor TChessBoardList.Create;
var
 i: integer;
 wmove: boolean;
begin
  inherited Create;
  for i := 0 to High(PBrdList) do PBrdList[i] := @BrdList[i];
  SetPosition('',wmove);
  PV := TPrincipalVar.Create;
  @PollFunction := nil;
  PollParam := nil;
  WBStatPonder := '';
end;

destructor TChessBoardList.Destroy;
begin
  PV.Free;
  inherited;
end;

function TChessBoardList.SetPosition(sfen:string; var wmove:boolean): boolean;
begin
  Idx := 0;
  PCurBrd := PBrdList[0];
  result := BoardSetFEN(PCurBrd,sfen,wmove);
end;

function TChessBoardList.CountNodes(depth,incheck:integer; btm:boolean): int64;
var
 i,n: integer;
 movelist: array[0..99] of integer;
begin
  result := 0;
  if depth <= 0 then exit;
  if btm
     then n := BGenerateX(PCurBrd,@movelist,incheck,true)
     else n := AGenerateX(PCurBrd,@movelist,incheck,true);
  if n < 1 then exit;
  if depth > 1 then
  for i := 0 to n-1 do begin
      PlayMove(movelist[i]);
      incheck := BoardMoveChecked(PCurBrd,movelist[i]);
      result := result + CountNodes(depth-1,incheck,not btm);
      TakeBack;
  end
  else result := result + n;
end;

procedure TChessBoardList.ClearKillers;
var i: integer;
begin
  for i := 0 to 63 do begin
      MateKillers[i] := 0;
      PosKillers[i] := 0;
  end;
end;

function TChessBoardList.Search(timethink,timegame,maxdepth,elo:integer;
                                btm,ponder:boolean): integer;
var
 t,i,depth,movechk,easymove: integer;
 ct: single;
 analyze: boolean;
begin
  result := 0;
  EloSearch := elo;
  analyze := false;
  D1BlackToMove := btm;
  HashTable.ClearHistory;
  if timethink < 0 then begin
     timethink := 3600*24*1000;
     timegame := timethink;
     analyze := true;
  end
  else begin
     if ponder then begin
        PonderStatus := psWait;
        timethink := round(timethink * 0.74);
     end
     else begin
        PonderStatus := psNone;
        timethink := round(timethink * 0.68);
     end;
  end;
  ct := 0.3;
  t := timethink * 12;
  if t > timegame then t := round(timegame*0.80);
  TickStart := GetTickCount;
  TickAbort := TickStart + cardinal(t);
  if (maxdepth <= 0) or (maxdepth > 40) then maxdepth := 40;
  t := 0;
  SearchNodes := t; D1MoveCount := t; StatCnt := t; PollCnt := t;
  depth := t; D1Moves[0] := t; PVMoves[0] := t; PVMoves[1] := t;
  D1Current := 1;
  Cut1 := 1; CutX := 1;
  AbortSearch := false;
  SearchIdx := Idx;
  PV.Reset;
  HashTable.Clear(false);
  BoardUpdateData(PCurBrd);
  for i := 0 to High(D1Scores) do D1Scores[i] := 0;
  movechk := UnderCheckX(PCurBrd,btm);
  if btm
     then D1MoveCount := BGenerateX(PCurBrd,@D1Moves,movechk,true)
     else D1MoveCount := AGenerateX(PCurBrd,@D1Moves,movechk,true);
  easymove := SearchEasyMove(btm);
  repeat
    ClearKillers;
    t := SearchRoot(depth,btm);
    if (depth = 0) or (not AbortSearch) then result := t;
    if PV.PVMove(0) <> easymove then ct := 1;
    inc(depth);
    if (depth >= maxdepth) or
       (GetTickCount >= TickStart + round(timethink*ct)) then break;
    if (depth >= 4) and (not analyze) and
       ((D1MoveCount = 1) or (result > VALMOSTMATE)) then break;
    HashTable.InsertPV(PCurBrd,PV,btm);
  until AbortSearch;
  TickAbort := GetTickCount - TickStart;
  while ponder and (PonderStatus = PSWait)
     do PollFunction(PollParam,false);
end;

function TChessBoardList.SearchEasyMove(btm:boolean): integer;
var
 i,n,m,v1,v2: integer;
 scores: array[0..99] of integer;
begin
  result := 0;
  n := D1MoveCount;
  if n < 2 then exit;
  for i := 0 to n-1 do begin
      m := D1Moves[i];
      PlayMove(m);
      scores[i] := -SearchQuiesce(-VINFINITE,VINFINITE,not btm);
      TakeBack;
  end;
  i := FindMaxArray(scores,n);
  v1 := scores[i];
  scores[i] := -VINFINITE;
  v2 := scores[FindMaxArray(scores,n)];
  if v1 >= v2+200 then result := D1Moves[i];
end;

function TChessBoardList.SearchRoot(depth:integer; btm:boolean): integer;
var
 i,j,v,m,n,aa,ibest,nd,movechk: integer;
const
 BIGSCORE = $40000000;
 FAILHIGH = 64;
label
 SkipRepRoot;
begin
  inc(SearchNodes);
  D1Depth := depth + 1;
  SearchMaxDepth := depth + 12;
  ibest := 0;
  aa := -VINFINITE;
  n := D1MoveCount;
  if n < 0 then begin
     if n = -1 then result := -VMATE else result := -001;
     exit;
  end;
  for i := 0 to n-1 do begin
      nd := SearchNodes;
      D1Current := i+1;
      m := D1Moves[i];
      PlayMove(m);
      if PCurBrd.DrawingMoves > 3 then
      if IsRepetition(false) then begin
         v := 0;
         pv.Leaf(1);
         goto SkipRepRoot;
      end;
      movechk := BoardMoveChecked(PCurBrd,m);
      if i = 0 then begin
         v := -SearchTree(depth-1,-VINFINITE,VINFINITE, not btm,false,movechk);
      end
      else begin
         v := -SearchTree(depth-1,-aa-1,-aa,not btm,false,movechk);
         if v > aa
            then v := -SearchTree(depth-1,-VINFINITE,-aa,not btm,false,movechk);
      end;
     SkipRepRoot:
      TakeBack;
      if AbortSearch then break;
      if v > aa then begin
         ibest := i;
         aa := v;
         D1Score := v;
         pv.Store(0,m);
         if ((D1Depth > 1) or (EloSearch <> MAXELO)) and
            Assigned(PollFunction) then PollFunction(PollParam,true);
      end;
      inc(D1Scores[i],(SearchNodes - nd) and (BIGSCORE-1));
  end;
  if D1Depth = 1 then inc(StatCnt);
  HashTable.IncHistory(D1Moves[ibest],depth,FAILHIGH,PCurBrd);
  result := aa;
  if n < 2 then exit;
  inc(D1Scores[ibest],BIGSCORE);
  for j := 2 to n do begin
      v := D1Scores[j-1];
      m := D1Moves[j-1];
      i := j-1;
      while (i > 0) and (D1Scores[i-1] < v) do begin
         D1Scores[i] := D1Scores[i-1];
         D1Moves[i] := D1Moves[i-1];
         dec(i);
      end;
      D1Scores[i] := v;
      D1Moves[i] := m;
  end;
  dec(D1Scores[0],BIGSCORE);
end;

procedure TChessBoardList.PlayMove(m: integer);
var
 pcb: PChessBrd;
 (*i: integer;
 s: string; *)
begin
  if Idx >= MAXBRDLIST then exit;
  inc(Idx);
  pcb := PBrdList[Idx];
  pcb^ := PCurBrd^;
  PCurBrd := pcb;
  if m = 0 then begin
     pcb.EPSquare := 255;
     pcb.DrawingMoves := 0;
     pcb.LastCapSquare := 255;
     pcb.LastMove := 0;
     exit;
  end;
  BoardPlayMove(pcb,m);
end;

procedure TChessBoardList.TakeBack;
begin
  if Idx = 0 then exit;
  dec(Idx);
  PCurBrd := PBrdList[Idx];
end;

function TChessBoardList.IsRepetition(triple:boolean): boolean;
var
 i,index,nfound,nmax,nrep,hb1,hp1: integer;
 pcb: PChessBrd;
begin
  result := false;
  index := Idx;
  if index < 2 then exit;
  if not triple then begin
     nmax := PCurBrd.DrawingMoves shr 1;
     if nmax > 20 then nmax := 20;
     nrep := 1;
  end
  else begin
     nmax := VINFINITE;
     nrep := 2;
  end;
  nfound := 0;
  hb1 := PCurBrd.HashB1;
  hp1 := PCurBrd.HashP1;
  for i := 1 to nmax do begin
      dec(index,2);
      if index < 0 then break;
      pcb := PBrdList[index];
      if (hb1 <> pcb.HashB1) or (hp1 <> pcb.HashP1) then continue;
      inc(nfound);
      if nfound < nrep then continue;
      result := true;
      break;
  end;
end;

procedure TChessBoardList.EvaluateMoves(var movelist: array of integer;
                                        n:integer; btm:boolean);
var
 i,m,v,s,vmax,imax,killerm,killerp: integer;
 pcb: PChessBrd;
 onlypawn: boolean;
const
 SCOREOFF = 100;
begin
  pcb := PCurBrd;
  vmax := -VINFINITE;
  killerm := MateKillers[Idx-SearchIdx];
  killerp := PosKillers[Idx-SearchIdx];
  onlypawn := pcb.AMatStage + pcb.BMatStage <= 16;
  if movelist[SCOREOFF] = 0 then s := 0 else s := 1;
  imax := s;
  for i := s to n-1 do begin
      m := movelist[i];
      v := 0;
      if (m and (MPROMOTION + MCAPTURE)) <> 0 then begin
         if not onlypawn
            then v := EstimateExchangeX(pcb,m,-VINFINITE,btm)
            else v := 50;
         if (m and MPROMOTEDQUEEN) = MPROMOTEDQUEEN then inc(v,700);
         if v > -150 then v := v * 256 + VINFINITE div 100;
      end;
      if m = killerm then v := VINFINITE div 20;
      if m = killerp then v := VINFINITE div 1000;
      if (m and MPROMOTIONTHREAT) <> 0 then inc(v,1000);
      inc(v,HashTable.GetHistory(m,PCurBrd));
      movelist[i+SCOREOFF] := v;
      if v > vmax then begin
         vmax := v;
         imax := i;
      end;
  end;
  v := movelist[imax]; movelist[imax] := movelist[s]; movelist[s] := v;
  v := movelist[imax+SCOREOFF]; movelist[imax+SCOREOFF] := movelist[s+SCOREOFF];
  movelist[s+SCOREOFF] := v;
end;

procedure DoubleFastSort(arr:pintarray; n:integer);
var
 i,j,a,b: integer;
const
 SCOREOFF = 100;
begin
  if n < 2 then exit;
  for j := 1 to n-1 do begin
      a := arr[j];
      b := arr[j-SCOREOFF];
      i := j;
      while (arr[i-1] < a) do begin
         arr[i] := arr[i-1];
         arr[i-SCOREOFF] := arr[i-(1+SCOREOFF)];
         dec(i);
      end;
      arr[i] := a;
      arr[i-SCOREOFF] := b;
  end;
end;

function TChessBoardList.SearchTree(depth,aa,bb:integer;
                         btm,isnull:boolean; incheck:integer): integer;
var
 i,n,v,vbest,hkind,movehash,movebest,m,mcount,deepidx: integer;
 incr,extension,movechk,matstage,pieces: integer;
 ispv,invmove,islazy,pruned: boolean;
 movelist: array[0..199] of integer;
const
 SCOREOFF = 100;
 MAXTREEDEPTH = 31;
 FAILHIGH = 64;
 FAILLOW = 128;
 TTFHIGH = 1;
 TTFLOW = 2;
 TTEXACT = 3;
 TTNONULL = 4;
label
 SkipRep,Verify;
begin
  deepidx := Idx-SearchIdx;
  if bb = aa+1 then ispv := false
  else begin
     ispv := true;
     pv.Leaf(deepidx);
  end;
  result := 0;
  if deepidx >= SearchMaxDepth then exit;
  if AbortSearch then if D1Depth > 1 then exit;
  with PCurBrd^ do begin
     matstage := AMatStage + BMatStage;
     pieces := (AMatCnt + BMatCnt) and 63;
  end;
  if (pieces <= 5) and (*deepidx > middepth*) (depth < 2) then begin
     v := (PCurBrd.BMatStage shl 16) + PCurBrd.AMatStage;
     if v = 0 then begin
        result := 0;
        exit;
     end;
     if IsKnownEG(v) then begin
        result := DBEndgameX(PCurBrd,btm,invmove);
        if not invmove then begin
           if btm then result := -result;
           exit;
        end;
     end;
  end;
  if incheck <> 0 then begin
     inc(depth);
     if depth < 0 then depth := 0;
  end;
  if depth < 0 then begin
     result := SearchQuiesce(aa,bb,btm);
     exit;
  end;
  inc(SearchNodes);
  if HashTable.GetPosition(PCurBrd,bb,depth,deepidx,btm,hkind,v,movehash)
  then begin
    result := v;
    case hkind of
     TTEXACT: if not ispv then exit;
     TTFHIGH: if v >= bb then exit;
     TTFLOW:
      begin
        if aa >= v then exit;
        if v >= bb then hkind := TTNONULL;
        movehash := 0;
      end;
    end;
  end;
  if (depth > 0) and (matstage > 550) and (hkind <> TTNONULL) and
     not (ispv or isnull or boolean(InCheck) (*or nullthreat*))
     and (bb < VALMOSTMATE) then
  begin
     PlayMove(0);
     if depth < 6 then v := 2 else v := 3;
     result := -SearchTree(depth-v-1,-bb,-bb+1,not btm,true,0);
     TakeBack;
     if result >= bb then begin
        if result > VALMOSTMATE then result := VALMOSTMATE;
        HashTable.StorePosition(PCurBrd,depth,TTFHIGH,result,0,deepidx,btm);
        exit;
     end;
  end;
  if btm
     then n := BGenerateX(PCurBrd,@movelist,incheck,false)
     else n := AGenerateX(PCurBrd,@movelist,incheck,false);
  if n < 0 then begin
     if n = -1
        then result := -VMATE + deepidx
        else result := -001;
     exit;
  end;
  if n > 1 then begin
     if (movehash = 0) and (depth >= 3) then begin
        v := depth-2;
        if incheck <> 0 then dec(v);
        SearchTree(v,aa,bb,btm,true,incheck);
        movehash := HashTable.LastMove and $3fff;
     end;
     if movehash = 0 then movehash := MateKillers[deepidx] and $3fff;
     movelist[SCOREOFF] := 0;
     if movehash <> 0 then
        for i := 0 to n-1 do begin
           if (movelist[i] and $3fff) <> movehash then continue;
           movehash := movelist[i];
           movelist[i] := movelist[0];
           movelist[0] := movehash;
           movelist[SCOREOFF] := VINFINITE;
           movehash := $ffff;
           break;
        end;
     if movehash <> $ffff then EvaluateMoves(movelist,n,btm);
  end;
  hkind := TTFLOW;
  vbest := -VINFINITE;
  movebest := 0;
  extension := 0;
  if (n = 1) and (incheck <> 0) then extension := 1;
  mcount := 0;
  for i := 0 to n-1 do begin
    incr := extension;
    if (i = 1) then begin
       if movehash = $ffff then EvaluateMoves(movelist,n,btm);
       DoubleFastSort(@movelist[SCOREOFF],n);
    end;
    v := PCurBrd.LastCapSquare;
    m := movelist[i];
    if (m and MKINGNLEGAL) <> 0 then begin
       if btm
          then invmove := BIsLegalMoveX(PCurBrd,m,false)
          else invmove := AIsLegalMoveX(PCurBrd,m,false);
       if not invmove then continue;
       movelist[i] := m and (not MKINGNLEGAL);
    end;
    PlayMove(m);
    inc(mcount);
    if PCurBrd.DrawingMoves > 3 then
    if IsRepetition(false) then begin
       v := 0;
       pv.Leaf(deepidx+1);
       goto SkipRep;
    end;
    if ispv then begin
       if (m and MPROMOTIONTHREAT <> 0) then inc(incr);
       if (m and 127) = v then begin
          inc(incr);
          if matstage < 50 then inc(incr,2);
       end;
    end;
    movechk := BoardMoveChecked(PCurBrd,m);
    if (depth = 1) and (mcount >= 1) and (not ispv) and (incheck = 0) and
       (extension = 0) and (movechk = 0) and
       (m and (MCAPTURE + MPROMOTION + MPROMOTIONTHREAT) = 0) then begin
       v := EvaluateRel(PCurBrd,VINFINITE,btm,islazy);
       if v + 200 < aa then goto SkipRep;
    end;
    pruned := false;
    if (depth >= 3) and (mcount >= 4) and (not ispv) and (incheck = 0) and
       (extension = 0) and (movechk = 0) and
       (m and (MCAPTURE + MPROMOTION + MPROMOTIONTHREAT) = 0) and
       HashTable.PruneHistory(m,PCurBrd) then
       begin
          pruned := true;
          dec(incr);
       end;
Verify:
    if i = 0 then
       v := -SearchTree(depth-1+incr,-bb,-aa,not btm,false,movechk)
    else begin
       v := -SearchTree(depth-1+incr,-aa-1,-aa,not btm,false,movechk);
       if (v > aa) and (v < bb) then
          v := -SearchTree(depth-1+incr,-bb,-aa,not btm,false,movechk);
    end;
    if pruned and (v >= bb) then begin
       inc(incr);
       pruned := false;
       goto Verify;
    end;
SkipRep:
    TakeBack;
    if AbortSearch then break;
    if v > vbest then vbest := v;
    if v > aa then begin
       movebest := m;
       aa := v;
       if v >= bb then begin
          hkind := TTFHIGH;
          if v <= VALMOSTMATE then begin
             if (m and (MPROMOTION + MCAPTURE) = 0)
                then PosKillers[deepidx] := m;
          end
          else MateKillers[deepidx] := m;
          HashTable.IncHistory(m,depth,FAILHIGH,PCurBrd);
          if (m and (MCAPTURE + MPROMOTION) = 0) then
             for v := 0 to i-1 do
                 HashTable.IncHistory(movelist[v],0,FAILLOW,PCurBrd);
          inc(CutX);
          if i = 0 then inc(Cut1);
          break;
       end;
       hkind := TTEXACT;
       pv.Store(deepidx,movebest);
    end;
    if PollCnt < 500 then inc(PollCnt)
    else begin
       PollCnt := 0;
       if GetTickCount > TickAbort then AbortSearch := true;
       if Assigned(PollFunction) then PollFunction(PollParam,false);
    end;
  end;
  if mcount <= 0 then begin
     result := -001;
     exit;
  end;
  MateKillers[deepidx+1] := 0;
  PosKillers[deepidx+1] := 0;
  if AbortSearch then exit;
  HashTable.StorePosition(PCurBrd,depth,hkind,vbest,movebest,deepidx,btm);
  result := vbest;
end;

function TChessBoardList.SearchQuiesce(aa,bb:integer; btm:boolean): integer;
var
 i,n,m,v,vtarget,nmoves,eval:integer;
 pboard: PByteArray;
 movelist: array[0..63] of integer;
 scorelist: array[0..19] of integer;
 islazy: boolean;
const
 VMATERIAL: array[0..7] of integer =
  (0, 0, VKING, VPAWN, VKNIGHT, VBISHOP, VROOK, VQUEEN);
begin
  inc(SearchNodes);
  if bb <> aa+1 then PV.Leaf(Idx-SearchIdx);
  result := EvaluateRel(pCurBrd,bb,btm,islazy);
  if (PCurBrd.AMatStage > VROOK) and (PCurBrd.BMatStage > VROOK) and
     (btm = D1BlackToMove) then inc(result,10);
  if result >= bb then exit;
  if result > aa then aa := result;
  eval := result;
  if btm
     then n := BGenerateCapX(pCurBrd,@movelist)
     else n := AGenerateCapX(pCurBrd,@movelist);
  if n < 1 then exit;
  nmoves := 0;
  pboard := @(PCurBrd.BoardArray[BOARDOFFSET]);
  for i := 0 to n-1 do begin
      m := movelist[i];
      v := pboard[m and 127];
      vtarget := VMATERIAL[v and 7];
      if vtarget = VKING then begin
         result := 9900;
         exit;
      end;
      if vtarget = 0 then vtarget := VPAWN;
      if (m and MPROMOTEDQUEEN) <> MPROMOTEDQUEEN then begin
         if ((v = APAWN) and (m and 127 >= _A7)) or
            ((v = BPAWN) and (m and 127 <= _H2)) then inc(vtarget,100);
         if (m and MPAWNQUICAP <> 0) then inc(vtarget,10);
      end
      else inc(vtarget,600);
      if eval + 120 + vtarget < aa then continue;
      dec(vtarget,VMATERIAL[pboard[(m shr 7) and 127] and 7] shr 5);
      scorelist[nmoves] := (5000 - vtarget) shl 8 + nmoves;
      movelist[nmoves] := m;
      inc(nmoves);
      if nmoves > 19 then break;
  end;
  if nmoves = 0 then exit;
  if nmoves > 1 then QSort(scorelist,nmoves);
  for i := 0 to nmoves-1 do begin
      m := movelist[scorelist[i] and $ff];
      if (m and MPAWNQUICAP = 0) then begin
         vtarget := EstimateExchangeX(PCurBrd,m,-VINFINITE,btm);
         if vtarget < -50 then continue;
         if eval + 120 + vtarget < aa then continue;
      end;
      PlayMove(m);
      v := -SearchQuiesce(-bb,-aa,not btm);
      TakeBack;
      if v > result then result := v;
      if v > aa then begin
         aa := v;
         if aa >= bb then break;
         PV.Store(Idx-SearchIdx,m);
      end;
  end;
end;

procedure QSort(var arr:array of integer; n:integer);
var i,j,a: integer;
begin
  for j := 2 to n do begin
      a := arr[j-1];
      i := j-1;
      while (i > 0) and (arr[i-1] > a) do begin
         arr[i] := arr[i-1];
         dec(i);
      end;
      arr[i] := a;
  end;
end;

procedure TChessBoardList.StorePV(offs: integer);
var
 i,c1,c2,m: integer;
 newpv,b: boolean;
begin
  c1 := SearchIdx + 1;
  c2 := Idx + offs;
  newpv := false;
  for i := c1 to c2 do begin
      m := PBrdList[i].LastMove;
      b := PVMoves[i-c1] <> m;
      newpv := newpv or b;
      PVMoves[i-c1] := m;
  end;
  if newpv then PVMoves[c2-c1+1] := 0;
end;

function WBInfoStringX(bl: TChessBoardList): string;
var t,v: integer;
begin
  t := integer(GetTickCount-bl.TickStart) div 10;
  v := bl.D1Score;
  if bl.WBAbsoluteScore and bl.D1BlackToMove then v := -v;
  result := format('%d %d %d %d',[bl.D1Depth,v,t,bl.SearchNodes]);
  if bl.EloSearch <> MAXELO then
     result := result + ' [' + inttostr(bl.EloSearch) + ' ELO]';
  if bl.WBStatPonder <> '' then begin
     result := result + ' [' + bl.WBStatPonder + ']';
     if bl.PonderStatus = psHit then result := result + '*';
  end;
  result := result + PVToStringSAN(bl.PCurBrd,bl.PV.PVArray);
end;

function UCIInfoStringX(bl: TChessBoardList): string;
var t: integer;
begin
  t := integer(GetTickCount-bl.TickStart);
  result := format('info depth %d score cp %d time %d nodes %d pv',
                   [bl.D1Depth,bl.D1Score,t,bl.SearchNodes]);
  result := result + PVToString(bl.PV.PVArray);
end;

function WBInfoString2(bl: TChessBoardList): string;
var n: integer;
begin
  n := bl.D1MoveCount;
  with bl do result := format('stat01:  %d %d %d %d %d',
       [(GetTickCount-TickStart) div 10,SearchNodes,D1Depth,n-D1Current,n]);
end;

function UCIInfoString2(bl: TChessBoardList): string;
var
 t: cardinal;
 nps: double;
begin
  t := GetTickCount - bl.TickStart;
  if t < 1 then t := 1;
  nps := t / 1000;
  nps := bl.SearchNodes / nps;
  if nps > 1e9 then nps := 1e9;
  with bl do result :=
       format('info nodes %d time %d nps %d currmovenumber %d currmove %s',
       [SearchNodes,t,round(nps),D1Current,MoveToString(D1Moves[D1Current-1])]);
end;

end.
