unit uEval;

interface

uses uMisc;

type

TPawnData = record
 VPStruct,VPos,CenterBlock: integer;
 APass,BPass,ACol,BCol,Pwb,Pbw,Shield,Storm: integer;
end;

TEvalData = record
 ABishopPos,ARookPos1,ARookPos2,AQueenPos,AKingPos: integer;
 BBishopPos,BRookPos1,BRookPos2,BQueenPos,BKingPos: integer;
 AMat,BMat,VPos,AClose,BClose: integer;
 Mater,Pos,AMatStage,BMatStage: integer;
 pBoard: pointer;
end;

TEvalCache = class
 private
   dummy: integer;
   data: array[0..262144-1] of integer;
 public
   procedure Reset;
   procedure Store(key1,key2,value:integer);
   function Get(key1,key2:integer): integer;
end;

TPawnCache = class
 private
  ftable,table: pintarray;
 public
  constructor Create;
  destructor Destroy; override;
  procedure StorePosition(hp1,hp2:integer; const pdat:TPawnData);
  function GetPosition(hp1,hp2:integer; var pdat:TPawnData): boolean;
  procedure Reset;
end;

TEvalFunction = procedure(var edat:TEvalData; var pdat:TPawnData);

function EvaluateRel(pcbe:pointer; pbeta:integer; btm:boolean;
                     var islazy:boolean): integer;
procedure EvaluateNoPawns(var edat:TEvalData; var pdat:TPawnData);
function EvaluateBishopEndgame(var edat:TEvalData; var pdat:TPawnData): integer;
procedure EvaluatePawns(var pdat:TPawnData; pboard:PByteArray; HP1,HP2:integer);
procedure EvaluatePassedPawns(var edat:TEvalData; var pdat:TPawnData);
function PawnRace(var edat:TEvalData; blacktomove:boolean;
                  apawns,bpawns:integer): integer;
procedure EvaluateAKnight(var edat:TEvalData; var pdat:TPawnData);
function AKnightVAttack(mempos:pointer): integer;
procedure EvaluateBKnight(var edat:TEvalData; var pdat:TPawnData);
function BKnightVAttack(mempos:pointer): integer;
procedure EvaluateABishop(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateBBishop(var edat:TEvalData; var pdat:TPawnData);
procedure AAttackBishop(pboard:pbytearray; pos:integer;
                          var amat,atkd:integer; var atkking: boolean);
procedure BAttackBishop(pboard:pbytearray; pos:integer;
                          var bmat,atkd:integer; var atkking: boolean);
function AttackRook(pboard:pbytearray; pos:integer): integer;
function AAttackedByPawn(pboard:pointer; pos:integer): integer;
function BAttackedByPawn(pboard:pointer; pos:integer): integer;
procedure EvaluateARook(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateBRook(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateAQueen(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateBQueen(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateNull(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluateRookQueen(var edat:TEvalData; var pdat:TPawnData);
procedure EvaluatePawnMiddlegame(var edat:TEvalData; var pdat:TPawnData);
function AQueenAttack(pboard:pbytearray; pos:integer; kside:boolean): boolean;
function BQueenAttack(pboard:pbytearray; pos:integer; kside:boolean): boolean;
Function ADevelop(pboard:pointer): integer;
Function BDevelop(pboard:pointer): integer;
procedure SetRandom(isrand:boolean);

var
 PawnCache: TPawnCache;
 EvalCache: TEvalCache;

 VKNIGHTPOS: array[0..63] of byte;

const
 COLORSQUARE: array[0..31] of byte =
 (0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,
  1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0);


implementation


uses uEndgame, uBoard;

var
 AEvalFunction,BEvalFunction: array[0..7] of TEvalFunction;

const
 TO8: array[0..127] of byte = (
 00, 01, 02, 03, 04, 05, 06, 07, 00, 00, 00, 00, 00, 00, 00, 00,
 08, 09, 10, 11, 12, 13, 14, 15, 00, 00, 00, 00, 00, 00, 00, 00,
 16, 17, 18, 19, 20, 21, 22, 23, 00, 00, 00, 00, 00, 00, 00, 00,
 24, 25, 26, 27, 28, 29, 30, 31, 00, 00, 00, 00, 00, 00, 00, 00,
 32, 33, 34, 35, 36, 37, 38, 39, 00, 00, 00, 00, 00, 00, 00, 00,
 40, 41, 42, 43, 44, 45, 46, 47, 00, 00, 00, 00, 00, 00, 00, 00,
 48, 49, 50, 51, 52, 53, 54, 55, 00, 00, 00, 00, 00, 00, 00, 00,
 56, 57, 58, 59, 60, 61, 62, 63, 00, 00, 00, 00, 00, 00, 00, 00
);

TO16: array[0..63] of byte = (
 00, 01, 02, 03, 04, 05, 06, 07,
 16, 17, 18, 19, 20, 21, 22, 23,
 32, 33, 34, 35, 36, 37, 38, 39,
 48, 49, 50, 51, 52, 53, 54, 55,
 64, 65, 66, 67, 68, 69, 70, 71,
 80, 81, 82, 83, 84, 85, 86, 87,
 96, 97, 98, 99, 100,101,102,103,
 112,113,114,115,116,117,118,119
);

 DIST16: array[0..239] of byte = (
 00, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07,
 00, 07, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 07,
 00, 07, 06, 05, 05, 05, 05, 05, 05, 05, 05, 05, 05, 05, 06, 07,
 00, 07, 06, 05, 04, 04, 04, 04, 04, 04, 04, 04, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 03, 03, 03, 03, 03, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 02, 02, 02, 02, 02, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 02, 01, 01, 01, 02, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 02, 01, 00, 01, 02, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 02, 01, 01, 01, 02, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 02, 02, 02, 02, 02, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 03, 03, 03, 03, 03, 03, 03, 04, 05, 06, 07,
 00, 07, 06, 05, 04, 04, 04, 04, 04, 04, 04, 04, 04, 05, 06, 07,
 00, 07, 06, 05, 05, 05, 05, 05, 05, 05, 05, 05, 05, 05, 06, 07,
 00, 07, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 07,
 00, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07
);

 VPAWNSCORE: array[0..7] of byte =
 (0,96,96,98,100,102,104,0);

 VPAWNCENTER: array[0..63] of byte = (
 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 2, 2, 0, 0, 0,
 0, 0, 2, 8, 8, 2, 0, 0,
 0, 0, 8,18,18, 8, 0, 0,
 0, 0, 4,17,17, 4, 0, 0,
 0, 0, 0,10,10, 0, 0, 0,
 0, 0, 0, 4, 4, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0
 );

 VPAWNBACKWARD: array[0..7] of byte =
 (0,23,20,12,6,1,0,0);

 VPROTECTEDPASSER: array[0..7] of byte =
 (0,2,2,4,12,35,90,0);

 VISOLATEDPAWN: array[0..7] of byte =
 (6,11,13,16,16,13,11,6);

 VPAWNSHIELD: array[0..63] of byte = (          // -f -g -h
  9,9,8,7,   9,8,6,6,   7,6,5,4,   5,5,4,2,     // 16 04 01
  9,8,8,7,   7,7,6,6,   6,5,4,4,   5,4,3,2,     // 32 08 02
  8,7,7,7,   7,6,5,5,   5,4,3,2,   5,4,3,2,     // 48 12 03
  7,6,6,6,   6,6,5,5,   5,4,2,2,   4,2,1,0);    // || --  -

 VPASSEDPAWN: array[0..63] of byte = (
   0,  0,  0,  0,  0,  0,  0,  0,
  22, 21, 18, 16, 16, 18, 21, 22,
  22, 21, 18, 16, 16, 18, 21, 22,
  36, 34, 31, 28, 28, 31, 34, 36,
  58, 56, 52, 48, 48, 52, 56, 58,
  82, 80, 76, 72, 72, 76, 80, 82,
 110,108,104,100,100,104,108,110,
   0, 0,  0,  0,  0,  0,   0,  0
 );

 VKINGPOS_: array[0..63] of byte = (
 00,05,10,16,16,10,05,00,
 05,17,22,27,27,22,17,05,
 10,22,34,39,39,34,22,10,
 15,34,47,49,49,47,34,15,
 17,36,49,52,52,49,36,17,
 14,30,43,48,48,43,30,14,
 09,22,29,34,34,29,22,09,
 03,08,12,18,18,12,08,03
);

 VKINGPOSX: array[0..8*15-1] of byte = (
 08,08,07,04,02,01,00,00,
 08,08,07,04,02,01,00,00,
 07,08,08,07,04,02,01,00,
 07,08,08,07,04,02,01,00,
 05,07,08,08,07,05,03,01,
 05,07,08,08,07,05,03,01,
 05,06,07,08,08,07,06,05,
 05,06,07,08,08,07,06,05,
 05,06,07,08,08,07,06,05,
 01,03,05,07,08,08,07,05,
 01,03,05,07,08,08,07,05,
 00,01,02,04,07,08,08,07,
 00,01,02,04,07,08,08,07,
 00,00,01,02,04,07,08,08,
 00,00,01,02,04,07,08,08
);

 VKINGPOSY: array[0..7] of byte = (
 01,03,05,07,07,06,05,03
);

 VKINGPPASSED: array[0..63] of byte = (
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
  12,14, 9, 4, 1, 0, 0, 0,
  19,22,14, 6, 2, 0, 0, 0,
  33,38,25,10, 4, 1, 0, 0,
  44,52,33,14, 8, 3, 1, 0,
   0, 0, 0, 0, 0, 0, 0, 0
 );

 VOKINGPPASSED: array[0..63] of byte = (
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   6, 5, 4, 3, 1, 0, 0, 0,
  11, 9, 7, 5, 2, 0, 0, 0,
  18,16,13, 7, 2, 1, 0, 0,
  23,20,14, 8, 4, 2, 1, 0,
   0, 0, 0, 0, 0, 0, 0, 0
 );

 VKINGDANGER: array[0..63] of byte = (
 04,00,20,42,42,20,00,04,
 06,06,23,46,46,23,06,06,
 13,27,39,52,52,39,27,13,
 22,34,48,60,60,48,34,22,
 22,34,48,60,60,48,34,22,
 13,27,39,52,52,39,27,13,
 06,06,23,46,46,23,06,06,
 04,00,20,42,42,20,00,04
);

 VCENTERBLOCK: array[0..7] of byte =
 (95,100,105,110,115,120,120,120);

 VKNIGHTOPPKING: array[0..7] of byte =
 (8,8,8,6,2,0,0,0);

 VKNIGHTOWNKING: array[0..7] of byte =
 (4,4,4,0,0,0,0,0);

 VBISHOPPOS: array[0..63] of byte = (
 03,01,00,01,01,00,01,03,
 02,10,11,11,11,11,10,02,
 06,11,13,14,14,13,11,06,
 08,11,14,17,17,14,11,08,
 08,11,14,17,17,14,11,08,
 06,11,13,14,14,13,11,06,
 02,10,11,11,11,11,10,02,
 00,01,02,03,03,02,01,00
);

 VBISHOPMOB: array[0..15] of byte =
 (080,085,090,094, 098,102,105,108, 111,114,116,118, 119,120,120,120);

 VROOKMOB: array[0..15] of byte =
 (087,091,094,096, 098,100,102,104, 106,107,108,109, 110,110,110,110);

 VBISHOPATTACK: array[0..7] of byte =
 (0,0,1,0,0,0,1,1);

 VBISHOPPIN: array[0..7] of byte =
 (0,0,9,0,0,0,4,7);

 VROOKPOS: array[0..63] of byte = (
 00,03,05,08,08,05,03,00,
 00,02,04,07,07,04,02,00,
 00,01,03,05,05,03,01,00,
 00,01,03,05,05,03,01,00,
 00,01,03,05,05,03,01,00,
 00,01,03,05,05,03,01,00,
 00,02,04,07,07,04,02,00,
 00,03,05,08,08,05,03,00
 );

 VROOKKING: array[0..7] of byte =
 (10,10,6,0,0,0,0,0);

 VQUEENPOS: array[0..63] of byte = (
 0, 2, 4, 5, 5, 4, 2, 0,
 1, 4, 9,10,10, 9, 4, 1,
 2, 6,10,12,12,10, 6, 2,
 4, 9,14,16,16,14, 9, 4,
 6,11,16,18,18,16,11, 6,
 6,11,16,18,18,16,11, 6,
 4, 8,11,13,13,11, 8, 4,
 2, 5, 7, 9, 9, 7, 5, 2
);

 VQUEENATTACK: array[0..5] of integer =
 (16,16,16,10,6,0);

 VKBNK: array[0..63] of byte = (
 56,48,40,32,24,16,08,00,
 48,40,32,24,16,08,00,08,
 40,32,24,16,08,00,08,16,
 32,24,16,08,00,08,16,24,
 24,16,08,00,08,16,24,32,
 16,08,00,08,16,24,32,40,
 08,00,08,16,24,32,40,48,
 00,08,16,24,32,40,48,56
 );

 VBORDER: array[0..63] of byte = (
 41,36,32,31,31,32,36,41,
 36,26,21,20,20,21,26,36,
 32,21,04,02,02,04,21,32,
 31,20,02,00,00,02,20,31,
 31,20,02,00,00,02,20,31,
 32,21,04,02,02,04,21,32,
 36,26,21,20,20,21,26,36,
 41,36,32,31,31,32,36,41
);

 PCHASH1 = 0;
 PCHASH2 = 1;
 PCVALUE = 2;
 PCPASSA = 3;
 PCPASSB = 4;
 PCSHIELD = 5;
 PCSTORM = 6;

function APawnShield(a,b,c:integer): integer; forward;
function SwapInt(n:integer): integer; forward;

// TPawnCache

constructor TPawnCache.Create;
begin
  inherited Create;
  getmem(ftable,131072*32 + 512);
  table := pointer((cardinal(ftable) + 255) and $ffffff00);
end;

destructor TPawnCache.Destroy;
begin
  freemem(ftable);
  inherited Destroy;
end;

procedure TPawnCache.Reset;
begin
  FillChar(table[0],131072*32,0);
end;

procedure TPawnCache.StorePosition(hp1,hp2:integer; const pdat:TPawnData);
var offs: integer;
begin
  offs := (hp1 and $1ffff) shl 3;
  table[offs+PCHASH1] := (hp1 and $ffff0000) + cardinal(pdat.CenterBlock);
  table[offs+PCHASH2] := hp2;
  table[offs+PCVALUE] := (pdat.Pwb shl 28) + (pdat.Pbw shl 24) +
                         ((pdat.vpstruct+2048) shl 12) + (pdat.vpos+2048);
  table[offs+PCPASSA] := (pdat.Acol shl 24) + pdat.APass;
  table[offs+PCPASSB] := (pdat.Bcol shl 24) + pdat.BPass;
  table[offs+PCSHIELD] := pdat.Shield;
  table[offs+PCSTORM] := pdat.Storm;
end;

function TPawnCache.GetPosition(hp1,hp2:integer; var pdat:TPawnData): boolean;
var offs,v: integer;
begin
  result := false;
  if hp1 = 0 then begin
     result := true;
     v := 0;
     pdat.vpstruct := v; pdat.vpos := v;  pdat.CenterBlock := v;
     pdat.APass := v; pdat.BPass := v; pdat.Acol := v; pdat.Bcol := v;
     pdat.Pwb := v; pdat.Pbw := v; pdat.Shield := v; pdat.Storm := v;
     exit;
  end;
  offs := (hp1 and $1ffff) shl 3;
  if ((table[offs+PCHASH1] xor hp1) and $ffff0000 <> 0) or (offs = 0) then exit;
  if table[offs+PCHASH2] <> hp2 then exit;
  pdat.CenterBlock := table[offs+PCHASH1] and 7;
  v := table[offs+PCVALUE];
  pdat.Pwb := (v shr 28) and $0f;
  pdat.Pbw := (v shr 24) and $0f;
  pdat.vpstruct := ((v shr 12) and $fff) - 2048;
  pdat.vpos := (v and $fff) - 2048;
  pdat.Acol := (table[offs+PCPASSA] shr 24) and $ff;
  pdat.Bcol := (table[offs+PCPASSB] shr 24) and $ff;
  pdat.APass := table[offs+PCPASSA] and $ffffff;
  pdat.BPass := table[offs+PCPASSB] and $ffffff;
  pdat.Shield := table[offs+PCSHIELD];
  pdat.Storm := table[offs+PCSTORM];
  result := true;
end;

procedure EvaluatePawns(var pdat:TPawnData; pboard:PByteArray; HP1,HP2:integer);
var
 i,i8,x,y,v,amat,bmat,nbp,at,bt,napawns,nbpawns,minx,maxx: integer;
 pintboard: pintarray;
 aprank,bprank: array[0..9] of integer;
const
 OPENFILE = $ff;
begin
  if pawncache.GetPosition(HP1,HP2,pdat) then exit;
  pintboard := pointer(pboard);
  for i := 0 to 9 do begin
      aprank[i] := OPENFILE;
      bprank[i] := -OPENFILE;
  end;
  pdat.Acol := 0;
  pdat.Bcol := 0;
  napawns := 0;
  nbpawns := 0;
  minx := 7;
  maxx := 0;
  for i := _A2 to _H7 do begin
      x := -1;
      if pboard[i] = APAWN then begin
         x := i and 7;
         y := i shr 4;
         pdat.Acol := pdat.Acol or (1 shl x);
         if y < aprank[x + 1] then
            aprank[x + 1] := y;
         inc(napawns);
      end;
      if pboard[i] = BPAWN then begin
         x := i and 7;
         y := i shr 4;
         pdat.Bcol := pdat.Bcol or (1 shl x);
         if y > bprank[x + 1] then
            bprank[x + 1] := y;
         inc(nbpawns);
      end;
      if x >= 0 then begin
         if x > maxx then maxx := x;
         if x < minx then minx := x;
      end;
  end;
  x := 0; nbp := x; amat := x; bmat := x;
  if pdat.ACol = $ff then dec(amat,4);
  if pdat.BCol = $ff then dec(bmat,4);
  pdat.Pwb := x;   pdat.Pbw := x;
  pdat.vpos := x; pdat.CenterBlock := x;
  pdat.APass := x; pdat.BPass := x;
  for i := _A2 to _H7 do begin
      if pboard[i] and PAWNMASK = 0 then continue;
      x := (i and 7) + 1;
      y := i shr 4;
      i8 := TO8[i];
      if pboard[i] = APAWN then begin
         if pboard[i+16] = BPAWN then begin
            if COLORSQUARE[i and 31] = 1 then inc(pdat.Pwb)  // blocked pawns
               else inc(pdat.Pbw);
            if (x = 4) or (x = 5) then inc(pdat.CenterBlock);
         end;
         inc(amat,VPAWNSCORE[y]);
         if (aprank[x+1] = y) or (aprank[x-1] = y) then inc(amat,2); // duo
         if aprank[x] <> y then dec(amat,6);                         // doubled
         if (x = 4) or (x = 5) then inc(pdat.vpos,VPAWNCENTER[i8]);    // center
         if (bprank[x] <= y) and (bprank[x+1] <= y) and (bprank[x-1] <= y) then
         begin
            if i >= _A4 then begin                       // passed
               pdat.APass := (pdat.APass shl 6) + i8;
            end;
            v := VPASSEDPAWN[i8];
            if (abs(aprank[x+1] - y) < 2) or (abs(aprank[x-1] - y) < 2)
               then inc(v,VPROTECTEDPASSER[y]);            // protected
            inc(amat,v);
            dec(pdat.vpos,v shr 3);
            if (aprank[x+1] = OPENFILE) and (aprank[x-1] = OPENFILE)
               then dec(amat,VISOLATEDPAWN[x-1]);           // isolated
            if aprank[x] <> y then dec(amat,30);            // doubled
            continue;
         end;
         if (aprank[x+1] = OPENFILE) and (aprank[x-1] = OPENFILE) then begin
            dec(amat,VISOLATEDPAWN[x-1]);               // isolated pawn
            if bprank[x] = -255 then dec(amat,8);          // open file
            if aprank[x] <> y then dec(amat,8);            // doubled
            continue;
         end;
         if bprank[x] <> -OPENFILE then continue;      // if closed continue
         v := 0;
         if bprank[x+1] = -OPENFILE then inc(v);
         if bprank[x-1] = -OPENFILE then inc(v);
         if (v = 1) and ((aprank[x+1] = y) or (aprank[x-1] = y)) then begin
            v := (VPASSEDPAWN[i8] - 10) shr 1;
            inc(amat,v);                            // candidate
            dec(pdat.vpos,v shr 2);
         end;
         if (aprank[x+1] > y) and (aprank[x-1] > y) then begin
            if (pboard[i+31] = BPAWN) or (pboard[i+33] = BPAWN)
               then dec(amat,VPAWNBACKWARD[y]);             // backward
         end;
      end
      else begin
         inc(bmat,VPAWNSCORE[7-y]);
         if bprank[x] <> y then dec(bmat,6);                          // doubled
         if (bprank[x+1] = y) or (bprank[x-1] = y) then inc(bmat,2);  // duo
         if (x = 4) or (x = 5) then dec(pdat.vpos,VPAWNCENTER[63-i8]);  // center
         if (aprank[x] >= y) and (aprank[x+1] >= y) and (aprank[x-1] >= y) then
         begin
            if (i <= _H5) and (nbp <= 3) then begin   // passed
               inc(nbp);
               pdat.BPass := (pdat.BPass shl 6) + i8;
            end;
            v := VPASSEDPAWN[63-i8];
            if (abs(bprank[x+1] - y) < 2) or (abs(bprank[x-1] - y) < 2)
                then inc(v,VPROTECTEDPASSER[7-y]);          // protected
            inc(bmat,v);
            inc(pdat.vpos,v shr 3);
            if (bprank[x+1] = -OPENFILE) and (bprank[x-1] = -OPENFILE)
               then dec(bmat,VISOLATEDPAWN[x-1]);              // isolated
            if bprank[x] <> y then dec(bmat,30);               // doubled
            continue;
         end;
         if (bprank[x+1] = -OPENFILE) and (bprank[x-1] = -OPENFILE) then begin
            dec(bmat,VISOLATEDPAWN[x-1]);          // isolated pawns
            if aprank[x] = OPENFILE then dec(bmat,8);      // open file
            if bprank[x] <> y then dec(bmat,8);       // doubled
            continue;
         end;
         if aprank[x] <> OPENFILE then continue;   // if closed continue
         v := 0;
         if aprank[x+1] = OPENFILE then inc(v);
         if aprank[x-1] = OPENFILE then inc(v);
         if (v = 1) and ((bprank[x+1] = y) or (bprank[x-1] = y)) then begin
            v := (VPASSEDPAWN[63-i8] - 10) shr 1;
            inc(bmat,v);                              // candidate
            inc(pdat.vpos,v shr 2);
         end;
         if (bprank[x+1] < y) and (bprank[x-1] < y) then begin
            if (pboard[i-31] = APAWN) or (pboard[i-33] = APAWN)
               then dec(bmat,VPAWNBACKWARD[7-y]);            // backward
         end;
      end;
  end;
  if (napawns > nbpawns+1) then dec(amat,(napawns-nbpawns-1) * 8);
  if (nbpawns > napawns+1) then dec(bmat,(nbpawns-napawns-1) * 8);
  if pdat.CenterBlock = 2 then pdat.CenterBlock := 3
     else if (pdat.ACol and pdat.BCol and $18) = $18
          then inc(pdat.CenterBlock);
  pdat.APass := pdat.APass and $ffffff;
  pdat.BPass := pdat.BPass and $ffffff;
  pdat.Shield := APawnShield(pintboard[5],pintboard[9],pintboard[13]) +
                 APawnShield(SwapInt(pintboard[4]),SwapInt(pintboard[8]),
                             SwapInt(pintboard[12])) shl 8 +
                 APawnShield(pintboard[25] shl 1,pintboard[21] shl 1,
                             pintboard[17] shl 1) shl 16 +
                 APawnShield(SwapInt(pintboard[24]) shl 1,
                             SwapInt(pintboard[20]) shl 1,
                             SwapInt(pintboard[16]) shl 1) shl 24;
  pdat.vpstruct := amat-bmat;
  at := 5; bt:= 5;
  if pdat.ACol and _COLA = 0 then inc(bt,2);
  if pdat.ACol and _COLB = 0 then inc(bt,2);
  if pdat.ACol and _COLC = 0 then inc(bt);
  dec(bt,((integer(pboard[_A2] or pboard[_A3]) and P_APAWNS) +
          (integer(pboard[_B2] or pboard[_B3]) and P_APAWNS)) shr 5);
  dec(bt,(integer(pboard[_C2] or pboard[_C3]) and P_APAWNS) shr 6);
  if pdat.BCol and _COLH = 0 then inc(at,2);
  if pdat.BCol and _COLG = 0 then inc(at,2);
  if pdat.BCol and _COLF = 0 then inc(at);
  dec(at,((integer(pboard[_H7] or pboard[_H6]) and P_BPAWNS) +
          (integer(pboard[_G7] or pboard[_G6]) and P_BPAWNS)) shr 4);
  dec(at,(integer(pboard[_F7] or pboard[_F6]) and P_BPAWNS) shr 5);
  v := (at + (bt shl 4)) shl 8;
  at := 5; bt:= 5;
  if pdat.ACol and _COLH = 0 then inc(bt,2);
  if pdat.ACol and _COLG = 0 then inc(bt,2);
  if pdat.ACol and _COLF = 0 then inc(bt);
  dec(bt,((integer(pboard[_H2] or pboard[_H3]) and P_APAWNS) +
          (integer(pboard[_G2] or pboard[_G3]) and P_APAWNS)) shr 5);
  dec(bt,(integer(pboard[_F2] or pboard[_F3]) and P_APAWNS) shr 6);
  if pdat.BCol and _COLA = 0 then inc(at,2);
  if pdat.BCol and _COLB = 0 then inc(at,2);
  if pdat.BCol and _COLC = 0 then inc(at);
  dec(at,((integer(pboard[_A7] or pboard[_A6]) and P_BPAWNS) +
          (integer(pboard[_B7] or pboard[_B6]) and P_BPAWNS)) shr 4);
  dec(at,(integer(pboard[_C7] or pboard[_C6]) and P_BPAWNS) shr 5);
  pdat.Storm := v + at + (bt shl 4) + ((minx + maxx) shl 16);
  pawncache.StorePosition(HP1,HP2,pdat);
end;

function APawnShield(a,b,c:integer): integer;
var
 x: integer;
 t,r: byte;
begin
  x := $4000;
  t := 0;
  if a and x <> 0 then t := 3*16
     else if b and x <> 0 then t := 2*16
        else if c and x <> 0 then t := 1*16;
  r := t;
  x := x shl 8;
  t := 0;
  if a and x <> 0 then t := 3*4
     else if b and x <> 0 then t := 2*4
        else if c and x <> 0 then t := 1*4;
  inc(r,t);
  x := x shl 8;
  t := 0;
  if a and x <> 0 then t := 3
     else if b and x <> 0 then t := 2
        else if c and x <> 0 then t := 1;
  inc(r,t);
  result := VPAWNSHIELD[r];
end;

function SwapInt(n:integer): integer;
asm
  bswap eax
end;


// TEvalCache


procedure TEvalCache.Reset;
begin
  dummy := 0;
  FillChar(data[0],262144*4,0);
end;

procedure TEvalCache.Store(key1,key2,value: integer);
begin
  if (value < -511) or (value > 511) then exit;
  data[key1 and $3ffff] := (key1 and integer($fffc0000)) +
                           (key2 and $0003fc00) + value + 512;
end;

function TEvalCache.Get(key1,key2:integer): integer;
var vdata: integer;
begin
  result := $ffff;
  vdata := data[key1 and $3ffff];
  if (vdata = 0) or (((vdata xor key1) and $fffc0000) <> 0) then exit;
  if ((vdata xor key2) and $0003fc00) <> 0 then exit;
  result := (vdata and $3ff) - 512;
end;


// Eval functions


function EvaluateRel(pcbe:pointer; pbeta:integer; btm:boolean;
                     var islazy:boolean): integer;
var
 pboard: PByteArray;
 pcb: PChessBrd;
 i,v,w,at,bt,kax,kbx,rmult: integer;
 napawn,nbpawn,nabis,nbbis,npieces,nheavy,mater,drawfactor,hasheval: integer;
 edat: TEvalData;
 pdat: TPawnData;
 wrongside: boolean;
const
 KWING = $e0;
 QWING = $07;
 NONE = $ff;
 AHPAWNS = $81;
 NOPAWNS = $fff0;
 VPAWNSTAGE: array[0..15] of byte =
    (0,0,10,18, 24,26,27,28, 29,30,30,30, 30,30,30,30);   // 0 1024 2048 3072
 VDEVELOPMENT: array[0..9] of integer =
    (-36,-24,-14,-6, 0, 6,14,24,36, 0);
 VKINGATTACK: array[0..31] of byte =
    (0,0, 1,1, 2,3, 5,7, 9,12, 15,19, 23,28, 34,41, 49,58,
     64,75, 85,95, 105,115, 125,135, 140,145, 150,150, 150,150);
label
 evexit;
begin
  pcb := pcbe;
  islazy := false;
  if btm then rmult := -1 else rmult := 1;
  napawn := pcb.AMatStage and 15;
  nbpawn := pcb.BMatStage and 15;
  mater := (pcb.AMatStage + pcb.BMatStage) shr 1;

  if mater = VROOK then begin
    (*
    if (pcb.AMatStage = pcb.BMatStage + 1) or (pcb.AMatStage = pcb.BMatStage - 1) then begin
       result := 0;//EvaluateKRPKR16(pboard,true);
       if pcb.BMatStage > pcb.AMatStage then result := -result;
       exit;
    end;    *)
    if (pcb.AMatStage = VROOK) and (pcb.BMatStage = VROOK) then begin
       result := ((VBORDER[TO8[pcb.BKingPos]] -
                   VBORDER[TO8[pcb.AKingPos]]) div 4)*rmult;
       exit;
    end;
  end;
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  if (mater >= VROOK) and (napawn > 1) and (nbpawn > 1) then begin
     result := ((napawn-nbpawn)*100 + pcb.AMatStage - pcb.BMatStage)*rmult;
     if result >= pbeta+400 then begin
        result := result-400;
        islazy := true;
        exit;
     end;
     EvaluatePawns(pdat,pboard,pcb.HashP1,pcb.HashP2);
     result := (pdat.vpstruct + pdat.vpos + pcb.AMatStage - pcb.BMatStage)*rmult;
     if result >= pbeta+200 then begin
        result := result-200;
        islazy := true;
        exit;
     end;
  end
  else EvaluatePawns(pdat,pboard,pcb.HashP1,pcb.HashP2);

  hasheval := 0;
  if mater > 1100 then
  with pcb^ do begin
     hasheval := HashB2;
     result := evalcache.Get(HashB1,hasheval);
     if result <> $ffff then begin
        result := result*rmult;
        exit;
     end;
  end;

  w := 0;
  edat.VPos := pdat.vpos;
  edat.AMat := w;
  edat.BMat := w;
  edat.AClose := w;
  edat.BClose := w;
  v := pcb.AMatStage - pcb.BMatStage;
  if (v > 250)
     then inc(edat.AMat,36 + (napawn shl 3) - (pcb.BMatStage shr 6));
  if (v < -250)
     then inc(edat.BMat,36 + (nbpawn shl 3) - (pcb.AMatStage shr 6));
  nabis := (pcb.AMatCnt shr 12) and 15;
  nbbis := (pcb.BMatCnt shr 12) and 15;
  inc(edat.AMat,(nabis shr 1) * 40);
  inc(edat.BMat,(nbbis shr 1) * 40);
  npieces := (pcb.AMatCnt shr 24) + (pcb.BMatCnt shr 24);
  w := NONE;
  kax := pcb.AKingPos and 7;
  kbx := pcb.BKingPos and 7;
  edat.AKingPos := pcb.AKingPos;
  edat.BKingPos := pcb.BKingPos;
  edat.AMatStage := pcb.AMatStage;
  edat.BMatStage := pcb.BMatStage;
  edat.pBoard := pboard;
  edat.Mater := mater;
  edat.ARookPos1 := w;
  edat.ARookPos2 := w;
  edat.AQueenPos := w;
  edat.ABishopPos := w;
  edat.BRookPos1 := w;
  edat.BRookPos2 := w;
  edat.BQueenPos := w;
  edat.BBishopPos := w;

  i := 0;
  while true do begin
    w := pcb.ALoc[i];
    inc(i);
    if w = 129 then break;
    v := pboard[w];
    if (v and 4) = 0 then continue;
    edat.Pos := w;
    AEvalFunction[v and 7](edat,pdat);
  end;
  i := 0;
  while true do begin
    w := pcb.BLoc[i];
    inc(i);
    if w = 129 then break;
    v := pboard[w];
    if (v and 4) = 0 then continue;
    edat.Pos := w;
    BEvalFunction[v and 7](edat,pdat);
  end;

  if pdat.APass + pdat.BPass <> 0
     then EvaluatePassedPawns(edat,pdat);
  if edat.AQueenPos + edat.BQueenPos < NONE*2
     then EvaluateRookQueen(edat,pdat);

  if (edat.AMatStage > edat.BMatStage - 100) and (napawn > nbpawn)
     then dec(edat.AMat,VPAWNSTAGE[(mater shr 8) and 15]);
  if (edat.BMatStage > edat.AMatStage - 100) and (nbpawn > napawn)
     then dec(edat.BMat,VPAWNSTAGE[(mater shr 8) and 15]);

  nheavy := (((pcb.AMatCnt shr 16) + (pcb.BMatCnt shr 16)) and 15) +
            (((pcb.AMatCnt shr 20) + (pcb.BMatCnt shr 20)) and 15)*2;
  if nheavy >= 4 then inc(mater,350);
  drawfactor := 0;

  if mater < 1600 then begin
     w := pcb.AMatStage + pcb.BMatStage;
     if w = 0 then begin
        result := 0;
        exit;
     end;
     if w = 1 then begin
        result := DBEndgameX(pcb,btm,wrongside)*rmult;
        exit;
     end;
     if (napawn+nbpawn) = 0 then begin
        EvaluateNoPawns(edat,pdat);
     end
     else begin
        w := ((pdat.Storm shr 16) and 15) * 8;
        edat.VPos := VKINGPOSY[pcb.AKingPos shr 4] * VKINGPOSX[w+kax] -
                     VKINGPOSY[7-(pcb.BKingPos shr 4)] * VKINGPOSX[w+kbx];
        if (npieces <= 2) and (nabis+nbbis >= 1)
           then drawfactor := EvaluateBishopEndgame(edat,pdat);
     end;
     if mater > 900 then edat.VPos := edat.VPos * (1600-mater) div 700;
     if mater < 50 then begin
        if pdat.APass + pdat.BPass <> 0 then
           inc(edat.VPos,PawnRace(edat,btm,pdat.APass,pdat.BPass));
        inc(edat.AMat,napawn * 16);
        inc(edat.BMat,nbpawn * 16);
     end;
     inc(edat.VPos,(nabis-nbbis)*12);
  end
  else begin
     EvaluatePawnMiddlegame(edat,pdat);
     if (kax shr 2 <> kbx shr 2) and
        (edat.ARookPos1 <> NONE) and (edat.BRookPos1 <> NONE)
     then begin
        v := pdat.Storm;
        if kax >= 4 then v := v shr 8;
        at := v and $0f;
        bt := (v shr 4) and $0f;
        inc(edat.VPos,(bt-at)*3);
        inc(edat.AClose,at);
        inc(edat.BClose,bt);
     end;
     bt := VKINGDANGER[TO8[pcb.BKingPos]];
     dec(bt,bt*pdat.CenterBlock shr 4);
     at := VKINGDANGER[TO8[pcb.AKingPos]];
     dec(at,at*pdat.CenterBlock shr 4);
     w := bt - at;
     if (edat.AQueenPos = NONE) and (edat.BQueenPos = NONE) then w := w div 2;
     at := ADevelop(pboard);
     bt := BDevelop(pboard);
     inc(edat.VPos,w + VDEVELOPMENT[at-bt+4]);
     if (pboard[_D1] = AQUEEN) then inc(edat.VPos,(4-at)*2);
     if (pboard[_D8] = BQUEEN) then dec(edat.VPos,(4-bt)*2);
     if ((pboard[_G1] = AROOK) or (pboard[_H1] = AROOK)) and
        ((pcb.AKingPos = _G1) or (pcb.AKingPos = _F1)) and
        ((pdat.ACol and _COLH) <> 0) then dec(edat.VPos,24);
     if ((pboard[_G8] = BROOK) or (pboard[_H8] = BROOK)) and
        ((pcb.BKingPos = _G8) or (pcb.BKingPos = _F8)) and
        ((pdat.BCol and _COLH) <> 0) then inc(edat.VPos,24);
     if mater < 2200 then edat.VPos := edat.VPos * (mater-1600) div 600;
  end;

  if edat.BClose > 1 then begin
     if edat.BClose > 63 then edat.BClose := 63;
     inc(edat.AMat,VKINGATTACK[edat.BClose shr 1]);
  end;
  if edat.AClose > 1 then begin
     if edat.AClose > 63 then edat.AClose := 63;
     inc(edat.BMat,VKINGATTACK[edat.AClose shr 1]);
  end;
  result := edat.AMat - edat.BMat + edat.VPos + pdat.vpstruct;
  if drawfactor <> 0 then result := result * drawfactor div 8;
  if result > 7000 then result := 7000;
  if result <-7000 then result := -7000;
  if hasheval <> 0 then evalcache.Store(pcb.HashB1,hasheval,result);
  if npieces > 3 then goto evexit;
  if (pcb.AMatStage xor pcb.BMatStage) and NOPAWNS = 0 then goto evexit;
  if pcb.AMatStage > pcb.BMatStage then begin
     if (napawn > 0) or (nbpawn > 3) then goto evexit;
     w := (pcb.BMatStage shl 16 + pcb.AMatStage) and VNOPAWNS;
  end
  else begin
     if (nbpawn > 0) or (napawn > 3) then goto evexit;
     w := (pcb.AMatStage shl 16 + pcb.BMatStage) and VNOPAWNS;
  end;
  v := 0;
  for i := 0 to EGCOUNT-1 do if EGSCORES[i].pp = w then
  begin
     case (napawn+nbpawn) of
      0: begin
            v := EGSCORES[i].p0 * 10;
            if (w = VKNIGHT) or (w = VBISHOP)
               then v := abs(result) - 1;
         end;
      1: v := EGSCORES[i].p1 * 10;
      2: v := EGSCORES[i].p2 * 10;
      3: v := EGSCORES[i].p3 * 10;
     end;
     break;
  end;
  if pcb.AMatStage > pcb.BMatStage then dec(result,v) else inc(result,v);
  if hasheval <> 0 then evalcache.Store(pcb.HashB1,hasheval,result);
 evexit:
  result := result * rmult;
end;

procedure EvaluateNoPawns(var edat:TEvalData; var pdat:TPawnData);
var w,v: integer;
begin
  edat.VPos := VBORDER[TO8[edat.BKingPos]] - VBORDER[TO8[edat.AKingPos]];
  w := DIST16[edat.AKingPos-edat.BKingPos+120];
  v := edat.AMatStage - edat.BMatStage;
  if v > 0 then dec(edat.VPos,w*2);
  if v < 0 then inc(edat.VPos,w*2);
  if (edat.AMatStage = VKNIGHT+VBISHOP) and (edat.BMatStage = 0) then begin
     w := TO8[edat.BKingPos];
     if (COLORSQUARE[edat.ABishopPos and 31] = 1) then w := VFlipSquare8(w);
     edat.AMat := 540 + VKBNK[w]*3;
     edat.BMat := 0;
  end;
  if (edat.BMatStage = VKNIGHT+VBISHOP) and (edat.AMatStage = 0) then begin
     w := TO8[edat.AKingPos];
     if (COLORSQUARE[edat.BBishopPos and 31] = 1) then w := VFlipSquare8(w);
     edat.BMat := 540 + VKBNK[w]*3;
     edat.AMat := 0;
  end;
end;

function EvaluateBishopEndgame(var edat:TEvalData; var pdat:TPawnData): integer;
var
 AMatStage,BMatStage,x,ca,cb,ap,bp: integer;
const
 KWING = $e0;
 QWING = $07;
 NOPAWNS = $fff0;
 NONE = $ff;
 AHPAWNS = $81;
begin
  result := 0;
  AMatStage := edat.AMatStage;
  BMatStage := edat.BMatStage;
  ap := AMatStage and NOPAWNS;
  bp := BMatStage and NOPAWNS;
  if (ap = VBISHOP) and (bp = VKNIGHT) and
     (pdat.ACol and QWING <> 0) and (pdat.ACol and KWING <> 0)
     then inc(edat.AMat,20);
  if (bp = VBISHOP) and (ap = VKNIGHT) and
     (pdat.BCol and QWING <> 0) and (pdat.BCol and KWING <> 0)
     then inc(edat.BMat,20);
  ca := COLORSQUARE[edat.ABishopPos and 31];
  cb := COLORSQUARE[edat.BBishopPos and 31];
  if (ap = VBISHOP) and (bp = VBISHOP) and (ca <> cb)
     then result := 6;
  if (AMatStage = (VBISHOP + 1)) and (AMatStage > BMatStage) and
     (pdat.ACol and AHPAWNS <> 0) and (edat.BKingPos >= _A7) then begin
     x := edat.BKingPos and 7;
     if ((x <= 1) and (pdat.ACol and QWING <> 0) and (ca = 0)) or
        ((x >= 6) and (pdat.ACol and KWING <> 0) and (ca = 1))
        then result := 1;
  end;
  if (BMatStage = (VBISHOP + 1)) and (BMatStage > AMatStage) and
     (pdat.BCol and AHPAWNS <> 0) and (edat.AKingPos <= _H2) then begin
     x := edat.AKingPos and 7;
     if ((x <= 1) and (pdat.BCol and QWING <> 0) and (cb = 1)) or
        ((x >= 6) and (pdat.BCol and KWING <> 0) and (cb = 0))
        then result := 1;
  end;
end;

procedure EvaluatePassedPawns(var edat:TEvalData; var pdat:TPawnData);
var
 t,x,w,w8,v: integer;
 pboard: PByteArray;
const
 NOPAWNS = $fff0;
begin
  pboard := edat.pBoard;
  t := pdat.APass;
  while t <> 0 do begin
     w8 := t and 63;
     w := TO16[w8];
     t := t shr 6;
     if pboard[w+16] <> 0 then begin
        v := VPASSEDPAWN[w8] shr 3; // 1..13
        dec(edat.AMat,v);
        if (pboard[w+16] and B_PIECES) <> 0 then dec(edat.AMat,v);
        if pboard[w+16] = BKNIGHT then dec(edat.AMat,v);
     end;
     x := w and 7;
     if (x = (edat.ARookPos1 and 15)) and (edat.ARookPos1 < w) then inc(edat.AMat,18);
     if (x = (edat.ARookPos2 and 15)) and (edat.ARookPos2 < w) then inc(edat.AMat,18);
     if ((edat.BMatStage and NOPAWNS) = VKNIGHT) and ((x = 0) or (x = 7))
        then inc(edat.AMat,20);
     x := DIST16[w + (16 + 120) - edat.AKingPos];
     inc(edat.AMat,VKINGPPASSED[(w8 and $f8) + x]);
     x := DIST16[w + (16 + 120) - edat.BKingPos];
     dec(edat.AMat,VOKINGPPASSED[(w8 and $f8) + x]);
   end;
  t := pdat.BPass;
  while t <> 0 do begin
     w8 := t and 63;
     w := TO16[w8];
     t := t shr 6;
     if pboard[w-16] <> 0 then begin
        v := VPASSEDPAWN[63-w8] shr 3; // 1..13
        dec(edat.BMat,v);
        if (pboard[w-16] and A_PIECES) <> 0 then dec(edat.BMat,v);
        if pboard[w-16] = AKNIGHT then dec(edat.BMat,v);
     end;
     x := w and 7;
     if (x = (edat.BRookPos1 and 15)) and (edat.BRookPos1 > w) then inc(edat.BMat,18);
     if (x = (edat.BRookPos2 and 15)) and (edat.BRookPos2 > w) then inc(edat.BMat,18);
     if ((edat.AMatStage and NOPAWNS) = VKNIGHT) and ((x = 0) or (x = 7))
        then inc(edat.BMat,20);
     x := DIST16[w + (-16 + 120) - edat.BKingPos];
     inc(edat.BMat,VKINGPPASSED[((63-w8) and $f8) + x]);
     x := DIST16[w + (-16 + 120) - edat.AKingPos];
     dec(edat.BMat,VOKINGPPASSED[((63-w8) and $f8) + x]);
  end;
end;

function PawnRace(var edat:TEvalData; blacktomove:boolean;
                  apawns,bpawns:integer): integer;
var kd,kod,dy,pos,pp,akd,bkd,bonus: integer;
begin
  result := 0;
  akd := edat.AKingPos + 120;
  bkd := edat.BKingPos + 120;
  if blacktomove then bonus := 1 else bonus := 0;
  while apawns <> 0 do begin
     pos := TO16[apawns and 63];
     apawns := apawns shr 6;
     dy := 7 - (pos shr 4);
     pp := _A8 + (pos and 7);
     kd := DIST16[bkd - pp] - bonus;
     kod := DIST16[akd - pp];
     if (dy < kd) or ((kod = 1) and (dy <= 2)) then begin;
        inc(result,600);
        break;
     end;
  end;
  if not blacktomove then bonus := 1 else bonus := 0;
  while bpawns <> 0 do begin
     pos := TO16[bpawns and 63];
     bpawns := bpawns shr 6;
     dy := (pos shr 4);
     pp := _A1 + (pos and 7);
     kd := DIST16[akd - pp] - bonus;
     kod := DIST16[bkd - pp];
     if (dy < kd) or ((kod = 1) and (dy <= 2)) then begin;
        dec(result,600);
        break;
     end;
  end;
end;

procedure EvaluateAKnight(var edat:TEvalData; var pdat:TPawnData);
var t,w: integer;
begin
  w := edat.Pos;
  inc(edat.AMat,(VKNIGHT-128) + VKNIGHTPOS[TO8[w]] + VCENTERBLOCK[pdat.CenterBlock]);
  t := AKnightVAttack(pointer(integer(edat.pBoard)+w));
  inc(edat.AMat,t and 31);
  inc(edat.BClose,VKNIGHTOPPKING[DIST16[edat.BKingPos-w+120]]);
  dec(edat.AClose,VKNIGHTOWNKING[DIST16[edat.AKingPos-w+120]]);
end;

function AKnightVAttack(mempos:pointer): integer;
const
 VAKNIGHTATK: array[0..15] of byte =
  (0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 32+4, 2, 2, 4, 32+4, 32+4);
asm
  push esi
  mov esi, offset VAKNIGHTATK
  movzx ecx, byte ptr [eax+14]
  movzx edx, byte ptr [eax+18]
  push edi
  mov edi, eax
  xor eax, eax
  and ecx, 15
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  movzx ecx, byte ptr [edi-14]
  movzx edx, byte ptr [edi-18]
  and ecx, 15
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  movzx ecx, byte ptr [edi+31]
  movzx edx, byte ptr [edi+33]
  and ecx, 15
  pop edi
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  pop esi
end;

procedure EvaluateBKnight(var edat:TEvalData; var pdat:TPawnData);
var t,w: integer;
begin
  w := edat.Pos;
  inc(edat.BMat,(VKNIGHT-128) + VKNIGHTPOS[63-TO8[w]] + VCENTERBLOCK[pdat.CenterBlock]);
  t := BKnightVAttack(pointer(integer(edat.pBoard)+w));
  inc(edat.BMat,t and 31);
  inc(edat.AClose,VKNIGHTOPPKING[DIST16[edat.AKingPos-w+120]]);
  dec(edat.BClose,VKNIGHTOWNKING[DIST16[edat.BKingPos-w+120]]);
end;

function BKnightVAttack(mempos:pointer): integer;
const
 VBKNIGHTATK: array[0..15] of byte =
  (0, 0, 32+4, 2, 2, 4, 32+4, 32+4,   0, 0, 0, 0, 0, 0, 0, 0);
asm
  push esi
  mov esi, offset VBKNIGHTATK
  movzx ecx, byte ptr [eax-14]
  movzx edx, byte ptr [eax-18]
  push edi
  mov edi, eax
  xor eax, eax
  and ecx, 15
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  movzx ecx, byte ptr [edi+14]
  movzx edx, byte ptr [edi+18]
  and ecx, 15
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  movzx ecx, byte ptr [edi-31]
  movzx edx, byte ptr [edi-33]
  and ecx, 15
  pop edi
  and edx, 15
  add al, byte ptr [esi + ecx]
  add al, byte ptr [esi + edx]
  pop esi
end;

procedure EvaluateABishop(var edat:TEvalData; var pdat:TPawnData);
var
 w,ww: integer;
 pboard: pbytearray;
 bisattack: boolean;
begin
  w := edat.Pos;
  edat.ABishopPos := w;
  pboard := edat.pBoard;
  inc(edat.AMat,(VBISHOP-16) + VBISHOPPOS[TO8[w]]);
  AAttackBishop(pBoard,w,edat.AMat,ww,bisattack);
  if COLORSQUARE[w and 31] = 1
     then dec(edat.AMat,pdat.Pwb*4)
     else dec(edat.AMat,pdat.Pbw*4);
  if ((w = _A7) and (pboard[_B6] = BPAWN)) or
     ((w = _H7) and (pboard[_G6] = BPAWN)) then dec(edat.Amat,100);
  if edat.Mater >= 900 then begin
     if (DIST16[edat.BKingPos-w+120] < 4) or
        (bisattack and (abs(abs((w and 7)-(edat.BKingPos and 7)) -
                        abs((w shr 4)-(edat.BKingPos shr 4))) < 3))
                        then inc(edat.BClose,8);
     if DIST16[edat.AKingPos-w+120] < 2 then dec(edat.AClose,4);
  end;
end;

procedure AAttackBishop(pboard:pbytearray; pos:integer;
                          var amat,atkd:integer; var atkking: boolean);
var
 p,v,mob,i,st:integer;
begin
  atkking := false;
  mob := 0;
  for i := 0 to 1 do begin
    st := 15 + i*2;
    p := pos;
    repeat
      inc(p,st);
      v := pboard[p];
      if v <> 0 then break;
      inc(mob);
    until false;
    if p >= _A6-3 then atkking := true;
    if v = BPAWN then begin
       if (pboard[p+15] <> BPAWN) and (pboard[p+17] <> BPAWN)
          then inc(amat,3);
       continue;
    end;
    if (v and B_PIECES) = 0 then begin
       if v = AKNIGHT then inc(mob);
       continue;
    end;
    inc(mob);
    if v = BBISHOP then continue;
    if (v = BKNIGHT) and (pboard[p+15] <> BPAWN) and (pboard[p+17] <> BPAWN)
       then inc(amat,6);
    dec(atkd,VBISHOPATTACK[v and 7]);
    repeat
       inc(p,st);
       v := pboard[p];
    until v <> 0;
    if (v and B_PIECES) = 0 then continue;
    inc(amat,VBISHOPPIN[v and 7]);
  end;
  v := 0;
  p := pos-15;
  repeat
    if pboard[p] <> 0 then break;
    dec(p,15);
    inc(v);
  until false;
  p := pos-17;
  repeat
    if pboard[p] <> 0 then break;
    dec(p,17);
    inc(v);
  until false;
  inc(amat,VBISHOPMOB[mob+v]-100);
end;

procedure EvaluateBBishop(var edat:TEvalData; var pdat:TPawnData);
var
 w,ww: integer;
 pboard: pbytearray;
 bisattack: boolean;
begin
  w := edat.Pos;
  edat.BBishopPos := w;
  pboard := edat.pBoard;
  inc(edat.BMat,(VBISHOP-16) + VBISHOPPOS[63-TO8[w]]);
  BAttackBishop(pBoard,w,edat.BMat,ww,bisattack);
  if COLORSQUARE[w and 31] = 0
     then dec(edat.BMat,pdat.Pwb*4)
     else dec(edat.BMat,pdat.Pbw*4);
  if ((w = _A2) and (pboard[_B3] = APAWN)) or
     ((w = _H2) and (pboard[_G3] = APAWN)) then dec(edat.Bmat,100);
  if edat.Mater >= 900 then begin
     if (DIST16[edat.AKingPos-w+120] < 4) or
        (bisattack and (abs(abs((w and 7)-(edat.AKingPos and 7)) -
                        abs((w shr 4)-(edat.AKingPos shr 4))) < 3))
                        then inc(edat.AClose,8);
     if DIST16[edat.BKingPos-w+120] < 2 then dec(edat.BClose,4);
  end;
end;

procedure BAttackBishop(pboard:pbytearray; pos:integer;
                          var bmat,atkd:integer; var atkking: boolean);
var p,v,mob,i,st: integer;
begin
  atkking := false;
  mob := 0;
  for i := 0 to 1 do begin
    st := 15 + i*2;
    p := pos;
    repeat
      dec(p,st);
      v := pboard[p];
      if v <> 0 then break;
      inc(mob);
    until false;
    if p <= _H3+3 then atkking := true;
    if v = APAWN then begin
       if (pboard[p-15] <> APAWN) and (pboard[p-17] <> APAWN)
          then inc(bmat,3);
       continue;
    end;
    if (v and A_PIECES) = 0 then begin
       if v = BKNIGHT then inc(mob);
       continue;
    end;
    inc(mob);
    if v = ABISHOP then continue;
    if (v = AKNIGHT) and (pboard[p-15] <> APAWN) and (pboard[p-17] <> APAWN)
       then inc(bmat,6);
    inc(atkd,VBISHOPATTACK[v and 7]);
    repeat
       dec(p,st);
       v := pboard[p];
    until v <> 0;
    if (v and A_PIECES) = 0 then continue;
    inc(bmat,VBISHOPPIN[v and 7]);
  end;
  v := 0;               
  p := pos+15;
  repeat
    if pboard[p] <> 0 then break;
    inc(p,15);
    inc(v);
  until false;
  p := pos+17;
  repeat
    if pboard[p] <> 0 then break;
    inc(p,17);
    inc(v);
  until false;
  inc(bmat,VBISHOPMOB[mob+v]-100);
end;

function AttackRook(pboard:pbytearray; pos:integer): integer;
var
 p,v:integer;
begin
  v := 0;
  p := pos+16;
  repeat
    if pboard[p] <> 0 then break;
    inc(p,16);
    inc(v);
  until false;
  p := pos-16;
  repeat
    if pboard[p] <> 0 then break;
    dec(p,16);
    inc(v);
  until false;
  p := pos+1;
  repeat
    if pboard[p] <> 0 then break;
    inc(p);
    inc(v);
  until false;
  p := pos-1;
  repeat
    if pboard[p] <> 0 then break;
    dec(p);
    inc(v);
  until false;
  result := VROOKMOB[v] - 100;
end;

function AAttackedByPawn(pboard:pointer; pos:integer): integer;
asm
  add edx, eax
  xor eax, eax
  cmp byte ptr[edx+15], BPAWN
  setz cl
  cmp byte ptr[edx+17], BPAWN
  setz dl
  or al, cl
  or al, dl
end;

function BAttackedByPawn(pboard:pointer; pos:integer): integer;
asm
  add edx, eax
  xor eax, eax
  cmp byte ptr[edx-17], APAWN
  setz cl
  cmp byte ptr[edx-15], APAWN
  setz dl
  or al, cl
  or al, dl
end;

procedure EvaluateARook(var edat:TEvalData; var pdat:TPawnData);
var
 w,x,y,kbx,kby,t,v: integer;
 pseven: PIntArray;
 pboard: PByteArray;
begin
  w := edat.Pos;
  x := w and 7;
  y := w shr 4;
  kbx := edat.BKingPos and 7;
  kby := edat.BKingPos shr 4;
  pboard := edat.pBoard;
  pseven := pointer(pboard);
  inc(edat.AMat,(VROOK-19) + AttackRook(pboard,w));
  inc(edat.VPos,VROOKPOS[TO8[w]]);
  if y =_R7 then begin
     if (kby = _R8) or ((pseven[24] or pseven[25]) and BPAWNMASK <> 0) then begin
        inc(edat.Amat,28);
        if (kby = _R8) and (pdat.APass <> 0) then inc(edat.AMat,15);
        if (kby >= _R7) and (edat.ARookPos1 shr 4 = _R7) then inc(edat.AMat,20);
     end;
  end;
  edat.ARookPos2 := edat.ARookPos1;
  edat.ARookPos1 := w;
  t := 1 shl x;
  if (t and pdat.ACol) = 0 then begin
     v := 8;
     if (t and pdat.BCol) = 0 then inc(v,16);
  end
  else begin
     v := 0;
     if (w >= _A3) and (pboard[_A2 + x] = APAWN) then inc(v,4);
  end;
  if abs(y-kby) <= 1 then inc(edat.BClose,8)
     else if (v > 0) then inc(edat.BClose,VROOKKING[abs(x-kbx)]);
  inc(edat.AMat,v);
  if (pboard[w-1] <> 0) and (pboard[w+1] <> 0) then begin
     dec(edat.AMat,10);
     if pboard[w+16] <> 0 then dec(edat.AMat,4);
     if pboard[w-16] <> 0 then dec(edat.AMat,6);
  end;
end;

procedure EvaluateBRook(var edat:TEvalData; var pdat:TPawnData);
var
 w,x,y,kax,kay,t,v: integer;
 pseven: PIntArray;
 pboard: PByteArray;
begin
  w := edat.Pos;
  x := w and 7;
  y := w shr 4;
  kax := edat.AKingPos and 7;
  kay := edat.AKingPos shr 4;
  pboard := edat.pBoard;
  pseven := pointer(pboard);
  inc(edat.BMat,(VROOK-19) + AttackRook(pboard,w));
  dec(edat.VPos,VROOKPOS[TO8[w]]);
  if y =_R2 then begin
     if (kay = _R1) or ((pseven[4] or pseven[5]) and APAWNMASK <> 0) then begin
        inc(edat.Bmat,28);
        if (kay = _R1) and (pdat.BPass <> 0) then inc(edat.BMat,15);
        if (kay <= _R2) and (edat.BRookPos1 shr 4 = _R2) then inc(edat.BMat,20);
     end;
  end;
  edat.BRookPos2 := edat.BRookPos1;
  edat.BRookPos1 := w;
  t := 1 shl x;
  if (t and pdat.BCol) = 0 then begin
     v := 8;
     if (t and pdat.ACol) = 0 then inc(v,16);
  end
  else begin
     v := 0;
     if (w <= _H6) and (pboard[_A7 + x] = BPAWN) then inc(v,4);
  end;
  if abs(y-kay) <= 1 then inc(edat.AClose,8)
     else if (v > 0) then inc(edat.AClose,VROOKKING[abs(x-kax)]);
  inc(edat.BMat,v);
  if (pboard[w-1] <> 0) and (pboard[w+1] <> 0) then begin
     dec(edat.BMat,10);
     if pboard[w-16] <> 0 then dec(edat.BMat,4);
     if pboard[w+16] <> 0 then dec(edat.BMat,6);
  end;
end;

procedure EvaluateAQueen(var edat:TEvalData; var pdat:TPawnData);
var w,x,y,kax,kbx,kby,d: integer;
begin
  w := edat.Pos;
  edat.AQueenPos := w;
  x := w and 7;
  y := w shr 4;
  kax := edat.AKingPos and 7;
  kbx := edat.BKingPos and 7;
  kby := edat.BKingPos shr 4;
  inc(edat.AMat,(VQUEEN+3) + VQUEENPOS[TO8[w]] - abs(x-kbx) - abs(x-kax));
  d := DIST16[Edat.BKingPos-w+120];
  if d > 4 then begin
     if (abs(x-kbx) <= 1) or (abs(y-kby) <= 1) then inc(edat.BClose,6)
        else if AQueenAttack(edat.pBoard,w,(x < kbx))
             then inc(edat.BClose,6);
  end
  else begin
     inc(edat.BClose,VQUEENATTACK[d]);
     if d = 1 then begin
        inc(edat.AMat,25);
     end;
  end;
end;

function AQueenAttack(pboard:pbytearray; pos:integer; kside:boolean): boolean;
var z:integer;
begin
  if kside then z := 17 else z := 15;
  result := false;
  repeat
     inc(pos,z);
     result := result or
              (pboard[pos+16] = BKING) or (pboard[pos-16] = BKING) or
              (pboard[pos+32] = BKING) or (pboard[pos-32] = BKING);
  until pboard[pos] <> 0;
  inc(pos,z);
  result := result or (pboard[pos] = BKING) or
           (pboard[pos-16] = BKING) or (pboard[pos-32] = BKING);
end;

procedure EvaluateBQueen(var edat:TEvalData; var pdat:TPawnData);
var w,x,y,kbx,kax,kay,d: integer;
begin
  w := edat.Pos;
  edat.BQueenPos := w;
  x := w and 7;
  y := w shr 4;
  kbx := edat.BKingPos and 7;
  kax := edat.AKingPos and 7;
  kay := edat.AKingPos shr 4;
  inc(edat.BMat,(VQUEEN+3) + VQUEENPOS[63-TO8[w]] - abs(x-kax) - abs(x-kbx));
  d := DIST16[Edat.AKingPos-w+120];
  if d > 4 then begin
     if (abs(x-kax) <= 1) or (abs(y-kay) <= 1) then inc(edat.AClose,6)
        else if BQueenAttack(edat.pBoard,w,(x < kax))
             then inc(edat.AClose,6);
  end
  else begin
     inc(edat.AClose,VQUEENATTACK[d]);
     if d = 1 then begin
        inc(edat.BMat,25);
     end;
  end;
end;

function BQueenAttack(pboard:pbytearray; pos:integer; kside:boolean): boolean;
var z:integer;
begin
  if kside then z := 15 else z := 17;
  result := false;
  repeat
     dec(pos,z);
     result := result or
               (pboard[pos+16] = AKING) or (pboard[pos-16] = AKING) or
               (pboard[pos+32] = AKING) or (pboard[pos-32] = AKING);
  until pboard[pos] <> 0;
  dec(pos,z);
  result := result or (pboard[pos] = AKING) or
           (pboard[pos+16] = AKING) or (pboard[pos+32] = AKING);
end;

procedure EvaluateRookQueen(var edat:TEvalData; var pdat:TPawnData);
var
 x: integer;
const
 NONE = $ff;
 COLUMNMASK: array[0..9] of byte = (255,1,2,4,8,16,32,64,128,255);
begin
  if (edat.BQueenPos <> NONE) then begin
     x := edat.BQueenPos and 7;
     if (pdat.ACol and pdat.BCol and COLUMNMASK[x + 1]) = 0 then begin
        if ((edat.ARookPos1 and 15) = x) then dec(edat.BMat,10);
        if ((edat.ARookPos2 and 15) = x) then dec(edat.BMat,10);
     end;
  end;
  if (edat.AQueenPos <> NONE) then begin
     x := edat.AQueenPos and 7;
     if (pdat.ACol and pdat.BCol and COLUMNMASK[x + 1]) = 0 then begin
        if ((edat.BRookPos1 and 15) = x) then dec(edat.AMat,10);
        if ((edat.BRookPos2 and 15) = x) then dec(edat.AMat,10);
     end;
  end;
end;

procedure EvaluatePawnMiddlegame(var edat:TEvalData; var pdat:TPawnData);
var
 t: integer;
 pboard: PByteArray;
const
 NONE = $ff;
 CLN_E = 4;
begin
  pboard := edat.pBoard;
  if (edat.AKingPos and 7) >= CLN_E then begin
     if ((pboard[_H3] or pboard[_G3] or pboard[_F3]) and P_BPAWNS) <> 0
         then inc(edat.AClose,4);
     if ((pboard[_E4] or pboard[_E3]) and P_BPAWNS) <> 0
        then inc(edat.AClose,2);
     if (pboard[_C3] = APAWN) then inc(edat.VPos,4);
     if (pboard[_C4] = APAWN) then inc(edat.VPos,8);
     if (pboard[_C5] = APAWN) then inc(edat.VPos,2);
     if edat.BQueenPos <> NONE then begin
        t := pdat.Shield and $ff;
        inc(edat.AClose,t);
        dec(edat.VPos,t*7);
        inc(edat.VPos,((integer(pboard[edat.AKingPos+15]) and 16) +
                       (integer(pboard[edat.AKingPos+16]) and 16)) shr 3);
     end;
  end
  else begin
     if ((pboard[_A3] or pboard[_B3] or pboard[_C3]) and P_BPAWNS) <> 0
        then inc(edat.AClose,4);
     if ((pboard[_D4] or pboard[_D3]) and P_BPAWNS) <> 0
        then inc(edat.AClose,2);
     if (pboard[_F3] = APAWN) then inc(edat.VPos,4);
     if (pboard[_F4] = APAWN) then inc(edat.VPos,8);
     if (pboard[_F5] = APAWN) then inc(edat.VPos,2);
     if edat.BQueenPos <> NONE then begin
        t := (pdat.Shield shr 8) and $ff;
        inc(edat.AClose,t);
        dec(edat.VPos,t*7);
        inc(edat.VPos,((integer(pboard[edat.AKingPos+17]) and 16) +
                       (integer(pboard[edat.AKingPos+16]) and 16)) shr 3);
     end;
  end;
  if (edat.BKingPos and 7) >= CLN_E then begin
     if ((pboard[_H6] or pboard[_G6] or pboard[_F6]) and P_APAWNS) <> 0
        then inc(edat.BClose,4);
     if ((pboard[_E5] or pboard[_E6]) and P_APAWNS) <> 0
        then inc(edat.BClose,2);
     if (pboard[_C6] = BPAWN) then dec(edat.VPos,4);
     if (pboard[_C5] = BPAWN) then dec(edat.VPos,8);
     if (pboard[_C4] = BPAWN) then dec(edat.VPos,2);
     if edat.AQueenPos <> NONE then begin
        t := (pdat.Shield shr 16) and $ff;
        inc(edat.BClose,t);
        inc(edat.VPos,t*7);
        dec(edat.VPos,((integer(pboard[edat.BKingPos-17]) and 8) +
                       (integer(pboard[edat.BKingPos-16]) and 8)) shr 2);
     end;
  end
  else begin
     if ((pboard[_A6] or pboard[_B6] or pboard[_C6]) and P_APAWNS) <> 0
        then inc(edat.BClose,4);
     if ((pboard[_D5] or pboard[_D6]) and P_APAWNS) <> 0
        then inc(edat.BClose,2);
     if (pboard[_F6] = BPAWN) then dec(edat.VPos,4);
     if (pboard[_F5] = BPAWN) then dec(edat.VPos,8);
     if (pboard[_F4] = BPAWN) then dec(edat.VPos,2);
     if edat.AQueenPos <> NONE then begin
        t := (pdat.Shield shr 24);
        inc(edat.BClose,t);
        inc(edat.VPos,t*7);
        dec(edat.VPos,((integer(pboard[edat.BKingPos-15]) and 8) +
                       (integer(pboard[edat.BKingPos-16]) and 8)) shr 2);
     end;
  end;
end;

function ADevelop(pboard:pointer): integer;
asm
  push esi
  mov esi,eax
  cmp byte ptr [eax + _B1], AKNIGHT
  setnz cl
  xor eax,eax
  cmp byte ptr [esi + _C1], ABISHOP
  setnz dl
  cmp byte ptr [esi + _G1], AKNIGHT
  setnz al
  add dl,cl
  cmp byte ptr [esi + _F1], ABISHOP
  setnz cl
  add al,dl
  pop esi
  add al,cl
end;

function BDevelop(pboard:pointer): integer;
asm
  push esi
  mov esi,eax
  cmp byte ptr [eax + _B8], BKNIGHT
  setnz cl
  xor eax,eax
  cmp byte ptr [esi + _C8], BBISHOP
  setnz dl
  cmp byte ptr [esi + _G8], BKNIGHT
  setnz al
  add dl,cl
  cmp byte ptr [esi + _F8], BBISHOP
  setnz cl
  add al,dl
  pop esi
  add al,cl
end;

procedure EvaluateNull(var edat:TEvalData; var pdat:TPawnData);
begin
end;

procedure SetRandom(isrand:boolean);
const
 ORIVKNIGHTPOS: array[0..63] of byte = (
 00,14,24,31,31,24,14,00,
 18,26,36,40,40,36,26,18,
 26,35,48,50,50,48,35,26,
 32,40,50,55,55,50,40,32,
 34,43,53,59,59,53,43,34,
 28,38,47,56,56,47,38,28,
 19,27,39,44,44,39,27,19,
 00,18,26,33,33,26,18,00
 );
var i,n: integer;
begin
  randomize;
  for i := 0 to 63 do begin
      if isrand then n := random(2) else n := 0;
      VKNIGHTPOS[i] := ORIVKNIGHTPOS[i] + n;
  end;
end;

initialization
  AEvalFunction[0] := EvaluateNull;
  AEvalFunction[1] := EvaluateNull;
  AEvalFunction[2] := EvaluateNull;
  AEvalFunction[3] := EvaluateNull;
  AEvalFunction[4] := EvaluateAKnight;
  AEvalFunction[5] := EvaluateABishop;
  AEvalFunction[6] := EvaluateARook;
  AEvalFunction[7] := EvaluateAQueen;
  BEvalFunction[0] := EvaluateNull;
  BEvalFunction[1] := EvaluateNull;
  BEvalFunction[2] := EvaluateNull;
  BEvalFunction[3] := EvaluateNull;
  BEvalFunction[4] := EvaluateBKnight;
  BEvalFunction[5] := EvaluateBBishop;
  BEvalFunction[6] := EvaluateBRook;
  BEvalFunction[7] := EvaluateBQueen;
  SetRandom(false);
  pawncache := TPawnCache.Create;
  evalcache := TEvalCache.Create;

finalization
  evalcache.Free;
  pawncache.Free;


end.
