unit uGenMove;

interface

uses Sysutils, uMisc, uBoard, uGenChk;


function AGenerateX(pcb:PChessBrd; pmovearray:pinteger;
                   incheck:integer; legal:boolean): integer;
function BGenerateX(pcb:PChessBrd; pmovearray:pinteger;
                   incheck:integer; legal:boolean): integer;
function AIsAttackedX(pcb:PChessBrd; pos:integer; slideonly:boolean): boolean;
function BIsAttackedX(pcb:PChessBrd; pos:integer; slideonly:boolean): boolean;
function AIsLegalMoveX(pcb:PChessBrd; m:integer; incheck:boolean): boolean;
function BIsLegalMoveX(pcb:PChessBrd; m:integer; incheck:boolean): boolean;
function UnderCheckX(pcb:PChessBrd; blacktomove:boolean): integer;
function CheckMoveX(pcb:PChessBrd; m:integer; btm:boolean): boolean;
function LegalMovesX(pcb:PChessBrd; btm:boolean): integer;
function PVToStringSAN(pcb:PChessBrd; pvarray:PIntArray): string;
function MoveToSAN(pcb:PChessBrd; m:integer): string;


implementation


const
 AKNIGHTVEC: array[0..8] of integer = (14,-14,18,-18,31,-31,33,-33,0);
 AKINGVEC: array[0..8] of integer = (1,-1,16,-16,17,-17,15,-15,0);
 DIAG16: array[0..239] of byte = (
  00, 04, 00, 00, 00, 00, 00, 00, 01, 00, 00, 00, 00, 00, 00, 08,
  00, 00, 04, 00, 00, 00, 00, 00, 01, 00, 00, 00, 00, 00, 08, 00,
  00, 00, 00, 04, 00, 00, 00, 00, 01, 00, 00, 00, 00, 08, 00, 00,
  00, 00, 00, 00, 04, 00, 00, 00, 01, 00, 00, 00, 08, 00, 00, 00,
  00, 00, 00, 00, 00, 04, 00, 00, 01, 00, 00, 08, 00, 00, 00, 00,
  00, 00, 00, 00, 00, 00, 04, 16, 01, 16, 08, 00, 00, 00, 00, 00,
  00, 00, 00, 00, 00, 00, 16, 04, 01, 08, 16, 00, 00, 00, 00, 00,
  00, 02, 02, 02, 02, 02, 02, 02, 00, 02, 02, 02, 02, 02, 02, 02,
  00, 00, 00, 00, 00, 00, 16, 08, 01, 04, 16, 00, 00, 00, 00, 00,
  00, 00, 00, 00, 00, 00, 08, 16, 01, 16, 04, 00, 00, 00, 00, 00,
  00, 00, 00, 00, 00, 08, 00, 00, 01, 00, 00, 04, 00, 00, 00, 00,
  00, 00, 00, 00, 08, 00, 00, 00, 01, 00, 00, 00, 04, 00, 00, 00,
  00, 00, 00, 08, 00, 00, 00, 00, 01, 00, 00, 00, 00, 04, 00, 00,
  00, 00, 08, 00, 00, 00, 00, 00, 01, 00, 00, 00, 00, 00, 04, 00,
  00, 08, 00, 00, 00, 00, 00, 00, 01, 00, 00, 00, 00, 00, 00, 04
 );

function AGeneratePawns(posmem:pbytearray; pcurrmove:pinteger): pointer; forward;
function AGeneratePromotePawns(posmem:pbytearray; pmoves:pintarray): pointer; forward;
function AGenerateRook(posmem,pcurrmove:pointer): pointer; forward;
function AGenerateBishop(posmem,pcurrmove:pointer): pointer; forward;
function AGenerateRound(posmem,pcurrmove,parray:pointer): pointer; forward;
function BGenerateRound(posmem,pcurrmove,parray:pointer): pointer; forward;
function BGeneratePawns(posmem:pbytearray; pcurrmove:pinteger): pointer; forward;
function BGeneratePromotePawns(posmem:pbytearray; pmoves:pintarray): pointer; forward;
function BGenerateRook(posmem,pcurrmove:pointer): pointer; forward;
function BGenerateBishop(posmem,pcurrmove:pointer): pointer; forward;
function BDRayDiag(posmem:integer): integer; forward;
function BYRayCross(posmem:integer): integer; forward;
function BGenerateKCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer; forward;
function BGenerateQCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer; forward;


function AGenerateQCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer;
var board: pintarray;
begin
  board := @(pcb.BoardArray[BOARDOFFSET]);
  if (board[0] = AROOK) and
     not AisAttackedX(pcb,_D1,false) and
     not AisAttackedX(pcb,_C1,false) then
     begin
        pcurrmove^ := _C1 + (_E1 shl 7) + MCASTLE;
        inc(integer(pcurrmove),4);
     end;
  result := pcurrmove;
end;

function AGenerateKCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer;
var board: pintarray;
begin
  board := @(pcb.BoardArray[BOARDOFFSET]);
  if (board[1] = AKING + AROOK shl 24) and
     not AisAttackedX(pcb,_F1,false) and
     not AisAttackedX(pcb,_G1,false) then
     begin
        pcurrmove^ := _G1 + (_E1 shl 7) + MCASTLE;
        inc(integer(pcurrmove),4);
     end;
  result := pcurrmove;
end;

function AYRayCross(posmem:integer): integer;
asm
  push ebx
  xor ebx, ebx
  push edi
  lea ecx, [eax + 16]  
  push esi
  mov edi, 1
  mov esi, eax
@xn:
  mov dl, byte ptr [ecx]
  add ecx, 16
  cmp dl, 0
  setnz al
  cmp dl, BROOK
  jz @yend
  add bl, al
  cmp dl, BQUEEN
  jz @yend
  cmp bl, 2
  jnz @xn

  lea ecx, [esi - 32]
@xs:
  mov dl, byte ptr [ecx]
  sub ecx, 16
  cmp dl, BROOK
  jz @yend
  cmp dl, BQUEEN
  jz @yend
  cmp dl, NBORDER
  jnz @xs

  xor edi, edi
@yend:
  pop esi
  mov eax, edi
  pop edi
  pop ebx
end;

function ADRayDiag(posmem:integer): integer;
asm
  push edi
  xor edi, edi
  lea ecx, [eax + 17]
  push ebx
  xor ebx, ebx
  push esi
  mov esi, eax

@xne:
  mov dl, byte ptr [ecx]
  add ecx, 17
  cmp dl, 0
  setnz al
  cmp dl, BBISHOP
  jz @xraysw
  add bl, al
  cmp dl, BQUEEN
  jz @xraysw
  cmp bl, 2
  jz @xsw1
  jmp @xne
@xrayne:
  or edi, 4

@xsw1:
  lea ecx, [esi - 34]
@xsw:
  mov dl, byte ptr [ecx]
  sub ecx, 17
  cmp dl, BBISHOP
  jz @xraysw
  cmp dl, BQUEEN
  jz @xraysw
  cmp dl, NBORDER
  jz @xnw1
  jmp @xsw
@xraysw:
  or edi, 4

@xnw1:
  lea ecx, [esi + 15]
  mov bl, 0
@xnw:
  mov dl, byte ptr [ecx]
  add ecx, 15
  cmp dl, 0
  setnz al
  cmp dl, BBISHOP
  jz @xrayse
  add bl, al
  cmp dl, BQUEEN
  jz @xrayse
  cmp bl, 2
  jz @xse1
  jmp @xnw
@xraynw:
  or edi, 8

@xse1:
  lea ecx, [esi - 30]
@xse:
  mov dl, byte ptr [ecx]
  sub ecx, 15
  cmp dl, BBISHOP
  jz @xrayse
  cmp dl, BQUEEN
  jz @xrayse
  cmp dl, NBORDER
  jz @xend
  jmp @xse
@xrayse:
  or edi, 8

@xend:
  pop esi
  mov eax, edi
  pop ebx
  pop edi
end;

function AIsLegalMoveX(pcb:PChessBrd; m:integer; incheck:boolean): boolean;
var
 startpos,endpos: integer;
 endpiece: byte;
 isking: boolean;
 pboard: pbytearray;
begin
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  startpos := (m shr 7) and 127;
  isking := pboard[startpos] = AKING;
  endpos := m and 127;
  endpiece := pboard[endpos];
  if (m and MENPASSANT <> 0) then pboard[endpos-16] := 0;
  pboard[endpos] := pboard[startpos];
  pboard[startpos] := 0;
  if isking
     then result := not AIsAttackedX(pcb,endpos,false)
     else result := not AIsAttackedX(pcb,pcb.AKingPos,not incheck);
  pboard[startpos] := pboard[endpos];
  pboard[endpos] := endpiece;
  if (m and MENPASSANT <> 0) then pboard[endpos-16] := BPAWN;
end;

function AIsAttackedX(pcb:PChessBrd; pos:integer; slideonly:boolean): boolean;
var
 mat: integer;
 pb: ^byte;
 t: byte;
 pboard: pbytearray;
const
 COFF = 34;
begin
  result := true;
  mat := pcb.BMatCnt;
  if not slideonly then begin
     pboard := @(pcb.BoardArray[BOARDOFFSET-COFF+pos]);
     if (pboard[COFF+15] = BPAWN) or (pboard[COFF+17] = BPAWN) then exit;
     if mat and $0f00 <> 0 then if
        (pboard[COFF+14] = BKNIGHT) or (pboard[COFF-14] = BKNIGHT) or
        (pboard[COFF+18] = BKNIGHT) or (pboard[COFF-18] = BKNIGHT) or
        (pboard[COFF+31] = BKNIGHT) or (pboard[COFF-31] = BKNIGHT) or
        (pboard[COFF+33] = BKNIGHT) or (pboard[COFF-33] = BKNIGHT)
        then exit;
     if (pcb.BKingPos-pos <= 17) and (pos-pcb.BKingPos <= 17) and
        (abs((pos and 7)-(pcb.BKingPos and 7)) <= 1) then exit;
  end;
  pboard := @(pcb.BoardArray[BOARDOFFSET+pos]);
  if (pcb.BXCross and ($0f shl ((pos and 7) shl 2))) <> 0 then begin
     pb := pointer(integer(pboard) + 16);
     repeat
       t := pb^;
       inc(integer(pb),16);
     until t <> 0;
     pb := pointer(integer(pboard) + -16);
     if (t = BROOK) or (t = BQUEEN) then exit;
     repeat
       t := pb^;
       dec(integer(pb),16);
     until t <> 0;
     if (t = BROOK) or (t = BQUEEN) then exit;
  end;
  if (pcb.BYCross and ($0f shl ((pos shr 4) shl 2))) <> 0 then begin
     pb := pointer(integer(pboard) + -1);
     repeat
       t := pb^;
       dec(integer(pb));
     until t <> 0;
     pb := pointer(integer(pboard) + 1);
     if (t = BROOK) or (t = BQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb));
     until t <> 0;
     if (t = BROOK) or (t = BQUEEN) then exit;
  end;
  if (mat and $f0f000 <> 0) then begin
     pb := pointer(integer(pboard) + 15);
     repeat
       t := pb^;
       inc(integer(pb),15);
     until t <> 0;
     pb := pointer(integer(pboard) + 17);
     if (t = BBISHOP) or (t = BQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb),17);
     until t <> 0;
     pb := pointer(integer(pboard) + -17);
     if (t = BBISHOP) or (t = BQUEEN) then exit;
     repeat
       t := pb^;
       dec(integer(pb),17);
     until t <> 0;
     pb := pointer(integer(pboard) + -15);
     if (t = BBISHOP) or (t = BQUEEN) then exit;
     repeat
       t := pb^;
       dec(integer(pb),15);
     until t <> 0;
     if (t = BBISHOP) or (t = BQUEEN) then exit;
  end;
  result := false;
end;

function AGenerateX(pcb:PChessBrd; pmovearray:pinteger;
                   incheck:integer; legal:boolean): integer;
var
 pos,posmem,i,j,m,movecntnk,mv,ds,de,xray,n,pinpos: integer;
 pcurrmove: pinteger;
 pboard: pbytearray;
 movelist: pintarray;
 thelist: array[0..255] of integer;
label SkipGen;
begin
  pinpos := -1;
  movelist := @(thelist);
  pcurrmove := @(thelist);
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  if incheck = 0 then begin
     xray := 0;
     pos := pcb.AKingPos;
     posmem := integer(pboard) + pos;
     if (pcb.BXCross and ($0f shl ((pos and 7) shl 2))) <> 0
        then inc(xray,AYRayCross(posmem));
     if (pcb.BYCross and ($0f shl ((pos shr 4) shl 2))) <> 0
        then inc(xray,2);
     if (pcb.BMatCnt and $f0f000) <> 0
        then inc(xray,ADRayDiag(posmem));
  end
  else begin
     xray := 15;
     legal := true;
     if (incheck >= 256) and (pcb.EPSquare = 255) then begin
        pos := (incheck shr 10) and 127;
        posmem := integer(pboard) + pos;
        pos := pos*(128+1);
        pcurrmove^ := pos;
        pcurrmove := ACCapPawn(posmem,pcurrmove);
        pcurrmove^ := pos;
        if (pcb.AMatCnt and $0f00) <> 0 then begin
           pcurrmove := CCapKnight(posmem,AKNIGHT,pcurrmove);
           pcurrmove^ := pos;
        end;
        if (pcb.AMatCnt and $f0f000) <> 0 then begin
           pcurrmove := ACCapBishop(posmem,pcurrmove);
           pcurrmove^ := pos;
        end;
        if (pcb.AMatCnt and $ff0000) <> 0 then
           pcurrmove := ACCapRook(posmem,pcurrmove);
        goto SkipGen;
     end;
  end;
  if pcb.EPSquare <> 255 then begin
     pos := pcb.EPSquare;
     if pboard[pos-17] = APAWN then begin
        pcurrmove^ := ((pos-17) shl 7 + pos) or
                      (MCAPTURE + MENPASSANT);
        inc(integer(pcurrmove),4);
     end;
     if pboard[pos-15] = APAWN then begin
        pcurrmove^ := ((pos-15) shl 7 + pos) or
                      (MCAPTURE + MENPASSANT);
        inc(integer(pcurrmove),4);
     end;
  end;
  n := 0;
  while true do begin
     pos := pcb.ALoc[n];
     inc(n);
     if pos = LOCLAST then break;
     posmem := integer(pboard) + pos;
     pcurrmove^ := pos*(128+1);
     case (pboard[pos] and 7) of
      PC_NULL0: continue;
      PC_NULL1: continue;
      PC_PAWN: pcurrmove := AGeneratePawns(pointer(posmem),pcurrmove);
      PC_KNIGHT: pcurrmove := AGenerateRound(pointer(posmem),pcurrmove,@AKNIGHTVEC);
      PC_BISHOP: pcurrmove := AGenerateBishop(pointer(posmem),pcurrmove);
      PC_ROOK: pcurrmove := AGenerateRook(pointer(posmem),pcurrmove);
      PC_QUEEN: begin
        m := pcurrmove^;
        pcurrmove := AGenerateBishop(pointer(posmem),pcurrmove);
        pcurrmove^ := m;
        pcurrmove := AGenerateRook(pointer(posmem),pcurrmove);
      end;
     end;
  end;
SkipGen:
  movecntnk := (integer(pcurrmove) - integer(movelist)) shr 2;
  pcurrmove^ := pcb.AKingPos * (128+1);
  posmem := pcb.AKingPos + integer(pboard);
  if not legal then inc(pcurrmove^,MKINGNLEGAL);
  pcurrmove := AGenerateRound(pointer(posmem),pcurrmove,@AKINGVEC);
  if incheck = 0 then begin
     if (pcb.Castle and AKCASTLE) <> 0 then
        pcurrmove := AGenerateKCastle(pcb,pcurrmove);
     if (pcb.Castle and AQCASTLE) <> 0 then
        pcurrmove := AGenerateQCastle(pcb,pcurrmove);
  end;
  m := (integer(pcurrmove) - integer(movelist)) shr 2;
  if m > 100 then m := 100;
  pcurrmove := pmovearray;
  j := pcb.AKingPos + 120;
  for i := 0 to m-1 do begin
     mv := movelist[i];
     pmovearray^ := mv;
     inc(integer(pmovearray),4);
     if (xray = 0) and (i < movecntnk) then continue;
     if i < movecntnk then begin
        ds := DIAG16[j - ((mv shr 7) and 127)];
        de := DIAG16[j - (mv and 127)];
        if incheck = 0 then begin
           ds := ds and 15;
           if (ds and xray = 0) or (ds = de) then continue;
           if (mv and $3f80) = pinpos then begin
              dec(integer(pmovearray),4);
              continue;
           end;
        end
        else begin
           if (de and incheck = 0) or ((mv and MCAPTURE) + ds = de) or
              ((mv and MCAPTURE) + de and $ff0f = 0)
           then begin
              dec(integer(pmovearray),4);
              continue;
           end;
        end;
     end;
     if (i < movecntnk) or legal then
        if not AisLegalMoveX(pcb,mv,incheck <> 0) then begin
           pinpos := mv and ($3f80 + MENPASSANT);
           dec(integer(pmovearray),4);
        end;
  end;
  m := (integer(pmovearray) - integer(pcurrmove)) shr 2;
  if m > 0 then result := m
  else begin
     if incheck <> 0 then result := -1 else result := -2;
  end;
end;

function AGenerateRound(posmem,pcurrmove,parray:pointer): pointer;
asm
  push esi
  mov esi, eax
  push edi
  mov edi, edx
  push ebx
  mov eax, [edx]
  push ebp
  mov ebp, ecx
  xor ebx, ebx
  mov edx, [ebp]

@cyc:
  mov dl, byte ptr [esi+edx]
  add ebp, 4
  test dl, A_OR_BORDER
  setz bl
  test dl, B_PIECES
  setnz dl
  mov ecx, [ebp-4]
  and edx, 1
  add ecx, eax
  shl edx, 14
  add ecx, edx
  mov edx, [ebp]
  mov dword ptr [edi], ecx
  lea edi, [edi + ebx*4]
  cmp edx, 0
  jnz @cyc

@end:
  pop ebp
  pop ebx
  mov eax, edi
  pop edi
  pop esi
end;

function AGeneratePawns(posmem:pbytearray; pcurrmove:pinteger): pointer;
var y,themove: integer;
begin
  themove := pcurrmove^;
  y := (themove shr 4) and 7;
  if y >= _R6 then begin
     if y = _R7 then begin
        result := AGeneratePromotePawns(posmem,pointer(pcurrmove));
        exit;
     end;
     themove := themove or MPROMOTIONTHREAT;
  end;
  if posmem[16] = 0 then begin
     pcurrmove^ := themove + 16;
     inc(integer(pcurrmove),4);
     if y = _R2 then if posmem[32] = 0 then begin
        pcurrmove^ := themove + 32;
        inc(integer(pcurrmove),4);
     end;
  end;
  if (posmem[15] and B_PIECES) <> 0 then begin
     pcurrmove^ := themove + (15 + MCAPTURE);
     inc(integer(pcurrmove),4);
  end;
  if (posmem[17] and B_PIECES) <> 0 then begin
     pcurrmove^ := themove + (17 + MCAPTURE);
     inc(integer(pcurrmove),4);
  end;
  result := pcurrmove;
end;

function AGeneratePromotePawns(posmem:pbytearray; pmoves:pintarray): pointer;
var themove: integer;
begin
  themove := pmoves[0];
  if posmem[16] = 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN + 16);
     pmoves[1] := themove + (MPROMOTEDROOK + 16);
     pmoves[2] := themove + (MPROMOTEDBISHOP + 16);
     pmoves[3] := themove + (MPROMOTEDKNIGHT + 16);
     inc(integer(pmoves),16);
  end;
  if (posmem[15] and B_PIECES) <> 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN + MCAPTURE + 15);
     pmoves[1] := themove + (MPROMOTEDROOK + MCAPTURE + 15);
     pmoves[2] := themove + (MPROMOTEDBISHOP + MCAPTURE + 15);
     pmoves[3] := themove + (MPROMOTEDKNIGHT + MCAPTURE + 15);
     inc(integer(pmoves),16);
  end;
  if (posmem[17] and B_PIECES) <> 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN + MCAPTURE + 17);
     pmoves[1] := themove + (MPROMOTEDROOK + MCAPTURE + 17);
     pmoves[2] := themove + (MPROMOTEDBISHOP + MCAPTURE + 17);
     pmoves[3] := themove + (MPROMOTEDKNIGHT + MCAPTURE + 17);
     inc(integer(pmoves),16);
  end;
  result := pmoves;
end;

function AGenerateRook(posmem,pcurrmove:pointer): pointer;
asm
  push ebx
  xor ebx, ebx
  push edi
  mov ecx, [edx];
  mov edi, edx
  push esi
  mov esi, eax

  lea edx, [esi+16]
  mov eax, ecx
@dn:
  add eax, 16
  cmp byte ptr [edx], 0
  jnz @dsi
  add edx, 16
  mov [edi], eax
  add edi, 4
  jmp @dn
@dsi:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-16]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@ds:
  sub eax, 16
  cmp byte ptr [edx], 0
  jnz @dwi
  sub edx, 16
  mov [edi], eax
  add edi, 4
  jmp @ds
@dwi:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-1]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dw:
  dec eax
  cmp byte ptr [edx], 0
  jnz @dei
  dec edx
  mov [edi], eax
  add edi, 4
  jmp @dw
@dei:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi+1]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@de:
  inc eax
  cmp byte ptr [edx], 0
  jnz @end
  inc edx
  mov [edi], eax
  add edi, 4
  jmp @de
@end:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  pop esi
  lea eax, [edi + ebx*4]
  pop edi
  pop ebx
end;

function AGenerateBishop(posmem,pcurrmove:pointer): pointer;
asm
  push ebx
  xor ebx, ebx
  push edi
  mov ecx, [edx];
  mov edi, edx
  push esi
  mov esi, eax

  lea edx, [esi+17]
  mov eax, ecx
@dne:
  add eax, 17
  cmp byte ptr [edx], 0
  jnz @dswi
  add edx, 17
  mov [edi], eax
  add edi, 4
  jmp @dne
@dswi:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-17]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dsw:
  sub eax, 17
  cmp byte ptr [edx], 0
  jnz @dsei
  sub edx, 17
  mov [edi], eax
  add edi, 4
  jmp @dsw
@dsei:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-15]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dse:
  sub eax, 15
  cmp byte ptr [edx], 0
  jnz @dnwi
  sub edx, 15
  mov [edi], eax
  add edi, 4
  jmp @dse
@dnwi:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi+15]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dnw:
  add eax, 15
  cmp byte ptr [edx], 0
  jnz @end
  add edx, 15
  mov [edi], eax
  add edi, 4
  jmp @dnw
@end:
  add eax, MCAPTURE
  test byte ptr [edx], B_PIECES;
  setnz bl
  mov [edi], eax

  pop esi
  lea eax, [edi + ebx*4]
  pop edi
  pop ebx
end;

function BGenerateX(pcb:PChessBrd; pmovearray:pinteger;
                   incheck:integer; legal:boolean): integer;
var
 pos,posmem,i,j,m,movecntnk,mv,ds,de,xray,n,pinpos: integer;
 pcurrmove: pinteger;
 pboard: pbytearray;
 movelist: pintarray;
 thelist: array[0..255] of integer;
label SkipGen;
begin
  pinpos := -1;
  movelist := @thelist;
  pcurrmove := @thelist;
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  if incheck = 0 then begin
     xray := 0;
     pos := pcb.BKingPos;
     posmem := integer(pboard) + pos;
     if (pcb.AXCross and ($0f shl ((pos and 7) shl 2))) <> 0
        then inc(xray,BYRayCross(posmem));
     if (pcb.AYCross and ($0f shl ((pos shr 4) shl 2))) <> 0
        then inc(xray,2);
     if (pcb.AMatCnt and $f0f000) <> 0
        then inc(xray,BDRayDiag(posmem));
  end
  else begin
     xray := 15;
     legal := true;
     if (incheck >= 256) and (pcb.EPSquare = 255) then begin
        pos := (incheck shr 10) and 127;
        posmem := integer(pboard) + pos;
        pos := pos*(128+1);
        pcurrmove^ := pos;
        pcurrmove := BCCapPawn(posmem,pcurrmove);
        pcurrmove^ := pos;
        if (pcb.BMatCnt and $0f00) <> 0 then begin
           pcurrmove := CCapKnight(posmem,BKNIGHT,pcurrmove);
           pcurrmove^ := pos;
        end;
        if (pcb.BMatCnt and $f0f000) <> 0 then begin
           pcurrmove := BCCapBishop(posmem,pcurrmove);
           pcurrmove^ := pos;
        end;
        if (pcb.BMatCnt and $ff0000) <> 0 then
           pcurrmove := BCCapRook(posmem,pcurrmove);
        goto SkipGen;
     end;
  end;
  if pcb.EPSquare <> 255 then begin
     pos := pcb.EPSquare;
     if pboard[pos+17] = BPAWN then begin
        pcurrmove^ := ((pos+17) shl 7 + pos) or
                      (MCAPTURE + MENPASSANT);
        inc(integer(pcurrmove),4);
     end;
     if pboard[pos+15] = BPAWN then begin
        pcurrmove^ := ((pos+15) shl 7 + pos) or
                      (MCAPTURE + MENPASSANT);
        inc(integer(pcurrmove),4);
     end;
  end;
  n := 0;
  while true do begin
     pos := pcb.BLoc[n];
     inc(n);
     if pos = LOCLAST then break;
     posmem := integer(pboard) + pos;
     pcurrmove^ := pos*(128+1);
     case (pboard[pos] and 7) of
      PC_NULL0: continue;
      PC_NULL1: continue;
      PC_PAWN: pcurrmove := BGeneratePawns(pointer(posmem),pcurrmove);
      PC_KNIGHT: pcurrmove := BGenerateRound(pointer(posmem),pcurrmove,@AKNIGHTVEC);
      PC_BISHOP: pcurrmove := BGenerateBishop(pointer(posmem),pcurrmove);
      PC_ROOK: pcurrmove := BGenerateRook(pointer(posmem),pcurrmove);
      PC_QUEEN: begin
         m := pcurrmove^;
         pcurrmove := BGenerateBishop(pointer(posmem),pcurrmove);
         pcurrmove^ := m;
         pcurrmove := BGenerateRook(pointer(posmem),pcurrmove);
      end;
    end;
  end;
SkipGen:
  movecntnk := (integer(pcurrmove) - integer(movelist)) shr 2;
  pcurrmove^ := pcb.BKingPos * (128+1);
  posmem := pcb.BKingPos + integer(pboard);
  if not legal then inc(pcurrmove^,MKINGNLEGAL);
  pcurrmove := BGenerateRound(pointer(posmem),pcurrmove,@AKINGVEC);
  if incheck = 0 then begin
     if (pcb.Castle and BKCASTLE) <> 0
        then pcurrmove := BGenerateKCastle(pcb,pcurrmove);
     if (pcb.Castle and BQCASTLE) <> 0
        then pcurrmove := BGenerateQCastle(pcb,pcurrmove);
  end;
  m := (integer(pcurrmove) - integer(movelist)) shr 2;
  if m > 100 then m := 100;
  pcurrmove := pmovearray;
  j := pcb.BKingPos + 120;
  for i := 0 to m-1 do begin
     mv := movelist[i];
     pmovearray^ := mv;
     inc(integer(pmovearray),4);
     if (xray = 0) and (i < movecntnk) then continue;
     if i < movecntnk then begin
        ds := DIAG16[j - ((mv shr 7) and 127)];
        de := DIAG16[j - (mv and 127)];
        if incheck = 0 then begin
           ds := ds and 15;
           if (ds and xray = 0) or (ds = de) then continue;
           if (mv and $3f80) = pinpos then begin
              dec(integer(pmovearray),4);
              continue;
           end;
        end
        else begin
           if (de and incheck = 0) or ((mv and MCAPTURE) + ds = de) or
              ((mv and MCAPTURE) + de and $ff0f = 0)
           then begin
              dec(integer(pmovearray),4);
              continue;
           end;
        end;
     end;
     if (i < movecntnk) or legal then
        if not BisLegalMoveX(pcb,mv,incheck <> 0) then begin
           pinpos := mv and ($3f80 + MENPASSANT);
           dec(integer(pmovearray),4);
        end;
  end;
  m := (integer(pmovearray) - integer(pcurrmove)) shr 2;
  if m > 0 then result := m
  else begin
     if incheck <> 0 then result := -1 else result := -2;
  end;
end;

function BIsLegalMoveX(pcb:PChessBrd; m:integer; incheck:boolean): boolean;
var
 startpos,endpos: integer;
 endpiece: byte;
 isking: boolean;
 pboard: pbytearray;
begin
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  startpos := (m shr 7) and 127;
  isking := pboard[startpos] = BKING;
  endpos := m and 127;
  endpiece := pboard[endpos];
  if (m and MENPASSANT <> 0) then pboard[endpos+16] := 0;
  pboard[endpos] := pboard[startpos];
  pboard[startpos] := 0;
  if isking
     then result := not BIsAttackedX(pcb,endpos,false)
     else result := not BIsAttackedX(pcb,pcb.BKingPos,not incheck);
  pboard[startpos] := pboard[endpos];
  pboard[endpos] := endpiece;
  if (m and MENPASSANT <> 0) then pboard[endpos+16] := APAWN;
end;

function BIsAttackedX(pcb:PChessBrd; pos:integer; slideonly:boolean): boolean;
var
 mat: integer;
 pb: ^byte;
 t: byte;
 pboard: pbytearray;
const
 COFF = 34;
begin
  result := true;
  mat := pcb.AMatCnt;
  if not slideonly then begin
     pboard := @(pcb.BoardArray[BOARDOFFSET-COFF+pos]);
     if (pboard[COFF-15] = APAWN) or (pboard[COFF-17] = APAWN) then exit;
     if mat and $0f00 <> 0 then if
        (pboard[COFF+14] = AKNIGHT) or (pboard[COFF-14] = AKNIGHT) or
        (pboard[COFF+18] = AKNIGHT) or (pboard[COFF-18] = AKNIGHT) or
        (pboard[COFF+31] = AKNIGHT) or (pboard[COFF-31] = AKNIGHT) or
        (pboard[COFF+33] = AKNIGHT) or (pboard[COFF-33] = AKNIGHT)
        then exit;
     if (pos-pcb.AKingPos <= 17) and (pcb.AKingPos-pos <= 17) and
        (abs((pos and 7)-(pcb.AKingPos and 7)) <= 1) then exit;
  end;
  pboard := @(pcb.BoardArray[BOARDOFFSET+pos]);
  if (pcb.AXCross and ($0f shl ((pos and 7) shl 2))) <> 0 then begin
     pb := pointer(integer(pboard) + -16);
     repeat
       t := pb^;
       dec(integer(pb),16);
     until t <> 0;
     pb := pointer(integer(pboard) + 16);
     if (t = AROOK) or (t = AQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb),16);
     until t <> 0;
     if (t = AROOK) or (t = AQUEEN) then exit;
  end;
  if (pcb.AYCross and ($0f shl ((pos shr 4) shl 2))) <> 0 then begin
     pb := pointer(integer(pboard) + -1);
     repeat
       t := pb^;
       dec(integer(pb));
     until t <> 0;
     pb := pointer(integer(pboard) + 1);
     if (t = AROOK) or (t = AQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb));
     until t <> 0;
     if (t = AROOK) or (t = AQUEEN) then exit;
  end;
  if (mat and $f0f000 <> 0) then begin
     pb := pointer(integer(pboard) + -15);
     repeat
       t := pb^;
       dec(integer(pb),15);
     until t <> 0;
     pb := pointer(integer(pboard) + -17);
     if (t = ABISHOP) or (t = AQUEEN) then exit;
     repeat
       t := pb^;
       dec(integer(pb),17);
     until t <> 0;
     pb := pointer(integer(pboard) + 17);
     if (t = ABISHOP) or (t = AQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb),17);
     until t <> 0;
     pb := pointer(integer(pboard) + 15);
     if (t = ABISHOP) or (t = AQUEEN) then exit;
     repeat
       t := pb^;
       inc(integer(pb),15);
     until t <> 0;
     if (t = ABISHOP) or (t = AQUEEN) then exit;
  end;
  result := false;
end;

function BYRayCross(posmem:integer): integer;
asm
  push ebx
  xor ebx, ebx
  push edi
  lea ecx, [eax - 16]
  push esi
  mov edi, 1
  mov esi, eax
@xs:
  mov dl, byte ptr [ecx]
  sub ecx, 16
  cmp dl, 0
  setnz al
  cmp dl, AROOK
  jz @yend
  add bl, al
  cmp dl, AQUEEN
  jz @yend
  cmp bl, 2
  jnz @xs

  lea ecx, [esi + 32]
@xn:
  mov dl, byte ptr [ecx]
  add ecx, 16
  cmp dl, AROOK
  jz @yend
  cmp dl, AQUEEN
  jz @yend
  cmp dl, NBORDER
  jnz @xn

  xor edi, edi
@yend:
  pop esi
  mov eax, edi
  pop edi
  pop ebx
end;

function BDRayDiag(posmem:integer): integer;
asm
  push edi
  xor edi, edi
  lea ecx, [eax - 17]
  push ebx
  xor ebx, ebx
  push esi
  mov esi, eax

@xne:
  mov dl, byte ptr [ecx]
  sub ecx, 17
  cmp dl, 0
  setnz al
  cmp dl, ABISHOP
  jz @xrayse
  add bl, al
  cmp dl, AQUEEN
  jz @xrayse
  cmp bl, 2
  jz @xse1
  jmp @xne
@xrayne:
  or edi, 4

@xse1:
  lea ecx, [esi + 34]
@xse:
  mov dl, byte ptr [ecx]
  add ecx, 17
  cmp dl, ABISHOP
  jz @xrayse
  cmp dl, AQUEEN
  jz @xrayse
  cmp dl, NBORDER
  jz @xnw1
  jmp @xse
@xrayse:
  or edi, 4

@xnw1:
  lea ecx, [esi - 15]
  mov bl, 0
@xnw:
  mov dl, byte ptr [ecx]
  sub ecx, 15
  cmp dl, 0
  setnz al
  cmp dl, ABISHOP
  jz @xraysw
  add bl, al
  cmp dl, AQUEEN
  jz @xraysw
  cmp bl, 2
  jz @xsw1
  jmp @xnw
@xraynw:
  or edi, 8

@xsw1:
  lea ecx, [esi + 30]
@xsw:
  mov dl, byte ptr [ecx]
  add ecx, 15
  cmp dl, ABISHOP
  jz @xraysw
  cmp dl, AQUEEN
  jz @xraysw
  cmp dl, NBORDER
  jz @xend
  jmp @xsw
@xraysw:
  or edi, 8

@xend:
  pop esi
  mov eax, edi
  pop ebx
  pop edi
end;

function BGenerateRound(posmem,pcurrmove,parray:pointer): pointer;
asm
  push esi
  mov esi, eax
  push edi
  mov edi, edx
  push ebx
  mov eax, [edx]
  push ebp
  mov ebp, ecx
  xor ebx, ebx
  mov edx, [ebp]

@cyc:
  mov dl, byte ptr [esi+edx]
  add ebp, 4
  test dl, B_OR_BORDER
  setz bl
  test dl, A_PIECES
  setnz dl
  mov ecx, [ebp-4]
  and edx, 1
  add ecx, eax
  shl edx, 14
  add ecx, edx
  mov edx, [ebp]
  mov dword ptr [edi], ecx
  lea edi, [edi + ebx*4]
  cmp edx, 0
  jnz @cyc

@end:
  pop ebp
  pop ebx
  mov eax, edi
  pop edi
  pop esi
end;

function BGeneratePawns(posmem:pbytearray; pcurrmove:pinteger): pointer;
var y,themove: integer;
begin
  dec(integer(posmem),50);
  themove := pcurrmove^;
  y := (themove shr 4) and 7;
  if y <= _R3 then begin
     if y = _R2 then begin
        result := BGeneratePromotePawns(posmem,pointer(pcurrmove));
        exit;
     end;
     themove := themove or MPROMOTIONTHREAT;
  end;
  if posmem[50-16] = 0 then begin
     pcurrmove^ := themove - 16;
     inc(integer(pcurrmove),4);
     if y = _R7 then if posmem[50-32] = 0 then begin
        pcurrmove^ := themove - 32;
        inc(integer(pcurrmove),4);
     end;
  end;
  if (posmem[50-17] and A_PIECES) <> 0 then begin
     pcurrmove^ := themove + (MCAPTURE - 17);
     inc(integer(pcurrmove),4);
  end;
  if (posmem[50-15] and A_PIECES) <> 0 then begin
     pcurrmove^ := themove + (MCAPTURE - 15);
     inc(integer(pcurrmove),4);
  end;
  result := pcurrmove;
end;

function BGeneratePromotePawns(posmem:pbytearray; pmoves:pintarray): pointer;
var themove: integer;
begin
  themove := pmoves[0];
  if posmem[50-16] = 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN - 16);
     pmoves[1] := themove + (MPROMOTEDROOK - 16);
     pmoves[2] := themove + (MPROMOTEDBISHOP - 16);
     pmoves[3] := themove + (MPROMOTEDKNIGHT - 16);
     inc(integer(pmoves),16);
  end;
  if (posmem[50-17] and A_PIECES) <> 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN + MCAPTURE - 17);
     pmoves[1] := themove + (MPROMOTEDROOK + MCAPTURE - 17);
     pmoves[2] := themove + (MPROMOTEDBISHOP + MCAPTURE - 17);
     pmoves[3] := themove + (MPROMOTEDKNIGHT + MCAPTURE - 17);
     inc(integer(pmoves),16);
  end;
  if (posmem[50-15] and A_PIECES) <> 0 then begin
     pmoves[0] := themove + (MPROMOTEDQUEEN + MCAPTURE - 15);
     pmoves[1] := themove + (MPROMOTEDROOK + MCAPTURE - 15);
     pmoves[2] := themove + (MPROMOTEDBISHOP + MCAPTURE - 15);
     pmoves[3] := themove + (MPROMOTEDKNIGHT + MCAPTURE - 15);
     inc(integer(pmoves),16);
  end;
  result := pmoves;
end;

function BGenerateRook(posmem,pcurrmove:pointer): pointer;
asm
  push ebx
  xor ebx, ebx
  push edi
  mov ecx, [edx];
  mov edi, edx
  push esi
  mov esi, eax

  lea edx, [esi+16]
  mov eax, ecx
@dn:
  add eax, 16
  cmp byte ptr [edx], 0
  jnz @dsi
  add edx, 16
  mov [edi], eax
  add edi, 4
  jmp @dn
@dsi:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-16]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@ds:
  sub eax, 16
  cmp byte ptr [edx], 0
  jnz @dwi
  sub edx, 16
  mov [edi], eax
  add edi, 4
  jmp @ds
@dwi:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-1]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dw:
  dec eax
  cmp byte ptr [edx], 0
  jnz @dei
  dec edx
  mov [edi], eax
  add edi, 4
  jmp @dw
@dei:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi+1]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@de:
  inc eax
  cmp byte ptr [edx], 0
  jnz @end
  inc edx
  mov [edi], eax
  add edi, 4
  jmp @de
@end:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  pop esi
  lea eax, [edi + ebx*4]
  pop edi
  pop ebx
end;

function BGenerateBishop(posmem,pcurrmove:pointer): pointer;
asm
  push ebx
  xor ebx, ebx
  push edi
  mov ecx, [edx];
  mov edi, edx
  push esi
  mov esi, eax

  lea edx, [esi+17]
  mov eax, ecx
@dne:
  add eax, 17
  cmp byte ptr [edx], 0
  jnz @dswi
  add edx, 17
  mov [edi], eax
  add edi, 4
  jmp @dne
@dswi:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-17]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dsw:
  sub eax, 17
  cmp byte ptr [edx], 0
  jnz @dsei
  sub edx, 17
  mov [edi], eax
  add edi, 4
  jmp @dsw
@dsei:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi-15]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dse:
  sub eax, 15
  cmp byte ptr [edx], 0
  jnz @dnwi
  sub edx, 15
  mov [edi], eax
  add edi, 4
  jmp @dse
@dnwi:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  lea edx, [esi+15]
  mov eax, ecx
  lea edi, [edi + ebx*4]
@dnw:
  add eax, 15
  cmp byte ptr [edx], 0
  jnz @end
  add edx, 15
  mov [edi], eax
  add edi, 4
  jmp @dnw
@end:
  add eax, MCAPTURE
  test byte ptr [edx], A_PIECES;
  setnz bl
  mov [edi], eax

  pop esi
  lea eax, [edi + ebx*4]
  pop edi
  pop ebx
end;

function BGenerateKCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer;
var board: pintarray;
begin
  board := @(pcb.BoardArray[BOARDOFFSET]);
  if (board[29] = BKING + BROOK shl 24) and
     not BisAttackedX(pcb,_F8,false) and
     not BisAttackedX(pcb,_G8,false) then
     begin
        pcurrmove^ := _G8 + (_E8 shl 7) + MCASTLE;
        inc(integer(pcurrmove),4);
     end;
  result := pcurrmove;
end;

function BGenerateQCastle(pcb:PChessBrd; pcurrmove:pinteger): pointer;
var board: pintarray;
begin
  board := @(pcb.BoardArray[BOARDOFFSET]);
  if (board[28] = BROOK) and
     not BisAttackedX(pcb,_D8,false) and
     not BisAttackedX(pcb,_C8,false) then
     begin
        pcurrmove^ := _C8 + (_E8 shl 7) + MCASTLE;
        inc(integer(pcurrmove),4);
     end;
  result := pcurrmove;
end;

function UnderCheckX(pcb:PChessBrd; blacktomove:boolean): integer;
var chk: boolean;
begin
  if blacktomove
     then chk := BisAttackedX(pcb,pcb.BKingpos,false)
     else chk := AisAttackedX(pcb,pcb.AKingpos,false);
  if chk then result := 31 else result := 0;
end;

function CheckMoveX(pcb:PChessBrd; m:integer; btm:boolean): boolean;
var
 movelist: array[0..99] of integer;
 i,n,chk: integer;
begin
  result := false;
  if m = 0 then exit;
  chk := UnderCheckX(pcb,btm);
  if btm
     then n := BGenerateX(pcb,@movelist,chk,true)
     else n := AGenerateX(pcb,@movelist,chk,true);
  if n < 1 then exit;
  for i := 0 to n-1 do
      if (movelist[i] and $3fff) = (m and $3fff) then result := true;
end;

function LegalMovesX(pcb:PChessBrd; btm:boolean): integer;
var
 movelist: array[0..99] of integer;
 n,chk: integer;
begin
  chk := UnderCheckX(pcb,btm);
  if btm
     then n := BGenerateX(pcb,@movelist,chk,true)
     else n := AGenerateX(pcb,@movelist,chk,true);
  result := n;
end;

function MoveToIntSAN(ptcb:PChessBrd; m:integer; var btm:boolean): string;
var
 i,n,startpos,endpos,piece,im,ipiece,istartpos,iendpos,chk: integer;
 pboard: PByteArray;
 movelist: array[0..99] of integer;
 samerank,samecol,iscap: boolean;
begin
  result := '';
  pboard := @(ptcb.BoardArray[BOARDOFFSET]);
  endpos := m and 127;
  startpos := (m shr 7) and 127;
  piece := pboard[startpos];
  if (piece = 0) or (m = 0) then exit;
  btm := (piece and B_PIECES) <> 0;
  iscap := (pboard[endpos] <> 0) or
           (((piece and 7) = PC_PAWN) and (endpos = ptcb.EPSquare));
  if btm
     then n := BGenerateX(ptcb,@movelist,0,false)
     else n := AGenerateX(ptcb,@movelist,0,false);
  if n < 1 then exit;
  samerank := false;
  samecol := false;
  for i := 0 to n-1 do begin
      im := movelist[i];
      iendpos := im and 127;
      if iendpos <> endpos then continue;
      istartpos := (im shr 7) and 127;
      if istartpos = startpos then continue;
      ipiece := pboard[istartpos];
      if ipiece <> piece then continue;
      if (startpos and 7) = (istartpos and 7)
         then samecol := true
         else samerank := true;
  end;
  piece := piece and 7;
  if piece = PC_PAWN then begin
     result := SquareToString(endpos);
     if iscap then result := chr(ord('a') + (startpos and 7)) + 'x' + result;
     if (m and MPROMOTEDQUEEN) <> 0 then
        result := result + '=' + uppercase(FIGURES[(m shr 16) and 3 + 4]);
  end
  else
  if (piece = PC_KING) and (abs(startpos-endpos) = 2) then begin
     if (endpos and 7) = 6 then result := 'O-O' else result := 'O-O-O';
  end
  else begin
     result := uppercase(FIGURES[piece]);
     if samerank then result := result + chr(ord('a') + (startpos and 7));
     if samecol then result := result + inttostr(startpos shr 4 + 1);
     if iscap then result := result + 'x';
     result := result + SquareToString(endpos);
  end;
  BoardPlayMove(ptcb,m);
  chk := UnderCheckX(ptcb,not btm);
  if not btm
     then n := BGenerateX(ptcb,@movelist,chk,true)
     else n := AGenerateX(ptcb,@movelist,chk,true);
  if chk <> 0 then begin
     if n < 0
        then result := result + '#'
        else result := result + '+';
  end
  else
     if n < 0 then result := result + '=';
end;

function PVToStringSAN(pcb:PChessBrd; pvarray:PIntArray): string;
var
 tt: TChessBrd;
 i,m: integer;
 btm: boolean;
begin
  result := '';
  tt := pcb^;
  for i := 0 to 63 do begin
      m := pvarray[i];
      if m = 0 then break;
      result := result + ' ' + MoveToIntSAN(@tt,m,btm);
      if btm then result := result + '.';
  end;
end;

function MoveToSAN(pcb:PChessBrd; m:integer): string;
var
 tt: TChessBrd;
 btm: boolean;
begin
  tt := pcb^;
  result := MoveToIntSAN(@tt,m,btm);
end;

end.


