unit uInput;

interface

//{$DEFINE DelfiDebug}

uses Sysutils, Classes, Windows, uDelfi, uBrdList, uMisc, uGenMove,
     uTable, uEval;

procedure InputMain;


implementation


var
 IsConsole: boolean;
 stih,stoh: cardinal;
 sl,Queue: TStringList;

const
 MAXBUFF = 4096;

procedure WriteData(pbuff:pointer; n:integer);
var nw: cardinal;
begin
  WriteFile(stoh,pbuff^,n,nw,nil);
end;

procedure WriteString(s: string);
begin
  {$IFDEF DelfiDebug}
  sl.Add('>>>> ' + s);
  {$ENDIF}
  s := s + #10;
  WriteData(pchar(s),length(s));
end;

procedure SetupInput;
var mode: cardinal;
begin
  stih := GetStdHandle(STD_INPUT_HANDLE);
  stoh := GetStdHandle(STD_OUTPUT_HANDLE);
  IsConsole := GetConsoleMode(stih,mode);
  if IsConsole then begin
     mode := mode and not (ENABLE_WINDOW_INPUT or ENABLE_MOUSE_INPUT);
     SetConsoleMode(stih,mode);
     FlushConsoleInputBuffer(stih);
  end;
end;

function PeekInput: integer;
var
 n,nr,nl: cardinal;
 i: integer;
 data: array[0..MAXBUFF-1] of char;
begin
  result := 0;
  if IsConsole then begin
     GetNumberOfConsoleInputEvents(stih,n);
     if n > 1 then result := 256;
     exit;
  end;
  PeekNamedPipe(stih,@data,MAXBUFF,@nr,@n,@nl);
  if nr = 0 then exit;
  for i := 0 to nr-1 do
  if (data[i] = #10) or (data[i] = #13) then begin
     result := i+1;
     break;
  end;
  if nr > MAXBUFF then result := MAXBUFF;
end;

function ReadInput(nbytes:integer): string;
var
 n: cardinal;
 data: array[0..MAXBUFF-1] of char;
begin
  ReadFile(stih,data,nbytes,n,nil);
  data[n] := #0;
  result := data;
  result := trim(result);
end;

{ Parse functions }

function ParseUCI(dd:TDelfi; s:string; wf:TWriteFunction): boolean;
var
 n,v: integer;
 s1,s2: string;
begin
  result := true;
  if @wf = nil then exit;
  if s = '' then exit;
  if s = 'uci' then begin
     wf('id name ' + ENGINENAME);
     wf('id author Fabio Cavicchio');
     wf('option name Hash type spin default 4 min 4 max 64');
     wf('option name Elo type spin default 2600 min 1000 max 2600');
     wf('uciok');
     exit;
  end;
  if s = 'isready' then begin
     wf('readyok');
     exit;
  end;
  if pos('position ',s) = 1 then begin
     n := pos('moves ',s);
     if n > 0 then begin
        s2 := copy(s,n+5,length(s));
        s1 := copy(s,1,n-1);
     end
     else begin
        s2 := '';
        s1 := s;
     end;
     n := pos('fen ',s1);
     if n > 0 then s1 := copy(s1,n+4,length(s1)) else s1 := '';
     result := dd.SetPosition(s1,s2);
     exit;
  end;
  if pos('go ',s) = 1 then begin
     if pos('infinite',s) > 0 then begin
        dd.GoAnalyze;
     end
     else begin
        dd.SetTimeUCI(GetIntegerParam(s,'wtime',0),GetIntegerParam(s,'btime',0),
                      GetIntegerParam(s,'winc',0),GetIntegerParam(s,'binc',0),
                     GetIntegerParam(s,'movestogo',-1));
        if pos('ponder',s) > 0
           then dd.GoPonder
           else dd.GoThink;
     end;
     s := MoveToString(dd.bl.PV.PVMove(0));
     n := dd.bl.PV.PVMove(1);
     if (n <> 0) and (not dd.GoBook)
        then s := s + ' pondermove ' + MoveToString(n);
     wf('bestmove ' + s);
     exit;
  end;
  if pos('setoption ',s) = 1 then begin
     s := lowercase(s);
     n := pos('value ',s);
     if n > 1 then begin
        s1 := trim(copy(s,n+6,length(s)));
        v := StrToIntDef(s1,0);
        if pos('name hash',s) > 0 then dd.SetHash(v);
        if pos('name elo',s) > 0 then dd.SetElo(v);
     end;
     if pos('name clearhash',s) > 0 then dd.ClearHash;
     exit;
  end;
  result := false;
end;

function MakeMoveStatusWB(dd:TDelfi; m:integer; wf:TWriteFunction): boolean;
var s: string;
begin
  result := true;
  s := MoveToString(m);
  dd.PlayMove(s);
  wf('move ' + s);
  s := dd.CheckStatus;
  if s <> '' then begin
     wf(s);
     result := false;
  end;
end;

function CheckResignWB(dd:TDelfi; v:integer; wf:TWriteFunction): boolean;
var s: string;
begin
  result := false;
  s := dd.ShouldResign(v);
  if s <> '' then begin
     wf(s);
     result := true;
  end;
end;

function ParseWB(dd:TDelfi; s:string; wf:TWriteFunction): boolean;
var
 n,m,v: integer;
 narr: array[0..3] of integer;
begin
  result := true;
  if @wf = nil then exit;
  if s = '' then exit;
  if s = 'force' then begin
     dd.ForcedMode := true;
     exit;
  end;
  if s = 'remove' then begin
     dd.TakeBack;
     dd.TakeBack;
     exit;
  end;
  if s = 'undo' then begin
     dd.TakeBack;
     if not dd.WBAnalyzing then exit;
     s := 'analyze';
  end;
  if pos('usermove ',s) = 1 then begin
     delete(s,1,9);
     if not dd.PlayMove(s) then exit;
     if dd.LegalMoves < 0 then begin
        s := dd.CheckStatus;
        if (s <> '') and (not dd.WBAnalyzing) and (not dd.ForcedMode)
           then wf(s);
        exit;
     end;
     if not dd.ForcedMode then s := 'go'
     else begin
       if not dd.WBAnalyzing then exit;
       s := 'analyze';
     end;
  end;
  if s = 'analyze' then begin
     dd.ForcedMode := true;
     if dd.LegalMoves < 0 then exit;
     dd.WBAnalyzing := true;
     dd.GoAnalyze;
     exit;
  end;
  if s = 'go' then begin
     dd.ForcedMode := false;
     if dd.LegalMoves < 0 then exit;
     v := dd.GoThink;
     if CheckResignWB(dd,v,wf) then exit;
     if not MakeMoveStatusWB(dd,dd.bl.PV.PVMove(0),wf) then exit;
     if dd.WBPonder then s := 'ponder' else exit;
  end;
  if s = 'ponder' then begin
     repeat
       m := dd.bl.PV.PVMove(1);
       if m = 0 then exit;
       if not CheckMoveX(dd.bl.PCurBrd,m,dd.BlackToMove) then exit;
       if dd.GoBook then exit;
       dd.bl.WBStatPonder := MoveToSAN(dd.bl.PCurBrd,m);
       s := MoveToString(m);
       dd.PlayMove(s);
       dd.WBPonderString := 'usermove ' + s;
       dd.GoPonder;
       dd.WBPonderString := '';
       dd.bl.WBStatPonder := '';
       if dd.bl.PonderStatus <> psHit then break;
       if not MakeMoveStatusWB(dd,dd.bl.PV.PVMove(0),wf) then exit;
     until false;
     dd.TakeBack;
     exit;
  end;
  if s = 'new' then begin
     result := dd.SetPosition('','');
     HashTable.Clear(true);
     dd.ForcedMode := false;
     SetRandom(true);
     exit;
  end;
  if s = 'post' then begin
     dd.PostMode := true;
     exit;
  end;
  if s = 'nopost' then begin
     dd.PostMode := false;
     exit;
  end;
  if s = 'easy' then begin
     dd.WBPonder := false;
     exit;
  end;
  if s = 'hard' then begin
     dd.WBPonder := true;
     exit;
  end;
  if s = '.' then begin
     wf(WBInfoString2(dd.bl));
     exit;
  end;
  if pos('time ',s) = 1 then begin
     n := GetIntegerParam(s,'time',0);
     dd.SetTimeWB(n*10,-1,-1);
     exit;
  end;
  if pos('level ',s) = 1 then begin
     s := copy(s,7,length(s));
     n := ParseNums(s,4,narr);
     if n < 3 then exit;
     narr[1] := narr[1]*60;
     if n = 4 then begin
        inc(narr[1],narr[2]);
        narr[2] := narr[3];
     end;
     dd.SetTimeWB(narr[1]*1000,narr[2]*1000,narr[0]);
     exit;
  end;
  if pos('setboard ',s) = 1 then begin
     delete(s,1,9);
     dd.SetPosition(s,'');
     exit;
  end;
  if pos('ping ',s) = 1 then begin
     s[2] := 'o';
     wf(s);
     exit;
  end;
  if pos('protover ',s) = 1 then begin
     dd.WBMode := true;
     wf('feature myname="' + ENGINENAME + '" variants="normal" ' +
                'usermove=1 setboard=1 ping=1 done=1');
     exit;
  end;
  result := false;
end;

procedure WBPollFunction(pdd:pointer; dostats:boolean);
var
 dd: TDelfi;
 n: integer;
 s: string;
 ismove: boolean;
begin
  dd := pdd;
  if dostats then WriteString(WBInfoStringX(dd.bl));
  n := PeekInput;
  if n = 0 then exit;
  s := ReadInput(n);
  {$IFDEF DelfiDebug}
  sl.Add('*Poll* ' + s);
  {$ENDIF}
  if (s = 'exit') or (s = 'quit') or (s = '?') or (s = 'new') then begin
     Queue.Add(s);
     dd.bl.PonderStatus := psNone;
     dd.bl.AbortSearch := true;
     exit;
  end;
  if (s = '.') or (s = 'post') or (s = 'nopost') or (s = 'easy') or
     (s = 'hard') or (s = 'force') or
     (pos('ping ',s) = 1) or (pos('time ',s) = 1) then begin
     ParseWB(dd,s,WriteString);
     exit;
  end;
  ismove := pos('usermove ',s) = 1;
  if ismove and (dd.bl.PonderStatus = psWait) then begin
     if s = dd.WBPonderString then begin
        dd.bl.PonderStatus := psHit;
     end
     else begin
        Queue.Add(s);
        dd.bl.PonderStatus := psFail;
        dd.bl.AbortSearch := true;
     end;
     exit;
  end;
  Queue.Add(s);
  if dd.WBAnalyzing and ((s = 'undo') or ismove)
  then begin
     dd.bl.AbortSearch := true;
     exit;
  end;
end;

procedure UCIPollFunction(pdd:pointer; dostats:boolean);
var
 dd: TDelfi;
 n: integer;
 s: string;
begin
  dd := pdd;
  if dostats then WriteString(UCIInfoStringX(dd.bl));
  n := PeekInput;
  if n = 0 then exit;
  s := ReadInput(n);
  {$IFDEF DelfiDebug}
  sl.Add('*Poll* ' + s);
  {$ENDIF}
  if (s = 'ponderhit') and (dd.bl.PonderStatus = psWait) then begin
     dd.bl.PonderStatus := psHit;
  end;
  Queue.Add(s);
  if (s = 'stop') or (s = 'quit') then begin
     dd.bl.PonderStatus := psFail;
     dd.bl.AbortSearch := true;
     exit;
  end;
end;

procedure InputMain;
var
 n: integer;
 s: string;
 MDelfi: TDelfi;
begin
  Queue := TStringList.Create;
  sl := TStringList.Create;
  SetupInput;
  MDelfi := TDelfi.Create;
  MDelfi.WBMode := true;
  MDelfi.bl.PollFunction := WBPollFunction;
  MDelfi.bl.PollParam := MDelfi;
  if not MDelfi.LoadIni then begin
     MDelfi.Free;
     MessageBox(0,'Cannot find INI file','ERROR',MB_ICONERROR);
     exit;
  end;
  WriteString('; ' + ENGINENAME + ' - Winboard/UCI chess engine');
  SetRandom(true);
  repeat
     if Queue.Count > 0 then begin
        {$IFDEF DelfiDebug}
        sl.Add('*Queue*');
        {$ENDIF}
        s := Queue[0];
        Queue.Delete(0);
     end
     else begin
        n := PeekInput;
        if n = 0 then begin
           sleep(20);
           continue;
        end;
        s := ReadInput(n);
     end;
     {$IFDEF DelfiDebug}
     sl.Add(s);
     {$ENDIF}
     if s = 'quit' then break;
     if s = 'xboard' then begin
        MDelfi.WBMode := true;
        MDelfi.bl.PollFunction := WBPollFunction;
        MDelfi.SetHash(MDelfi.WBIniHash);
     end;
     if s = 'uci' then begin
        MDelfi.WBMode := false;
        MDelfi.bl.PollFunction := UCIPollFunction;
     end;
     if MDelfi.WBMode
        then ParseWB(MDelfi,s,@WriteString)
        else ParseUCI(MDelfi,s,@WriteString);
  until false;
  {$IFDEF DelfiDebug} 
  sl.SaveToFile('debug.txt');
  {$ENDIF}
  Queue.Free;
  sl.Free;
  MDelfi.Free;
end;

end.
