unit uMisc;

interface

uses SysUtils;

type

 TIntArray = array[0..$12345678] of integer;
 PIntArray = ^TIntArray;
 TByteArray = array[0..$12345678] of byte;
 PByteArray = ^TByteArray;
 PByte = ^Byte;
 PInteger = ^integer;
 TCallBackPV = procedure(const s:string);

const

 MAXELO = 0;
 ENGINENAME: string = 'Delfi 5.1';

 A_PIECES = $10;
 B_PIECES = $08;
 ABDW_PIECES = $04040404;
 P_BORDER = $80;
 P_APAWNS = $40;
 P_BPAWNS = $20;
 PAWNMASK = $40 + $20;
 BPAWNMASK = $20202020;
 APAWNMASK = $40404040;
 ABPAWNMASK = APAWNMASK + BPAWNMASK;
 AB_PIECES = A_PIECES + B_PIECES;
 NBORDER = 1 + P_BORDER;
 A_OR_BORDER = A_PIECES + P_BORDER;
 B_OR_BORDER = B_PIECES + P_BORDER;

 PC_NULL0 = 0;
 PC_NULL1 = 1;
 PC_KING = 2;
 PC_PAWN = 3;
 PC_KNIGHT = 4;
 PC_BISHOP = 5;
 PC_ROOK = 6;
 PC_QUEEN = 7;
 PC_NULL8 = 8;
 PC_NULL9 = 9;
 PC_BKING = 10;
 PC_BPAWN = 11;
 PC_BKNIGHT = 12;
 PC_BBISHOP = 13;
 PC_BROOK = 14;
 PC_BQUEEN = 15;

 AKING =   A_PIECES + 2;
 APAWN =   A_PIECES + 3 + P_APAWNS;
 AKNIGHT = A_PIECES + 4;
 ABISHOP = A_PIECES + 5;
 AROOK =   A_PIECES + 6;
 AQUEEN =  A_PIECES + 7;

 BKING =   B_PIECES + 2;
 BPAWN =   B_PIECES + 3 + P_BPAWNS;
 BKNIGHT = B_PIECES + 4;
 BBISHOP = B_PIECES + 5;
 BROOK =   B_PIECES + 6;
 BQUEEN =  B_PIECES + 7;


 MSOURCE = $000000;
 MDEST   = $000000;
 MCAPTURE = $004000;
 MPROMOTION = $008000;
 MPROMOTEDKNIGHT = MPROMOTION;
 MPROMOTEDBISHOP = MPROMOTION * 3;
 MPROMOTEDROOK = MPROMOTION * 5;
 MPROMOTEDQUEEN = MPROMOTION * 7;
 MENPASSANT = $040000;
 MCASTLE = $080000;
 MPROMOTIONTHREAT = $100000;
 MKINGNLEGAL =  $200000;
 MPAWNQUICAP =  $400000;
 MXUNASSIGNED = $800000;
 MSTARTPIECE = $F000000;

 VPAWN = 95;
 VKNIGHT = 20*16;    // 320
 VBISHOP = 21*16;    // 336
 VROOK = 32*16;      // 512
 VONEPIECE = 35*16;  // 560
 VQUEEN = 63*16;     // 1008
 VKING = 512*16;     // 8192
 VMATE = 10000;
 VALMOSTMATE = 9000;
 VROOKHMOBILITY = 2;
 VROOKHATTACK = 8;
 VKNIGHTATTACK = 5;
 VINFINITE = 999999999;

 CNTPIECES = 256*256*256;
 CNTN = CNTPIECES + 256;
 CNTB = CNTPIECES + 256*16;
 CNTR = CNTPIECES + 256*256;
 CNTQ = CNTPIECES + 256*256*16;

 KCastXCross = (1 shl (5*4)) - (1 shl (7*4));
 QCastXCross = (1 shl (3*4)) - 1;

 _A1 = 0;  _B1 = 1;  _C1 = 2;  _D1 = 3;  _E1 = 4;  _F1 = 5;  _G1 = 6;  _H1 = 7;
 _A2 = 16; _B2 = 17; _C2 = 18; _D2 = 19; _E2 = 20; _F2 = 21; _G2 = 22; _H2 = 23;
 _A3 = 32; _B3 = 33; _C3 = 34; _D3 = 35; _E3 = 36; _F3 = 37; _G3 = 38; _H3 = 39;
 _A4 = 48; _B4 = 49; _C4 = 50; _D4 = 51; _E4 = 52; _F4 = 53; _G4 = 54; _H4 = 55;
 _A5 = 64; _B5 = 65; _C5 = 66; _D5 = 67; _E5 = 68; _F5 = 69; _G5 = 70; _H5 = 71;
 _A6 = 80; _B6 = 81; _C6 = 82; _D6 = 83; _E6 = 84; _F6 = 85; _G6 = 86; _H6 = 87;
 _A7 = 96; _B7 = 97; _C7 = 98; _D7 = 99; _E7 =100; _F7 =101; _G7 =102; _H7 =103;
 _A8 =112; _B8 =113; _C8 =114; _D8 =115; _E8 =116; _F8 =117; _G8 =118; _H8 =119;

 _R1 = 0;  _R2 = 1; _R3 = 2;  _R4 = 3;
 _R5 = 4;  _R6 = 5; _R7 = 6;  _R8 = 7;

 _COLH = 128; _COLG = 64; _COLF = 32; _COLE = 16;
 _COLD = 8;   _COLC = 4;  _COLB = 2;  _COLA = 1;


 FIGURES: array[0..15] of char =
     (' ','?','k','p','n','b','r','q',
      '?','?','k','p','n','b','r','q'
      );

 FIGURESCASE: array[0..15] of char =
     (' ','?','K','P','N','B','R','Q',
      ' ','?','k','p','n','b','r','q');


const

  TO16X: array[0..63] of byte = (
  00, 01, 02, 03, 04, 05, 06, 07,
  16, 17, 18, 19, 20, 21, 22, 23,
  32, 33, 34, 35, 36, 37, 38, 39,
  48, 49, 50, 51, 52, 53, 54, 55,
  64, 65, 66, 67, 68, 69, 70, 71,
  80, 81, 82, 83, 84, 85, 86, 87,
  96, 97, 98, 99, 100,101,102,103,
  112,113,114,115,116,117,118,119
 );

 TO8X: array[0..127] of byte = (
  00, 01, 02, 03, 04, 05, 06, 07, 00, 00, 00, 00, 00, 00, 00, 00,
  08, 09, 10, 11, 12, 13, 14, 15, 00, 00, 00, 00, 00, 00, 00, 00,
  16, 17, 18, 19, 20, 21, 22, 23, 00, 00, 00, 00, 00, 00, 00, 00,
  24, 25, 26, 27, 28, 29, 30, 31, 00, 00, 00, 00, 00, 00, 00, 00,
  32, 33, 34, 35, 36, 37, 38, 39, 00, 00, 00, 00, 00, 00, 00, 00,
  40, 41, 42, 43, 44, 45, 46, 47, 00, 00, 00, 00, 00, 00, 00, 00,
  48, 49, 50, 51, 52, 53, 54, 55, 00, 00, 00, 00, 00, 00, 00, 00,
  56, 57, 58, 59, 60, 61, 62, 63, 00, 00, 00, 00, 00, 00, 00, 00
 );


HRANDOM: array[0..64*16+3] of cardinal = (
$97E72B07, $2F8E9756, $5557ACC6, $5E78761B,
$C0B32EE4, $4795C6C3, $FBAE0A4C, $B2583256,
$73EEC811, $80A47ECB, $551FFF9F, $A941B218,
$091508AF, $2FA889F5, $61B32AA8, $68338AE9,
$EC993873, $1A864401, $2F704315, $4CF2BC7D,
$E9262FC4, $74850311, $0E652D1B, $7BC96113,
$11228D76, $55A3DA89, $B7AE15B1, $4D0FEAA1,
$82CE3362, $5AC66F20, $78D9B035, $FB5D8A14,
$717C91D2, $1427860A, $2867CAB8, $8BC9AF48,
$763F1C2E, $494AFE4D, $48001E8C, $DDD27712,
$C00F3B30, $F32F543C, $C66EFF70, $A285AA00,
$195075C5, $B8CBADA7, $DD377D6D, $B6FF2011,
$0F84AA0A, $B71E2426, $DBECF05E, $7D7BCA2D,
$07AB0111, $3585DAAB, $678D11E5, $357B18B5,
$5CDB0D6E, $3BB98961, $AE1ECD28, $5D5A3E7E,
$59C229B1, $FB6FD0D8, $443AE63C, $0154137C,
$37250C71, $21ECBDF4, $90AF9B4B, $85ACFD63,
$4D43B710, $F78AABEB, $2457B45F, $9335B8F1,
$21266DF0, $548EB860, $46549B37, $43E4A19E,
$F8EA4DE9, $22FC2D8D, $5B3EDDE1, $A13A5AE7,
$B5D20752, $212F554A, $265FEAEC, $1AFDA596,
$F4D1BD32, $9C0E30AB, $8134AFE7, $91A71E87,
$F77E7D97, $520C11BB, $94977B44, $1DC72BEF,
$268A9649, $E9579A1B, $B01DD855, $A5F141FE,
$3394FD68, $46D3A0CB, $27367E91, $08810C10,
$2764ECCB, $E96ECFB2, $A2E2B325, $5597DAB5,
$5157BEEC, $0E5D3846, $B3F555B9, $5A5F3C07,
$A5AEED11, $57064E63, $5DB20CE3, $D28AE04F,
$20F64862, $144B7AA0, $753BB59C, $6DFA47DB,
$CAB78AE4, $A1379EA2, $F9BEBD62, $B75677FF,
$FD918B6D, $11220B89, $82966048, $93B1FB86,
$B44442B6, $04776D22, $A0BC6CDB, $B564B7E4,
$9F5CD398, $B87AFF27, $499DA232, $BF4C9B53,
$EC4BA037, $26F8D74A, $5929369E, $84869F67,
$E7131655, $D1A68AAA, $F435C61A, $E25331FD,
$2375DCDD, $43D3DD42, $D7BA2795, $BAB49EAC,
$2D6F7983, $BCE018F1, $5EACDA5B, $59B39FE0,
$14B977E2, $AF815BB8, $75394CE0, $6845E1BA,
$3E24089F, $2136A050, $54366787, $1E1D5CD1,
$D435CE14, $DF542DEB, $41ACAE6E, $436B5674,
$8F7F40DE, $9A6F6263, $5B253198, $5A699CDA,
$1E2879EE, $E7FED371, $8EEE8034, $D90ED7BE,
$66EDF385, $7F3236A7, $454A43C5, $083D59EF,
$3A967F4E, $9C9F4259, $80DBC8C9, $64818CD8,
$E727C8C0, $E287C964, $3D32EDFC, $C7896B6E,
$E0F238DE, $A246ECD8, $C7862FFC, $2AB3FC9E,
$0B1A0201, $C517C35C, $EB601C9D, $219CF91E,
$1C1FA7E2, $244D1E08, $17E3380F, $BCEEC204,
$DD426360, $816F3D73, $EF12B9C5, $90AFE552,
$DA33F6D3, $FEF4C677, $A3E9AC4E, $11CEE21C,
$31CBF768, $BBF9D252, $2CFC17EB, $D46EEFF8,
$8FAB6411, $AF618E04, $1685F7AE, $0A5DB5F3,
$6036FDD3, $7C1BD830, $479E8313, $64F1D611,
$853C999D, $433FF5CF, $F43AEF4B, $8AF00326,
$92B89201, $459A298C, $301060D1, $08FD4B55,
$92AAE7E4, $C5208CCF, $432D126A, $2991D880,
$91E2DC51, $31F2670B, $334D06CB, $71E0EF0F,
$6D322975, $6E163472, $AA926263, $8869728F,
$6CE60337, $9E537956, $B3D2D3EB, $3D93532E,
$B558DDB4, $3B768BC9, $1AEBB26F, $EC7A3C1A,
$DA186E09, $1B6DF5C0, $A2D7D9CA, $7AAA5F9C,
$1BA1022C, $CCF6F831, $6FDA950F, $44C68842,
$F16BB369, $1B540C1E, $6375B9CB, $036CD2B9,
$0337E8F1, $7816CF2C, $12EDDFDA, $024F3D97,
$2775EFE4, $A001FF76, $5B790102, $EA5514C8,
$BEB6275D, $672CD30A, $A365FCCC, $C62B4970,
$D879409D, $010577D4, $9530D84E, $AB6B0851,
$AF8A65F2, $C46BAA11, $EC0F5630, $7ACA51C6,
$5AFCD1C2, $81D24EF2, $E4F08C10, $56F52D90,
$EE01C842, $3EC4B5D6, $08222284, $19EF9C7A,
$F6256298, $83274B4D, $CCD18EC6, $6FB9358A,
$3F0324DA, $8E4033B7, $B2A9E6FF, $19420541,
$DCA81BF3, $30B38A5F, $89B597D8, $AABCEEC8,
$E5EE0E75, $F8292D98, $6B1ECA5B, $EF1CFF8B,
$3A7CE578, $C4AF9E33, $0F5F9480, $4A2C7DD0,
$1CDF7165, $FD710BA3, $D670B326, $2B394B1F,
$C533DC57, $76D6723B, $AA14E403, $E314C5CD,
$4C1CCF46, $08EC3393, $91277292, $FBDDDCB2,
$F2EBF9B3, $96523BB0, $4E071776, $1DD7E666,
$DBD74C69, $F36F6168, $A772F18D, $4699EFBE,
$8ACE96A4, $3763FE6A, $27748199, $2823271E,
$6B1A5251, $889A5E83, $CC99B4C1, $C1F945DE,
$1D93101B, $8E6957D7, $D1A5E04D, $B7240115,
$F530A507, $A0015642, $97A6931D, $CCF15F46,
$545E52CF, $8631DE6E, $7C4F399C, $F68B4279,
$2062004A, $6029D3FF, $4FBFDDD0, $0BA0E0CA,
$AAD639D8, $C766DCD9, $CF1AF3BC, $A22FF6E5,
$752E6CC8, $C31824E3, $A5E9157D, $FD36970F,
$A628E3DE, $4EBBF699, $6281950A, $B404516F,
$2FD5A581, $85B1107B, $6F07DCC5, $2D31FC78,
$A6F28320, $748F45CF, $2633C92A, $542A6A99,
$77FEDA33, $FFE1DD74, $1CBFE2F2, $0B7B0307,
$B53403DE, $ACEFA80D, $C1F5AA55, $1B2DA21B,
$19FECB5B, $7F345F48, $0BD49D30, $70339502,
$109FD31E, $8ED5E955, $26028FF3, $96FDF4E6,
$4188161C, $1EAF2460, $D39ECE78, $DF97C581,
$A8EF03D2, $6AB50386, $039FFE85, $CF6931BC,
$1BC0458D, $215967DE, $BE256B72, $90773663,
$0201556C, $BCE9318A, $F2055F8A, $88EBB707,
$D29FB123, $35CD823C, $ED9643C7, $56A10F7A,
$C448327C, $178CD144, $11596968, $C9163380,
$C3F7A230, $6555A094, $D4C5EF3D, $152A272D,
$1CD8B099, $94C922C1, $053A5F57, $C65C280E,
$A7004A35, $EBFD412B, $85F3C083, $963D3F0A,
$667B16F8, $855324F2, $97D78C11, $BFD7B9B8,
$1C561C5F, $9323AC02, $2B7AEB99, $395609BA,
$89A554F2, $71EBFB6F, $736F4D8A, $81AC70E5,
$38B6793F, $CB2468C8, $2D038721, $B9318A29,
$B71340B8, $B7085DD5, $2FF43259, $E99F773C,
$2CCCF031, $82B85CD0, $7E718192, $C2139995,
$F7FF686C, $F26E259F, $B9C5159A, $950A9449,
$119CB5DF, $8216AE3B, $41DB3E9F, $8E25014F,
$8E9BDF8A, $5D241569, $8899AC12, $D0B04E0B,
$78654AD8, $18F39C9C, $ED26A2ED, $BD0B1779,
$C84E1153, $7A1FE2E2, $59AEA1A9, $7089EC35,
$E393B61A, $0E107D8B, $9883457D, $6D90F0D2,
$760EC0E3, $EDBCD9F9, $7E64CBBD, $BF2714C2,
$AA1A3521, $0C34EE7B, $7773B0C3, $2CF4F2B6,
$8EA8BD02, $1A5C42E5, $555957ED, $975B807B,
$B1D05642, $91422834, $924CB6C5, $3A2358ED,
$BF39B95B, $21835153, $98B26FEA, $E85A13BE,
$E7E0D389, $B4CB1710, $3F281E33, $33A53A64,
$B9AC7E79, $AA48DCF2, $5A157BF9, $2287A174,
$D127ED0F, $E786F941, $46E30D05, $1F2ED9A6,
$6C6D4465, $4D437CCB, $F7EE9071, $68B76298,
$55FC13A8, $D153A076, $F2EDA061, $7CCE16CE,
$AA0B0E75, $84F5B920, $191D4D5B, $D2E4169D,
$52E92B84, $A622B54E, $C686B751, $3CE3E430,
$7260474C, $31DDED4A, $FBFBFBF0, $1C8DB171,
$525A7960, $03B78908, $F84CC12A, $67E40ED0,
$479F875C, $B373C70D, $83514D02, $94221962,
$9A1A5BBE, $B15057B9, $967E098F, $67894B08,
$E1B6D0E9, $97298877, $C76A70FE, $C96D74F3,
$7A53A6E3, $A792C74C, $0028A059, $735E8FC8,
$6C57113B, $DA83B489, $B77CDFD4, $3B4DDCED,
$C6294774, $A69B26D6, $34085D95, $D39891C2,
$52079C87, $1CBAB0F7, $3F75BCDF, $1E1EC638,
$3A0778E0, $D244533F, $B3D6F28A, $A77EAADA,
$A33865DB, $7666FAE9, $5534B6CB, $5CD517BB,
$482DC176, $B3B33BBA, $3EA60DED, $92FDEA85,
$E208C318, $279F6CD9, $88E228E7, $F43D084B,
$66D70639, $28F0EBD1, $731CDE47, $96DD2D25,
$FE099E13, $C8FE6DB8, $7D8C2EF2, $0B951356,
$777466DF, $7E6B83C1, $35FB1DDF, $419C2B46,
$CC549497, $0BFD1063, $F6523268, $386F940B,
$511E259E, $5E21C806, $E9566C1F, $1C59A62F,
$EBF12EFB, $2D03B54F, $A8B8DD37, $AEC18DE8,
$D35A8B31, $A7D5C9B9, $91D8214F, $05743DD9,
$A90AF2A9, $CBDD035E, $31BD1AC2, $6D27157B,
$71233409, $F57FAB39, $4F1FF55B, $08587D77,
$22026B69, $C2ECB43F, $A68DB7FE, $6D120371,
$2C5F4AED, $B4419A02, $80E6138F, $237C76A8,
$C0B541FE, $C3CD027A, $24C9BC9F, $BCD8AE8D,
$4F5F722A, $14ACEB72, $FE946148, $B4F8A4E1,
$1D846140, $58CC2F4D, $33113C58, $56D47864,
$A4AA6396, $E1C54524, $E86FE9DE, $6A714736,
$4E8AAF9E, $3F7FC739, $BD246EB9, $7CACB709,
$720AD060, $DC8E8C62, $08C81E89, $C53E964B,
$3651C808, $1717D95D, $2077549D, $FA2255D1,
$863F4111, $FE68D537, $CF7D4B99, $F20B84E7,
$3E1FEE39, $0EAB2E9F, $2845A12B, $EB0305DB,
$B13E72BF, $2796A72E, $CA4DA477, $65C359F9,
$8CCFD656, $D45BFCED, $6E4D989F, $E674013D,
$2B113845, $E52385C9, $5F105B71, $31FE13B2,
$6D89FBA3, $1A1CE932, $5610D6BA, $061C1A76,
$5FEE4FA8, $B3BAC927, $02E7C1C0, $E28A2827,
$5D32C467, $97663299, $C736D973, $E7D5E766,
$801C91CD, $42AE7FF8, $76F1D0B7, $EAA25CAA,
$180DDC8B, $10260316, $8E085229, $DED798D9,
$F9038694, $CEC31690, $BE106F17, $E79A3CDE,
$FFD59ACE, $15364697, $03A13EDD, $6039E6E2,
$C3CB1B41, $D063E7BD, $189404DA, $66902F74,
$537B8A12, $4CE94BD6, $9892EB96, $877B5C3B,
$B4226548, $15022273, $52C23F35, $934496A3,
$BF059692, $5CF385EF, $11BEBB66, $2FCEE6CF,
$F1250C9F, $D81AA53C, $677DE736, $7A4A7633,
$FE332864, $F1F3BD36, $BF6ACB35, $4EDDCEFB,
$43F0B93D, $DC9F8C01, $1015898D, $F9344942,
$013D1E47, $96E463C5, $A2733190, $F640756F,
$135FA182, $F1A01F75, $D27BD694, $ADF8ED9D,
$3C16F6A4, $39FBF892, $79E5A7E2, $19A1AA99,
$EE243B9E, $A06D9013, $90BCD231, $221FB99E,
$1EED77A7, $459C8EDA, $E7E32924, $D18756F1,
$3397014C, $DA262443, $C0BDD110, $13D0E783,
$F9A321AF, $4EB4FC02, $EA5225C6, $4AF84682,
$DECE6E98, $BE8AF230, $FC4179ED, $5F6969E5,
$2164E75E, $04809D70, $00385E17, $0A5BD7F8,
$AC802AFD, $9CE7605E, $79C336CB, $DD1D2EFB,
$68EED559, $731F13E8, $F569CBC9, $05297FFC,
$FE759EB9, $25428CE8, $C601AE43, $8C61C6A5,
$6F76DA72, $5E7FC54D, $8AEF15A2, $FC548B12,
$BF30D2AB, $6A3E435C, $3D43B09F, $59C72EE6,
$196189E7, $AF5C7EC5, $6C4CA383, $543C1622,
$1C2BB592, $1177F57A, $EE4A8AA7, $384B971E,
$208D14F5, $CD985A6B, $C04794FB, $9916673F,
$FE383807, $55B5C4FA, $967E1008, $640CEA07,
$B667D691, $B9BD6004, $A12F3138, $987BBB4B,
$C39BB9B4, $0E33D6CB, $971C34A7, $526794F1,
$3E444A04, $BB0A9AE4, $18474ECE, $E544596F,
$B8BE4A8C, $6E3DB121, $3277F738, $F4B81DA0,
$0747B2E9, $0DC9612A, $A8C4AD25, $8680263F,
$3D08B162, $A1F1B937, $C2E37527, $A7C66457,
$5987B7B2, $3DEAF5B8, $626671A9, $A501AF01,
$D0DDB69F, $3AAC32AC, $E024DD1B, $B89D1038,
$30946440, $D52D535A, $ADE38901, $518B9901,
$8D654944, $CB4442A8, $6C7395AA, $D0C89EB5,
$1ED789CF, $9E1B8C20, $3094DC43, $85FEEDF5,
$626E55BA, $3D289888, $2537D68C, $5185F809,
$9C9AF98D, $93A7A487, $32C2D112, $57010E1D,
$9FF5819D, $ADB12BCE, $45154B1F, $D7F47FF1,
$C203E085, $65EE8C21, $95FC6A9D, $F5337A0B,
$F2829A5A, $04ED3C99, $145737CF, $152F5CAD,
$7E264CD3, $BC7DC772, $11175F05, $DE1FF9F7,
$465737C8, $2640D04C, $1FFEB07C, $93A38BD1,
$FDD6C9A4, $5075F746, $26C3C8AF, $4923D0EF,
$819366AD, $875E4384, $A075A9E3, $4E4AC970,
$4DAF9708, $708E3A79, $385AA345, $171AF579,
$50463B6E, $CBC0FD3D, $1BF0B809, $B748BB4A,
$7F63E756, $F089F801, $39FCEFE9, $567ABF72,
$CC72EC93, $9D427A31, $B4FD93B4, $25231C35,
$E0E1F596, $964DD218, $C5F8DF87, $F1D3ECB3,
$08AD5C00, $59DF5D4A, $4FEDB8C5, $EBB88482,
$393A4836, $747FE195, $A76A8452, $87A06AEE,
$4CBE283B, $6E20BEFF, $5EC05798, $B73796BD,
$297FF78B, $98694BB2, $820CD42E, $DC9FD9A2,
$76F9B474, $6F6F4094, $B320EBA1, $63CC42CF,
$51BD3720, $B886D2C8, $335F04B2, $C56B4145,
$F738E2EB, $1C54C08F, $2049433B, $82315149,
$350B959C, $FA5582C8, $628979AC, $701C5EB5,
$8924ABA3, $D04E78C9, $35325A5C, $05FF8003,
$C9A8B0AE, $F384D347, $8E5217CE, $4CF970E4,
$29ED0035, $ADD3E845, $4A42C0FF, $0B11614A,
$6E90CA3E, $9A1C1528, $0836D321, $58316996,
$54F663E6, $65241A75, $112DB4A2, $D7C85485,
$93FAA471, $09D224DD, $2B870DC9, $44ACD8A6,
$7CA1A361, $76D9F67A, $0DE32E9A, $A238355F,
$F1D5218C, $44CD1D5A, $AC9A625B, $D6CB4879,
$AC9F1A17, $F9F57901, $0F02D6DA, $68A8D086,
$0113151A, $3A55063F, $28D732BF, $2788F80A,
$1A048766, $C707DE4D, $530CFBC9, $4BC7140A,
$61AB43FA, $758B3033, $C9133685, $D5663F65,
$DDA22EA8, $146DDBFD, $425AC623, $66E7BACD,
$A6F25EEB, $1E57A9A4, $D1E7D845, $30073EFD,
$549B2AF0, $FAD65C68, $64F7624E, $2ABBF64A,
$530DEE6D, $0192F1F1, $6FC61577, $D3B82989,
$F2377C6F, $354589D8, $E0B19589, $9FEA25EB,
$0DC36D49, $FCE458C4, $887863D1, $8209B8AB,
$23CF82F7, $6A071F08, $040DD1F1, $EC121A4D,
$0BB388A6, $18CF3122, $F283E7B8, $3A8D06B3,
$878D8347, $55ABA529, $AC775E37, $6AD649B8,
$44D37D28, $0C5654DD, $0579259F, $BF881760,
$E83E518D, $2607D079, $E8EB15CF, $5BE09625,
$9A18043E, $2D8FD50C, $3C1E7B2B, $85A4806E,
$4A9AEAB8, $935048C9, $B3B291F7, $D2CF9471,
$E27DF2A1, $DD626561, $5991B894, $20823B59,
$1280570E, $8EA87F62, $17CCFB1F, $F9DD965F,
$2876D784, $1AF035A2, $09326BE2, $D8642BC0
);

function SquareToString(n:integer): string;
function StringToSquare(const s:string): integer;
function MoveToStringEx(m:integer): string;
function PVToStringEx(pvarray: PIntArray): string;
function PVToString(pvarray: PIntArray): string;
function MoveToStringTrace(m:integer): string;
function MoveToString(m:integer): string;
function StringToMove(sm:string): integer;
function InvertMove8(m:integer): integer;
function VFlipSquare8(pos8:integer): integer;
function HFlipSquare8(pos8:integer): integer;
function distboard(pos1,pos2: integer): integer;
procedure MeasureStart;
procedure MeasureAdd;
procedure MeasureSub;
function MeasureGet: int64;
procedure diag4pieces(var ak,bk,az,bz: integer);
function ParseNums(s:string; nmax:integer; var numarr:array of integer): integer;
function GetStringParam(s,sp:string): string;
function GetIntegerParam(s,sp:string; ndefault:integer): integer;
function FindMaxArray(var a:array of integer; n:integer): integer;


implementation


var
 meastime: int64;

function SquareToString(n:integer): string;          // e.g. 119 to h8
var x,y: integer;
begin
  x := n and 7;
  y := n shr 4;
  result := chr(x + ord('a')) + chr (y + ord('1'));
  if (n < 0) or (n > 119) or ((n and 8) <> 0) then result := '';
end;

function StringToSquare(const s:string): integer;    //  e.g. H8 to 119
var x,y: integer;
begin
  result := -1;
  if length(s) <> 2 then exit;
  x := ord(s[1]) - ord('a');
  y := ord(s[2]) - ord('1');
  result := x + y*16;
  if result > 119 then result := -1;
end;

function MoveToStringEx(m:integer): string;
var
 startpos,endpos,n: integer;
 s: string;
begin
  endpos := m and 127;
  startpos := (m shr 7) and 127;
  if (m and MCAPTURE) <> 0 then s := 'x' else s := '';
  result := SquareToString(startpos) + s + SquareToString(endpos);
  if (m and MPROMOTION) <> 0 then begin
     n := (m shr 16) and 3 + 4;
     result := result + FIGURES[n];
  end;
  if m = 0 then result := 'Null';
end;

function PVToStringEx(pvarray: PIntArray): string;
var i,m: integer;
begin
  result := '';
  for i := 0 to 63 do begin
      m := pvarray[i];
      if m = 0 then break;
      result := result + ' ' + MoveToStringEx(m);
  end;
end;

function PVToString(pvarray: PIntArray): string;
var i,m: integer;
begin
  result := '';
  for i := 0 to 63 do begin
      m := pvarray[i];
      if m = 0 then break;
      result := result + ' ' + MoveToString(m);
  end;
end;

function MoveToStringTrace(m:integer): string;
var
 startpos,endpos,n: integer;
 s: string;
begin
  endpos := m and 127;
  startpos := (m shr 7) and 127;
  if (m and MCAPTURE) <> 0 then s := 'x' else s := '-';
  result := SquareToString(startpos) + s + SquareToString(endpos);
  n := (m shr 24) and 7;
  if n <> PC_PAWN then result := FIGURESCASE[n] + result;
  if (m and MPROMOTION) <> 0 then begin
     n := (m shr 16) and 3 + 4;
     result := result + FIGURES[n];
  end;
end;

function MoveToString(m:integer): string;
var startpos,endpos,n: integer;
begin
  endpos := m and 127;
  startpos := (m shr 7) and 127;
  result := SquareToString(startpos) + SquareToString(endpos);
  if (m and MPROMOTION) <> 0 then begin
     n := (m shr 16) and 3 + 4;
     result := result + FIGURES[n];
  end;
end;

function StringToMove(sm:string): integer;
var len,spos,dpos: integer;
begin
  result := -1;
  len := length(sm);
  if len < 4 then exit;
  spos := StringToSquare(copy(sm,1,2));
  dpos := StringToSquare(copy(sm,3,2));
  if (spos < 0) or (dpos < 0) then exit;
  result := (spos shl 7) + dpos;
  if len <> 5 then exit;
  case sm[5] of
   'n': result := result + MPROMOTEDKNIGHT;
   'b': result := result + MPROMOTEDBISHOP;
   'r': result := result + MPROMOTEDROOK;
   'q': result := result + MPROMOTEDQUEEN;
  end;
end;

function VFlipSquare8(pos8:integer): integer;
begin
  result := $38 - (pos8 and $38) + (pos8 and $c7);
end;

function HFlipSquare8(pos8:integer): integer;
begin
  result := (pos8 and $38) + 7 - (pos8 and 7);
end;

function DiagFlipSquare(pos8:integer): integer;
begin
  result := (pos8 and 7) shl 3 + (pos8 shr 3);
end;

function InvertMove8(m:integer): integer;
begin
  result := $0e38 - (m and $0e38) + (m and $7ffff1c7);
end;

function distboard(pos1,pos2: integer): integer;
var dy: integer;
begin
  result := abs((pos1 and 7) - (pos2 and 7));
  dy := abs((pos1 shr 3) - (pos2 shr 3));
  if dy > result then result := dy;
end;

function GetClock: int64; assembler;
asm
  DW $310F
end;

procedure MeasureStart;
begin
  meastime := 0;
end;

procedure MeasureAdd;
begin
  meastime := meastime + GetClock - 25;
end;

procedure MeasureSub;
begin
  meastime := meastime - GetClock;
end;

function MeasureGet: int64;
begin
  result := meastime;
end;

procedure diag4pieces(var ak,bk,az,bz: integer);
var bkx,bky: integer;
begin
  bky := bk shr 3;
  bkx := bk and 7;
  if bky > _R4 then begin
     ak := VFlipSquare8(ak);
     bk := VFlipSquare8(bk);
     az := VFlipSquare8(az);
     bz := VFlipSquare8(bz);
     bky := 7 - bky;
  end;
  if bkx > _R4 then begin
     ak := HFlipSquare8(ak);
     bk := HFlipSquare8(bk);
     az := HFlipSquare8(az);
     bz := HFlipSquare8(bz);
     bkx := 7 - bkx;
  end;
  if bky > bkx then begin
     ak := DiagFlipSquare(ak);
     bk := DiagFlipSquare(bk);
     az := DiagFlipSquare(az);
     bz := DiagFlipSquare(bz);
  end;
end;

function ParseNums(s:string; nmax:integer; var numarr:array of integer): integer;
var
 i,c: integer;
 st: string;
begin
  c := 0;
  st := '';
  s := s + '/';
  for i := 1 to length(s) do begin
      if s[i] in ['0'..'9'] then begin
         st := st + s[i];
         continue;
      end;
      if st <> '' then begin
         numarr[c] := StrToIntDef(st,0);
         st := '';
         inc(c);
      end;
      if c >= nmax then break;
  end;
  result := c;
end;

function GetStringParam(s,sp:string): string;
var n: integer;
begin
  result := '';
  n := pos(sp,s);
  if n < 1 then exit;
  delete(s,1,n+length(sp)-1);
  s := trim(s) + ' ';
  result := copy(s,1,pos(' ',s)-1);
end;

function GetIntegerParam(s,sp:string; ndefault:integer): integer;
begin
  result := StrToIntDef(GetStringParam(s,sp),ndefault);
end;

function FindMaxArray(var a:array of integer; n:integer): integer;
var i,mx: integer;
begin
  result := 0;
  mx := a[0];
  for i := 1 to n-1 do
      if a[i] > mx then begin
         mx := a[i];
         result := i;
      end;
end;

end.
