unit uPriV;

interface

uses Sysutils, uMisc;

const

MAXPVDEPTH = 35;
MAXTREEDEPTH = 31;
MAXQUIDEPTH = 32;
MAXITEDEPTH = 25;

type

TPVLine = record
 Moves: array[0..MAXPVDEPTH] of integer;
 Length: integer;
end;

TPrincipalVar = class
 private
  Data: array[0..MAXPVDEPTH-1] of TPVLine;
  BackLine: TPVLine;
 public
  procedure Store(plydepth,move:integer);
  procedure Reset;
  procedure SetPVOne(move:integer);
  function Size: integer;
  function PVMove(i:integer): integer;
  procedure Leaf(plydepth:integer);
  procedure Backup;
  procedure Restore;
  function PVArray: pointer;
  procedure CopyTo(var arr: array of integer);
end;


implementation


procedure TPrincipalVar.Reset;
var j,i: integer;
begin
  for j := 0 to MAXPVDEPTH-1 do begin
      Data[j].Length := 0;
      for i := 0 to MAXPVDEPTH-1 do Data[j].Moves[i] := 0;
  end;
end;

procedure TPrincipalVar.SetPVOne(move: integer);
begin
  Data[0].Moves[0] := move;
  Data[0].Moves[1] := 0;
  Data[0].Length := 1;
  Data[1].Length := 0;
end;

procedure TPrincipalVar.Store(plydepth,move:integer);
var i,n: integer;
begin
  if plydepth >= MAXPVDEPTH-1 then exit;
  for i := 0 to Data[plydepth+1].Length-1 do
      Data[plydepth].Moves[i+1] := Data[plydepth+1].Moves[i];
  Data[plydepth].Moves[0] := move;
  n := Data[plydepth+1].Length;
  Data[plydepth].Length := n + 1;
  Data[plydepth].Moves[n+1] := 0;
end;

procedure TPrincipalVar.Leaf(plydepth: integer);
begin
  if plydepth >= MAXPVDEPTH then exit;
  Data[plydepth].Length := 0;
  Data[plydepth].Moves[0] := 0;
end;

function TPrincipalVar.Size: integer;
begin
  result := Data[0].Length;
end;

function TPrincipalVar.PVMove(i:integer): integer;
begin
  if i >= Data[0].Length then result := 0
     else result := Data[0].Moves[i];
end;

function TPrincipalVar.PVArray: pointer;
begin
  result := @(Data[0].Moves);
end;

procedure TPrincipalVar.Backup;
var i: integer;
begin
  for i := 0 to MAXPVDEPTH-1 do BackLine.Moves[i] := Data[0].Moves[i];
  BackLine.Length := Data[0].Length;
end;

procedure TPrincipalVar.Restore;
var i: integer;
begin
  for i := 0 to MAXPVDEPTH-1 do Data[0].Moves[i] := BackLine.Moves[i];
  Data[0].Length := BackLine.Length;
end;

procedure TPrincipalVar.CopyTo(var arr: array of integer);
var i,n: integer;
begin
  n := Size;
  for i := 0 to n-1 do arr[i] := Data[0].Moves[i];
  arr[n] := 0;
end;

end.
