unit uTable;

interface

uses Sysutils, uBoard, uGenMove, uPriV, uMisc, Windows;

type

THashTable = class
 private
  table,ftable: PIntArray;
  tablesize,blackoff,maskoff,counter,lastalloc: integer;
 public
  LastMove: integer;
  CntCollision,CntProbe,CntHit: integer;
  HistEval,HistCount,HistGood: array[0..128*16*2] of integer;
  constructor Create;
  destructor Destroy; override;
  procedure Clear(full:boolean);
  procedure Alloc(mb:integer);
  procedure StorePosition(pcb:PChessBrd; depth,kind,score,move,
                          plydepth:integer; btm:boolean);
  function GetPosition(pcb:PChessBrd; beta,depth,plydepth:integer; btm:boolean;
                       var kind,score,move:integer): boolean;
  function GetHistory(m:integer; pcb:PChessBrd): integer;
  procedure IncHistory(m,depth,nodekind: integer; pcb:PChessBrd);
  function PruneHistory(m:integer; pcb:PChessBrd): boolean;
  procedure ClearHistory;
  procedure InsertPV(pcb:PChessBrd; pv:TPrincipalVar; btm:boolean);   
end;

var
 HashTable: THashTable = nil;


implementation


const

 TTDATA = 0;
 TTMOVE = 1;
 TTNEXTDATA = 2;
 TTNEXTMOVE = 3;
 TTPV = 0;
 TTFHIGH = 1;
 TTFLOW = 2;
 TTEXACT = 3;
 TTNONULL = 4;
 TTNOTHING = 5;
 TTMOVEMASK = $efdc0000;
 TTCNT0 = $00000000;
 TTCNT1 = $00200000;
 TTCNT2 = $10000000;
 TTCNT3 = $10200000;
 TTCNTMASK = TTCNT3;
 HBMASK = $fff00000;
 HPMASK = $000fffe0;


 // ++++++++++++++  TChessTTable +++++++++++++++


constructor THashTable.Create;
begin
  inherited Create;
  ftable := nil;
  table := nil;
  counter := TTCNT1;
  lastalloc := 0;
end;

destructor THashTable.Destroy;
begin
  if ftable <> nil then FreeMem(ftable);
  inherited Destroy;
end;

procedure THashTable.Alloc(mb: integer);
begin
  if mb = lastalloc then exit;
  lastalloc := mb;
  if ftable <> nil then begin
     FreeMem(ftable);
     ftable := nil;
  end;
  if mb <= 4 then mb := 4;
  if mb >= 64 then mb := 64;
  case mb of
   4..7: mb := 4;
   8..15: mb := 8;
   16..31: mb := 16;
   32..63: mb := 32;
  end;
  tablesize := 1048576 * mb;
  blackoff := (tablesize div 16) div 2;
  maskoff := blackoff-1;
  GetMem(ftable,tablesize + 512);
  table := pointer((cardinal(ftable) + 255) and $ffffff00);
end;

procedure THashTable.Clear(full:boolean);
begin
  if full then begin
     FillChar(table[0],tablesize,0);
     counter := TTCNT1;
  end
  else begin
     if counter = TTCNT3 then counter := TTCNT0
       else begin
         if counter = TTCNT2 then counter := TTCNT3;
         if counter = TTCNT1 then counter := TTCNT2;
         if counter = TTCNT0 then counter := TTCNT1;
       end;
  end;
  CntCollision := 0;
  CntProbe := 0;
  CntHit := 0;
end;

procedure THashTable.IncHistory(m,depth,nodekind: integer; pcb:PChessBrd);
var
 idx,i,pc,k: integer;
 pboard: PByteArray;
const
 FAILLOW = 128;
begin
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  pc := integer(pboard[m shr 7 and 127]) and 15;
  idx := (m and 127) shl 4 + pc;
  if (m and (MCAPTURE + MPROMOTION)) <> 0 then inc(idx,128*16)
  else begin
    if HistCount[idx] + 1 > 16000 then
       for i := 0 to 128*32-1 do begin
         HistCount[i] := HistCount[i] shr 1 + 1;
         HistGood[i] := HistGood[i] shr 1 + 1;
       end;
    inc(HistCount[idx]);
    if nodekind = FAILLOW then exit;
    inc(HistGood[idx]);
  end;
  if depth <= 0 then exit;
  k := depth*depth;
  if HistEval[idx] + k > 16000 then
     for i := 0 to 128*32-1 do
         HistEval[i] := HistEval[i] shr 1;
  inc(HistEval[idx],k);
end;

function THashTable.GetHistory(m:integer; pcb:PChessBrd): integer;
var
 idx,pc: integer;
 pboard: PByteArray;
begin
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  pc := integer(pboard[m shr 7 and 127]) and 15;
  idx := (m and 127) shl 4 + pc;
  if (m and (MCAPTURE + MPROMOTION)) <> 0 then inc(idx,128*16);
  result := HistEval[idx];
end;

function THashTable.PruneHistory(m:integer; pcb:PChessBrd): boolean;
var
 idx,pc: integer;
 pboard: PByteArray;
begin
  pboard := @(pcb.BoardArray[BOARDOFFSET]);
  pc := integer(pboard[m and 127]) and 15;
  idx := (m and 127) shl 4 + pc;
  if (m and (MCAPTURE + MPROMOTION)) <> 0 then inc(idx,128*16);
  result := (HistGood[idx] shl 16) div HistCount[idx] < 15000;  // 0..65535
end;

function THashTable.GetPosition(pcb:PChessBrd; beta,depth,plydepth:integer;
                                btm:boolean; var kind,score,move:integer): boolean;
var
 offs,h,hashdepth,vdata,vmove: integer;
 pdata: pintarray;
begin
  result := false;
  LastMove := 0;
  inc(CntProbe);
  kind := TTNOTHING;
  move := 0;
  with pcb^ do h := HashB1 xor EPSquare xor Castle;
  if h = 0 then exit;
  offs := h and maskoff;
  if btm then inc(offs,blackoff);
  pdata := pointer(integer(table) + (offs shl 4));
  if ((h xor pdata[TTDATA]) and HBMASK) <> 0 then begin
     if ((h xor pdata[TTNEXTDATA]) and HBMASK) <> 0 then exit;
     inc(integer(pdata),8);
  end;
  vdata := pdata[TTDATA];
  if ((pcb.HashB2 xor vdata) and HPMASK) <> 0 then begin
     inc(CntCollision);
     exit;
  end;
  vmove := pdata[TTMOVE];
  move := (vmove and integer(TTMOVEMASK)) shr 18;
  score := (vmove and $7fff) - 16384;
  kind := (vmove shr 16) and $3;
  if kind = TTEXACT then begin
     if score > VALMOSTMATE then dec(score,plydepth);
     if score < -VALMOSTMATE then inc(score,plydepth);
  end;
  if kind = TTPV then exit;
  hashdepth := vdata and 31;
  if depth > hashdepth then begin
     if depth > hashdepth+4 then move := 0;
     if (depth <= hashdepth+2) and (kind <> TTFHIGH) and (score <= beta)
        then kind := TTNONULL;
     exit;
  end;
  inc(CntHit);
  result := (vmove and TTCNTMASK) = counter;
end;

procedure THashTable.StorePosition(pcb:PChessBrd; depth,kind,score,
                                   move,plydepth:integer; btm:boolean);
var
 offs,h: integer;
 pdata: pintarray;
 ok1,ok2: boolean;
label
 skip;
begin
  LastMove := move;
  if depth < 0 then exit;
  if (score > 11000) or (score < -11000) then exit;
  if score > VALMOSTMATE then begin
     if kind = TTFLOW then exit;
     if kind = TTEXACT then inc(score,plydepth) else score := VALMOSTMATE;
     inc(depth);
  end;
  if score < -VALMOSTMATE then begin
     if kind = TTFHIGH then exit;
     if kind = TTEXACT then dec(score,plydepth) else score := -VALMOSTMATE;
     inc(depth);
  end;
  with pcb^ do h := HashB1 xor EPSquare xor Castle;
  offs := h and maskoff;
  if btm then inc(offs,blackoff);
  pdata := pointer(integer(table) + (offs shl 4));
  h := (h and integer(HBMASK)) + (pcb.HashB2 and HPMASK);
  if (h xor pdata[TTDATA]) and (HBMASK+HPMASK) = 0 then goto skip;
  if (h xor pdata[TTNEXTDATA]) and (HBMASK+HPMASK) = 0 then begin
     inc(integer(pdata),8);
     goto skip;
  end;
  ok1 := (pdata[TTMOVE] and TTCNTMASK) = counter;
  ok2 := (pdata[TTNEXTMOVE] and TTCNTMASK) = counter;
  if ok1 and ok2 then begin
     if ((pdata[TTDATA] and 31) > (pdata[TTNEXTDATA] and 31)) and (kind <> TTPV)
        then inc(integer(pdata),8);
  end
  else begin
     if ok1 then inc(integer(pdata),8);
  end;
skip:
  if depth > 31 then depth := 31;
  pdata[TTDATA] := h + depth;
  pdata[TTMOVE] := ((move shl 18) and integer(TTMOVEMASK)) +
                   ((kind and $3) shl 16) + counter + score + 16384;
end;

procedure THashTable.InsertPV(pcb:PChessBrd; pv:TPrincipalVar; btm:boolean);
var
 i,m: integer;
 cboard: TChessBrd;
begin
  cboard := pcb^;
  if pv.Size <= 0 then exit;
  for i := 0 to pv.Size-1 do begin
      m := pv.PVMove(i);
      if not CheckMoveX(@cboard,m,btm) then break;
      StorePosition(@cboard,0,TTPV,0,m,0,btm);
      BoardPlayMove(@cboard,m);
      btm := not btm;
  end;
end;

procedure THashTable.ClearHistory;
var i: integer;
begin
  for i := 0 to High(HistEval) do begin
      HistEval[i] := 0;
      HistCount[i] := 1;
      HistGood[i] := 1;
  end;
end;

initialization
 HashTable := THashTable.Create;
 HashTable.Alloc(4);

finalization
 if HashTable <> nil then HashTable.Free;

end.
